import pdb
import re
import os
import sys
import collections
import errno
import logging
from pyang import plugin
from string import Template
import time
import optparse
import ordered_set
import ordereddict

C_TYPE = ('SAFplusTypes', 'c-type')

C_EXISTING = ('SAFplusTypes', 'c-existing')

yang_stmts = ['container', 'list', 'notification']

leaf_stmts = ['leaf', 'leaf-list']

module_stmts = ['module', 'submodule']

node_stmts = module_stmts + yang_stmts + leaf_stmts

mapping_stmts_client = {'notification':'SAFplus::MgtNotify', 'container':'SAFplus::MgtContainer', 'grouping':'SAFplus::MgtContainer', \
                          'list':'SAFplus::MgtContainer', 'leaf':'SAFplus::MgtProv', 'rpc': 'SAFplus::MgtRpc', 'module': 'SAFplus::MgtModule'}
# ^ List maps to MgtContainer because a list is really a list of containers.  So the type of the instance is MgtList<key>, but the superclass of the generated class is a Container.

class_hierarchy = {}
"""Dict that map namespace names to sets of names of classes to be generated"""

type2include = { 'int8_t':'#include <cstdint>',
'int16_t':'#include <cstdint>', 
'int32_t':'#include <cstdint>', 
'int64_t':'#include <cstdint>', 
'uint8_t':'#include <cstdint>', 
'uint16_t':'#include <cstdint>', 
'uint32_t':'#include <cstdint>', 
'uint64_t':'#include <cstdint>',
}

# Yang standard modules:
yangStandardRFC = ['ietf-interfaces','ietf-netconf-monitoring','ietf-system']

# uses the types defined in stdint.h
intY2C = { 'int8':'int8_t', 'int16':'int16_t', 'int32':'int32_t', 'int64':'int64_t', 'uint8':'uint8_t', 'uint16':'uint16_t', 'uint32':'uint32_t', 'uint64':'uint64_t','decimal64':'int64_t' }

# Maps type names to include file names
mapping_include = {'SAFplus::Transaction': 'clTransaction', 'std::string': 'string', 'std::bitset':'bitset', 'SAFplus::MgtProv':'clMgtProv', 'SAFplus::MgtContainer': 'clMgtContainer',\
                     'SAFplus::MgtRpc': 'clMgtRpc', 'SAFplus::MgtObject': 'clMgtObject', 'SAFplus::MgtNotify':'clMgtNotify', 'SAFplus::MgtProvList':'clMgtProvList', \
                     'SAFplus::MgtFactory':'MgtFactory', 'SAFplus::MgtEnumType':'MgtEnumType', 'SAFplus::MgtList': 'clMgtList', 'SAFplus::MgtDatabase': 'clMgtDatabase', \
                     'SAFplus::MgtIdentifier':'clMgtIdentifier', 'SAFplus::MgtIdentifierList':'clMgtIdentifierList', 'SAFplus::MgtModule':'clMgtModule', 'SAFplus::ClPlugin':'clPluginApi', \
                     'SAFplus::MgtHistoryStat':'clMgtHistoryStat'}

mapping_namespace = {'std': ['array', 'bitset', 'deque', 'forward_list', 'list', \
                               'map', 'queue', 'set', 'stack', 'unordered_map', \
                               'unordered_set', 'vector', 'algorithm', 'iterator', \
                               'memory', 'stdexcept', 'locale', 'string', 'ostream', 'iostream','istream']}

mapping_include_boost_lib = {'boost::container::set': 'boost/container/set.hpp'}

mapping_typedef_enum = {}

cstdTypes = ['bool','long int', 'int', 'short int', 'char', 'string', 'long', "float", "double"]
cstdintTypes = [ 'int8_t', 'int16_t', 'int32_t', 'int64_t', 'uint8_t', 'uint16_t', 'uint32_t', 'uint64_t', 'int64_t' ]

""" This mapping separate for currently module generator with import/prefix other modules
Example: 
mapping_module_import_prefix = {'<currently>': {'imported':'prefix'} }
"""
mapping_module_import_prefix = {}

CUSTOMIZE_STR_FOR_COMMON_HEADER = "/* This file will not be changed once created and is for custom extensions */"

""" 
File open mode constant
"""
F_WRITE=1
F_OVERWRITE=2
F_APPEND=3

camelized_stmt_args = {}
normalized_stmt_args = {}

def isNumber(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def ns2Filename(s):
  """Convert a namespace to a filename"""
  lst = s.split("::")
  if lst[0] == "": del lst[0]  # Remove global namespace indicator
  return "".join(lst)
  

def nsInc(s):
  try:
    t = s.split("::")
    include = t[-1]
    namespace = "::".join(t[0:-1])
  except ValueError, e:  # there is no ::
    namespace = None
    include = s
  return namespace, include
  

def enumize(x):
  value = search(x,'value')
  if value:
    assert(len(value)==1)  # having 2 values makes no sense
    return (camelize(x.arg),value[0].arg)
  else:
    return (camelize(x.arg),None)

def getBaseType(stmt):
    """Returns the built in type that stmt is derived from"""
    if stmt.keyword == 'type' and stmt.arg == 'union':
        return stmt
    type_stmt = search_one(stmt, 'type')
    if type_stmt is None:
        return stmt
    try:
        typedef = type_stmt.i_typedef
    except AttributeError:
        return type_stmt
    else:
        if typedef is not None:
            return getBaseType(typedef)
        else:
            return type_stmt

def leafType(stmt,ctx):
  nativeType = search_one(stmt, C_EXISTING)
  if nativeType:
    return (None,nativeType.arg)
  ctype = search_one(stmt, C_TYPE)
  if ctype:
    return (ctype.arg,ctype.arg)
  yangtype =  search_one(stmt, 'type') 
  return getTypes(yangtype,ctx) 

def cType(stmt,ctx):
  namespace = getNamespace(stmt,ctx)
  nativeType = search_one(stmt, C_EXISTING)
  if nativeType:
    theType = (None,nativeType.arg)
  else:
    theType = [namespace + "::" + capitalizeFirst(camelize(stmt.arg)), capitalizeFirst(camelize(stmt.arg))]
  return theType


def getTypes(yang_type, ctx):
    if yang_type.keyword in leaf_stmts:
        yang_type = search_one(yang_type, 'type')
    assert yang_type.keyword in ('type', 'typedef'), 'argument is type, typedef or leaf'
    if yang_type.arg == 'leafref':
        return getTypes(yang_type.parent.i_leafref.i_target_node, ctx)

    include = None
    primitive = normalize(yang_type.arg)

    if yang_type.arg in ('boolean',):
        primitive = 'bool'
    elif yang_type.arg in ('string', 'binary', 'union', 'empty',
                           'instance-identifier', 'identityref'):
        include = 'std::string'
        primitive = 'string'
    elif yang_type.arg in ('enumeration',):
        primitive = mapping_typedef_enum[normalize(getParent(yang_type).arg)]
        include = primitive
    elif yang_type.arg in ('bits',):  # uint64 handled below
        include = 'std::bitset'
        primitive = 'bitset<%d>' % len(yang_type.substmts)
#    elif yang_type.arg == 'decimal64':
#        primitive = 'long'
    elif yang_type.arg in intY2C.keys():  # ('int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64'):
        include = primitive = intY2C[yang_type.arg]
        if not "::" in include:  # if no namespace is specified assume global ns
          include = "::" + include
          
    else:
        try:
            primitive = mapping_typedef_enum["%s" %yang_type.arg]
            include = primitive
        except:
            try:
                typedef = yang_type.i_typedef
                if typedef is None: raise AttributeError()
            except AttributeError:
                if yang_type.keyword == 'typedef':
                    primitive = normalize(yang_type.arg)
            else:
                if getParent(typedef).keyword in ('module', 'submodule'):
                    namespace = getNamespace(typedef, ctx)
                    typedef_arg = normalize(typedef.arg)
                    #Check to use c++ defined type
                    if search_one(typedef, C_EXISTING) is None:
                        include = namespace + '::' + typedef_arg
                    primitive = typedef_arg
                else:
                    basetype = getBaseType(typedef)
                    include, primitive = getTypes(basetype, ctx)
    return include, primitive

def write_file(d, file_name, file_content, ctx, mode = F_OVERWRITE):
    """Creates the directory d if it does not yet exist and writes a file to it
    named file_name with file_content in it.

    Copyright 2012 Tail-f Systems AB

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License. 

   """
    logging.debug("%s in %s" % (file_name, d))
    d = d.replace('.', os.sep)
    wd = os.getcwd()
    try:
        os.makedirs(d, 0o777)
    except OSError as exc:
        if exc.errno == errno.EEXIST:
            pass  # The directory already exists
        else:
            raise
    try:
        os.chdir(d)
    except OSError as exc:
        if exc.errno == errno.ENOTDIR:
            print_warning(msg = ('Unable to change directory to ' + d +
                '. Probably a non-directory file with same name as one of ' +
                'the subdirectories already exists.'), key = d, ctx = ctx)
        else:
            raise
    finally:
        os.chdir(wd)

    f = None
    if mode == F_OVERWRITE:
        f = open(d + os.sep + file_name, 'w+')
    elif mode == F_WRITE:
        #only create if it does not exist
        try:
            open(d + os.sep + file_name)
        except IOError:
            f = open(d + os.sep + file_name, 'w+')
    elif mode == F_APPEND:
        #only write if it is existing
        try:
            f = open(d + os.sep + file_name)
        except IOError:
            pass

    if f is not None:
        if isStringTypes(file_content):
            f.write(file_content)
        else:
            for line in file_content:
                f.write(line)
                f.write('\n')

def search(stmt, keywords):
    """Utility for calling Statement.search. If stmt has an i_children
    attribute, they are searched for keywords as well.

    stmt     -- Statement to search for children in
    keywords -- A string, or a tuple, list or set of strings, to search for

    Returns a set (without duplicates) of children with matching keywords.
    If choice or case is not in keywords, substatements of choice and case
    are searched as well.

    Copyright 2012 Tail-f Systems AB

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License. 

   """

    if isStringTypes(keywords):
        keywords = keywords.split()
    bypassed = ('choice', 'case')
    bypass = all(x not in keywords for x in bypassed)
    dict_ = ordereddict.OrderedDict()

    def iterate(children, acc):
        for ch in children:
            if bypass and ch.keyword in bypassed:
                _search(ch, keywords, acc)
                continue
            try:
                key = ' '.join([ch.keyword, ch.arg])
            except TypeError:
                if ch.arg is None:  # Extension
                    key = ' '.join(ch.keyword)
                else:
                    key = ' '.join([':'.join(ch.keyword), ch.arg])
            if key in acc:
                continue
            for keyword in keywords:
                if ch.keyword == keyword:
                    acc[key] = ch
                    break

    def _search(stmt, keywords, acc):
        if any(x in keywords for x in ('typedef', 'import',
                                       'augment', 'include')):
            old_keywords = keywords[:]
            keywords = ['typedef', 'import', 'augment', 'include']
            iterate(stmt.substmts, acc)
            keywords = old_keywords
        try:
            iterate(stmt.i_children, acc)
        except AttributeError:
            iterate(stmt.substmts, acc)

    _search(stmt, keywords, dict_)
    return list(dict_.values())

def search_one(stmt, keyword, arg = None):
    """
    Utility for calling Statement.search_one, including i_children.
    
    Copyright 2012 Tail-f Systems AB

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
        http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License. 

   """

    res = stmt.search_one(keyword, arg = arg)
    if res is None:
        try:
            res = stmt.search_one(keyword, arg = arg, children = stmt.i_children)
        except AttributeError:
            pass
    if res is None:
        try:
            return search(stmt, keyword)[0]
        except IndexError:
            return None
    return res

def is_config(stmt):
    """
    Returns True if stmt is a configuration data statement

    Copyright 2012 Tail-f Systems AB

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
        http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License. 

   """

    config = None
    while config is None and stmt is not None:
        if stmt.keyword == 'notification':
            return False # stmt is not config if part of a notification tree
        config = search_one(stmt, 'config')
        stmt = getParent(stmt)
    return config is None or config.arg == 'true'

def pairwise(iterable):
    """Returns an iterator that includes the next item also"""
    iterator = iter(iterable)
    item = next(iterator)  # throws StopIteration if empty.
    for next_item in iterator:
        yield (item, next_item)
        item = next_item
    yield (item, None)

def capitalizeFirst(string):
    """Returns string with its first character capitalized (if any)"""
    return string[:1].capitalize() + string[1:]

def lowerFirst(string):
    """Returns string with its first character capitalized (if any)"""
    return string[:1].lower() + string[1:]

def flatten(l):
    """Returns a flattened version of iterable l

    l must not have an attribute named values unless the return value values()
    is a valid substitution of l. Same applies to all items in l.

    Example: flatten([['12', '34'], ['56', ['7']]]) = ['12', '34', '56', '7']
    """
    res = []
    while hasattr(l, 'values'):
        l = list(l.values())
    for item in l:
        try:
            assert not isStringTypes(item)
            iter(item)
        except (AssertionError, TypeError):
            res.append(item)
        else:
            res.extend(flatten(item))
    return res

def camelize(string):
    """Removes hyphens and dots and replaces following character (if any) with
    its upper-case counterpart. Does not remove a trailing hyphen or dot.
    Returns an empty string if string argument is None.
    """
    try:  # Fetch from cache
        return camelized_stmt_args[string]
    except KeyError:
        pass
    camelized_str = collections.deque()
    if string is not None:
        iterator = pairwise(string)
        for character, next_character in iterator:
            if next_character and character in '-.':
                camelized_str.append(capitalizeFirst(next_character))
                next(iterator)
            else:
                camelized_str.append(character)

    res = ''.join(camelized_str)

    if re.match(r'\d', res):
        camelized_str.appendleft('_')
    res = ''.join(camelized_str)
    camelized_stmt_args[string] = res  # Add to cache
    return res

def normalize(string):
    """returns capitalizeFirst(camelize(string)), except if camelize(string)
    begins with and/or ends with a single underline
    """
    try:  # Fetch from cache
        return normalized_stmt_args[string]
    except KeyError:
        pass
    res = capitalizeFirst(camelize(string))
    normalized_stmt_args[string] = res  # Add to cache
    return res

"""
C++ keywords does not contain any specific characters
"""
def getCplusCplusString(str):
    return re.sub('[^a-zA-Z0-9]', '', str)

def getNamespace(stmt, ctx):
    rootpkg = ''
    parent = getParent(stmt)
    while parent is not None:
        stmt = parent
        parent = getParent(stmt)
        rootpkg = camelize(stmt.arg)
    return rootpkg

def getModule(stmt):
    """Returns the module to which stmt belongs to"""
    if stmt.top is not None:
        return getModule(stmt.top)
    elif stmt.keyword == 'module':
        return stmt
    else:  # stmt.keyword == 'submodule':
        belongs_to = search_one(stmt, 'belongs-to')
        for (module_name, revision) in stmt.i_ctx.modules:
            if module_name == belongs_to.arg:
                return stmt.i_ctx.modules[(module_name, revision)]

def getParent(stmt):
    """Returns closest parent which is not a choice, case or submodule
    statement. If the parent is a submodule statement, the corresponding main
    module is returned instead.

    """
    if stmt.parent is None:
        return None
    elif stmt.parent.keyword == 'submodule':
        return getModule(stmt)
    elif stmt.parent.parent is None:
        return stmt.parent
    elif stmt.parent.keyword in ('choice', 'case'):
        return getParent(stmt.parent)
    else:
        return stmt.parent

def fullPath(stmt):
    pathsep = "/"
    path = stmt.arg
    if (stmt.keyword != 'grouping') and (stmt.keyword != 'choice') and (stmt.keyword != 'case'):
        while stmt.parent is not None:
            stmt = stmt.parent
            if stmt.arg is not None:
                path = stmt.arg + pathsep + path
    return path

"""
Using full namespace for className
format [namespace::className, className]
"""
def getFullNamespace(returnType):
    #Search in class_hierarchy
    for module in class_hierarchy.keys(): 
        if getCplusCplusString(returnType[1]) in class_hierarchy[module]:
            return [module + "::" + returnType[1], returnType[1]]

    try:
        typedef = mapping_typedef_enum[returnType[1]]
        return [typedef, returnType[1]]
    except:
        return returnType

"""
Check if type is standard namespace 
"""
def isNamespaceStd(type):
    str = type
    try:
        str = type.split("::")[1]
    except:
        pass

    """ Remove all specific characters """
    str = re.sub('[^a-zA-Z]', '', str)
    if str in mapping_namespace['std']:
        return True
    return False

def isStringTypes(val):
    if isinstance(val, str):
        return True
    if isinstance(val, unicode):
        return True
    return False
