"""SMIv2 plugin"""

from pyang import plugin
from string import Template
import time
import optparse
import os
import sys
from pyang.statements import *

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), 'lib')))
from globals import *

smi_module_name = 'yang2smi'

OctetLengthDefault = "1..255"

scalarSMIObjectTemplate = Template("""\
${scalarName}Table OBJECT-TYPE
    SYNTAX      SEQUENCE OF ${type}
    MAX-ACCESS  ${maxaccess}
    STATUS      current
    DESCRIPTION
            "${description}"
    ::= { ${oid} }
""")

groupSMIObjectTemplate = Template("""\
${groupName} OBJECT-GROUP
    OBJECTS  { ${children} }
    STATUS   current
    DESCRIPTION
             "${description}"
    ::= { ${parentArg} ${inode} }
""")

ORGANIZATION = "OpenClovis Solutions, Inc"
CONTACT_INFO = "Andrew Stone <stone@openclovis.com>"
DESC = "Automatically generated description"

mibImportsTemplate = {"SNMPv2-SMI":["MODULE-IDENTITY", "OBJECT-TYPE", "OBJECT-IDENTITY", "enterprises", "OBJECT-TYPE", "NOTIFICATION-TYPE"],
                      "SNMPv2-CONF": ["MODULE-COMPLIANCE", "OBJECT-GROUP", "NOTIFICATION-GROUP"],
                      "SNMPv2-TC": ["TEXTUAL-CONVENTION"]
                      }

convertType = [

     #==========================================================================
     # Translation of the SMIng built - in types to the YANG
     # equivalents. 
     #==========================================================================

    "SNMPv2-SMI", "Integer32", None, "int32",
    "", "Integer64", None, "int64",
    "SNMPv2-SMI", "Unsigned32", None, "uint32",
    "SNMPv2-SMI", "Unsigned32", None, "uint16",
    # Still does not support on Unsigned64 on MIB
    "SNMPv2-SMI", "Unsigned32", None, "uint64",
    "SNMPv2-SMI", "OctetString", None, "binary",
    "SNMPv2-SMI", "Bits", None, "bits",
    "", "OBJECT IDENTIFIER", "ietf-yang-types", "object-identifier",
    "SNMPv2-SMI", "ObjectIdentifier", "ietf-yang-types", "object-identifier-128",
    "SNMPv2-SMI", "zeroDotZero", None, "zeroDotZero",

     #==========================================================================
     # We want to do these translations as well in order to retire the
     # SNMPv2 - SMI module which is not really an SMIv2 module but part
     # of the definition of SNMPv2 - SMI itself.
     #==========================================================================

#     "SNMPv2-SMI", "Integer32", None, "int32",
#     "SNMPv2-SMI", "Integer64", None, "int64",
#     "SNMPv2-SMI", "Unsigned32", None, "uint32",
    "SNMPv2-SMI", "Opaque", "ietf-yang-smiv2", "opaque",
    "SNMPv2-SMI", "Counter32", "ietf-yang-types", "counter32",
    "SNMPv2-SMI", "Counter64", "ietf-yang-types", "counter64",
    "SNMPv2-SMI", "Gauge32", "ietf-yang-types", "gauge32",
    "SNMPv2-SMI", "TimeTicks", "ietf-yang-types", "timeticks",
    "SNMPv2-SMI", "IpAddress", "ietf-inet-types", "ipv4-address",


     #==========================================================================
     # And we like to do the same for RFC1155 - SMI definitions...
     #==========================================================================


    "RFC1155-SMI", "Opaque", "ietf-yang-smiv2", "opaque",
    "RFC1155-SMI", "Counter", "ietf-yang-types", "counter32",
    "RFC1155-SMI", "Gauge", "ietf-yang-types", "gauge32",
    "RFC1155-SMI", "TimeTicks", "ietf-yang-types", "timeticks",
    "RFC1155-SMI", "IpAddress", "ietf-inet-types", "ipv4-address",
    "RFC1155-SMI", "NetworkAddress", "ietf-inet-types", "ipv4-address",

     #==========================================================================
     # We also translate frequently used SNMPv2 - TCs that have a YANG
     # equivalent. Note that DateAndTime is slightly different from
     # the ISO profile used by date - and -time. 
     #==========================================================================

    "SNMPv2-TC", "PhysAddress", "ietf-yang-types", "phys-address",
    "SNMPv2-TC", "MacAddress", "ietf-yang-types", "mac-address",
    "SNMPv2-TC", "TruthValue", None, "boolean",
    "SNMPv2-TC", "TimeStamp", "ietf-yang-types", "timestamp",

    "RMON2-MIB", "ZeroBasedCounter32", "ietf-yang-types", "zero-based-counter32",

    "HCNUM-TC", "ZeroBasedCounter64", "ietf-yang-types", "zero-based-counter64",
    "HCNUM-TC", "CounterBasedGauge64", "ietf-yang-types", "gauge64",

    "INET-ADDRESS-MIB", "InetVersion", "ietf-inet-types", "ip-version",
    "INET-ADDRESS-MIB", "InetPortNumber", "ietf-inet-types", "port-number",
    "INET-ADDRESS-MIB", "InetAutonomousSystemNumber", "ietf-inet-types", "as-number",
    "INET-ADDRESS-MIB", "InetAddress", "ietf-inet-types", "ip-address",

    "DIFFSERV-DSCP-TC", "Dscp", "ietf-inet-types", "dscp",

    "IPV6-FLOW-LABEL-MIB", "IPv6FlowLabel", "ietf-inet-types", "ipv6-flow-label",

    "URI-TC-MIB", "Uri", "ietf-inet-types", "uri",


    None, None, None, None
]

# This defined wellknown some OBJECT-IDENTITY in SNMPV2 
SNMPv2ObjectIdentifier = ['zeroDotZero']

# Keyword identifier should be translated when converting YANG to SMIV2 if conflict
TOKEN = 'y2mib'
SNMPv2KEYWORD_MAPPING = { 'current':'current%s'%TOKEN}

class YANG2SMIPlugin(plugin.PyangPlugin):

    def pre_load_modules(self,ctx):
      pass

    def add_opts(self, optparser):
        optlist = [
            optparse.make_option("--yang2smi-help",
                                 dest = "yang2smi_help",
                                 action = "store_true",
                                 help = "Print help on yang2smi symbols and exit"),
            optparse.make_option("--yang2smi-oid",
                                 action = "store",
                                 dest = "yang2smi_oid",
                                 help = "Enterprise OID specific for module"),
            optparse.make_option("--yang2smi-indent",
                                 action = "store",
                                 type = "int",
                                 dest = "yang2smi_indent",
                                 help = "Indent param"),
            ]

        if hasattr(optparser, 'yang2smi_opts'):
            g = optparser.yang2smi_opts
        else:
            g = optparser.add_option_group("YANG2SMI specific options")
            optparser.yang2smi_opts = g

        g.add_options(optlist)

    def add_output_format(self, fmts):
        self.multiple_modules = True
        fmts['yang2smi'] = self

    def emit(self, ctx, modules, fd):
        Yang2SmiDoc = YANG2SMIEmit(ctx, fd)
        Yang2SmiDoc.emit(modules)

    def setup_ctx(self, ctx):
        if ctx.opts.yang2smi_help:
            print_help()
            sys.exit(0)
        if ctx.opts.format == 'yang2smi':
            if ctx.opts.yang2smi_oid is None:
                sys.stderr.write("Please input enterprise OID specific for module with option --yang2smi-oid=<num|module>\n")
                sys.exit(0)

    def setup_fmt(self, ctx):
        ctx.implicit_errors = False


def print_help():
    print """
    Convert from .yang to SMIv2 (MIB)
    """

def pyang_plugin_init():
    """Called by pyang plugin framework at to initialize the plugin."""

    # Register the plugin
    plugin.register_plugin(YANG2SMIPlugin())

class YANG2SMIEmit:
    def __init__(self, ctx, fd):
        self.ctx = ctx
        self.leafrefs = []
        self.indent = 4
        self.thismod_prefix = None
        self.imports = ({})
        self.queueimports = mibImportsTemplate
        self.data = {"header": [], "import":[], "body":[]}
        self.io = fd
        self.identities = []
        self.augmentpaths = []
        self.augments = []
        self.baseid = []

        if ctx.opts.yang2smi_indent is not None:
            self.indent = ctx.opts.yang2smi_indent

        pass

    def emit(self, modules):

        for module in modules:
            # For modules in queue translate 
            pre = module.search_one('prefix').arg
            pkg = module.arg
            self.imports[pre] = pkg

            # For import of each module
            imports = module.search('import')
            for i in imports:
                pre = i.search_one('prefix').arg
                pkg = i.arg
                self.imports[pre] = pkg

        for i in imports:
            pre = i.search_one('prefix').arg
            pkg = i.arg
            self.imports[pre] = pkg

        for module in modules:
            self.emit_module_header(0, module)
            for substmt in module.substmts:
                self.emit_stmt(1, module, substmt)

        self.emit_smi_footer(0)

        #Print to io output
        self.dumpout()

    #===========================================================================
    # Output to io (std or file)
    #===========================================================================
    def dumpout(self):
        headerMsg = """--
-- This SMIv2 module has been generated by pyang and yang2smi plugin. Do not edit.
--

"""
        self.io.write(headerMsg)

        #=======================================================================
        # Imports header output
        #=======================================================================
        for h in self.data["header"]:
            self.io.write(h)

        self.io.write("\n")
        strImport = ""
        for i in self.queueimports:
            if i != self.thismod_prefix:
                for type in set(self.queueimports[i]):
                    strImport += " "*self.indent * 2 + "%s,\n" % type
                strImport = strImport[0:len(strImport) - 2]
                strImport += " "*self.indent + "FROM %s\n" % i
        strImport = strImport[0:len(strImport) - 1] + ";"
        self.io.write(strImport)

        #=======================================================================
        # Module, Scalar object, notification output
        #=======================================================================
        self.io.write("\n")
        for b in self.data["body"]:
            self.io.write(b)

    def emit_arg(self, level, output, arg):
        """Heuristically pretty print the argument string"""
        arg = arg.replace('\\', r'\\')
        arg = arg.replace('\t', r'\t')
        lines = arg.splitlines(1)

        try:
          if len(lines) <= 1:
              if len(lines[0].strip()) == 0:
                  self.data[output].append("\n")
              else:
                  self.data[output].append(" "*(self.indent * level) + arg)
          else:
              if len(lines[0].strip()) != 0:
                  self.data[output].append(" "*(self.indent * level) + lines[0])
              else:
                  self.data[output].append("\n")
              for line in lines[1:-1]:
                  self.data[output].append(" "*(self.indent * level) + line)
              if len(lines[-1].strip()) != 0:
                  self.data[output].append(" "*(self.indent * level) + lines[-1])
              else:
                  self.data[output].append("\n")
        except:
          import pdb
          pdb.set_trace()


    def emit_typedef(self, level, parent, t):
        e = t.search_one('type')
        self.emit_arg(level, 'body', "\n%s ::= TEXTUAL-CONVENTION" % (t.arg[0].capitalize() + t.arg[1:]))

        #display-hint
        displayhint = t.search_one(('yang-smi', 'display-hint'))
        if displayhint is not None:
            self.emit_arg(level + 1, 'body', "\n")
            self.emit_arg(level + 1, 'body', "DISPLAY-HINT \"%s\"" % displayhint.arg)

        #status
        status = t.search_one('status')
        statusValue = 'current'
        if status is not None:
            statusValue = status.arg
        self.emit_arg(level + 1, 'body', "\n")
        self.emit_arg(level + 1, 'body', "STATUS %s" % statusValue)

        #description
        description = t.search_one('description')
        if description is not None:
            self.emit_arg(level + 1, 'body', "\n")
            self.emit_arg(level + 1, 'body', "DESCRIPTION")
            self.emit_arg(level + 2, 'body', '\n"' + description.arg + '"\n')
        else:
            self.emit_arg(level + 2, 'body', '\n"%s"\n'%DESC)

        #reference
        reference = t.search_one('reference')
        if reference is not None:
            self.emit_arg(level + 1, 'body', "\n")
            self.emit_arg(level + 1, 'body', "REFERENCE")
            self.emit_arg(level + 1, 'body', "\n")
            self.emit_arg(level + 2, 'body', '"' + reference.arg + '"')

        self.emit_arg(level + 1, 'body', "\n")
        self.emit_arg(level + 1, 'body', "SYNTAX")

        if e.arg != 'enumeration':
            i = 0
            typeDefine = self.typegen(t, e)
            self.emit_arg(level, 'body', typeDefine)
        else:
            self.emit_arg(1, 'body', " INTEGER {")
            i = 0
            self.emit_arg(level + 2, 'body', "\n")
            while i < len(e.substmts) - 1:
                self.emit_arg(level + 4, 'body', "%s" % e.substmts[i].arg)
                value = e.substmts[i].search_one('value')
                if value is not None:
                    self.emit_arg(0, 'body', " (%s)," % value.arg)
                else:
                    self.emit_arg(0, 'body', ",")
                i += 1
                self.emit_arg(level + 2, 'body', "\n")

            self.emit_arg(level + 4, 'body', "%s" % e.substmts[i].arg)
            value = e.substmts[i].search_one('value')
            if value is not None:
                self.emit_arg(1, 'body', " (%s)" % value.arg)

            self.emit_arg(level + 2, 'body', "\n")
            self.emit_arg(level + 1, 'body', "}")

        self.emit_arg(level + 1, 'body', "\n")

    def emit_container(self, level, parent, node):
        self.emit_arg(level, 'body', "\n")
        self.emit_arg(level, 'body', "\n")
        if parent.parent is not None:
            self.emit_arg(level, 'body', "%s OBJECT IDENTIFIER ::= { %s %d }" % (self.translateArg(node), self.getParentArg(parent), self.findIndexNode(parent, node)))
        else:
            self.emit_arg(level, 'body', "%s OBJECT IDENTIFIER ::= { %s %d }" % (self.translateArg(node), self.thismod_prefix, self.findIndexNode(parent, node)))

    def emit_grouping(self, level, parent, stmt, suffix = ''):
        description = stmt.search_one('description')
        desc = "%s"%DESC
        if description is not None:
          desc = description.arg

        children = [self.translateArg(child) for child in stmt.i_children if child.keyword in (yang_stmts + leaf_stmts)]
        out = groupSMIObjectTemplate.safe_substitute(groupName=self.translateArg(stmt), children=", ".join(children), description = desc, parentArg = self.getParentArg(parent), inode = self.findIndexNode(parent, stmt))

        self.emit_arg(level, 'body', "\n")
        self.emit_arg(level, 'body', "\n")
        for arg in out.splitlines():
          self.emit_arg(level, 'body', "%s\n" %arg)

    def emit_list(self, level, parent, node, suffix = ''):
        nameTable = "%sTable" % node.arg
        nameTableEntry = node.arg[0].capitalize() + node.arg[1:]

        self.emit_arg(level, 'body', "\n")
        self.emit_arg(level, 'body', "\n")
        self.emit_arg(level, 'body', "%s OBJECT-TYPE" % nameTable)
        self.emit_arg(level, 'body', "\n")
        self.emit_arg(level + 1, 'body', "SYNTAX")
        self.emit_arg(1, 'body', "SEQUENCE OF %s" % nameTableEntry)
        self.emit_arg(level, 'body', "\n")
        self.emit_arg(level + 1, 'body', "MAX-ACCESS not-accessible")
        self.emit_arg(level, 'body', "\n")
        self.emit_arg(level + 1, 'body', "STATUS current")
        self.emit_arg(level, 'body', "\n")

        self.emit_arg(level, 'body', "\n")
        self.emit_arg(level + 1, 'body', "DESCRIPTION")

        #get description from list keyword
        description = parent.search_one('description')
        if description is not None:
            self.emit_arg(level + 2, 'body', '\n"' + description.arg + '"\n')
        else:
            self.emit_arg(level + 2, 'body', '\n"%s"\n'%DESC)

        self.emit_arg(level, 'body', "\n")
        self.emit_arg(level + 1, 'body', "::= { %s %d }" % (self.getParentArg(parent), self.findIndexNode(parent, node)))

        self.emit_arg(level, 'body', "\n\n%s OBJECT-TYPE" % node.arg)
        self.emit_arg(level + 1, 'body', "\nSYNTAX %s" % nameTableEntry)
        self.emit_arg(level + 1, 'body', "\nMAX-ACCESS not-accessible\n")
        self.emit_arg(level + 1, 'body', "STATUS current\n")

        #get description from list keyword
        self.emit_arg(level, 'body', "\n")
        self.emit_arg(level + 1, 'body', "DESCRIPTION")
        description = node.search_one('description')
        if description is not None:
            self.emit_arg(level + 2, 'body', '\n"' + description.arg + '"\n\n')
        else:
            self.emit_arg(level + 2, 'body', '\n"%s"\n'%DESC)

        key = node.search_one('key')
        if key is not None:
            self.emit_arg(level + 1, 'body', '\nINDEX { ' + ', '.join(key.arg.split(" ")) + ' }\n')

        #Scalar object, identifier should equal 1
        self.emit_arg(level, 'body', '\n::= { %s 1 }' % nameTable)

        #Now for data entry type
        self.emit_arg(level + 1, 'body', "\n")
        self.emit_arg(level + 1, 'body', "\n")
        self.emit_arg(level + 1, 'body', "%s ::= SEQUENCE  {" % nameTableEntry)
        str = "\n"
        for substmt in node.substmts:
            if substmt.keyword == 'leaf':
                type = substmt.search_one("type")
                #2 things: if ref our this module, ignore (and it is in key)
                if type.arg == 'leafref':
                    typeRef = self.typegen(node, type)
                    if typeRef != substmt.arg:
                        str += substmt.arg
                        str += " "*self.indent + self.typegen(node, type) + ",\n"

                else:
                    str += self.translateArg(substmt)
                    str += " "*self.indent + self.typegen(node, type) + ",\n"

        str = str[0: len(str) - 2]
        self.emit_arg(level + 2, 'body', str)
        self.emit_arg(level + 1, 'body', "\n")
        self.emit_arg(level + 1, 'body', "}")

    def emit_leaf(self, level, parent, node):
        key = parent.search_one('key')
        index = []
        if key is not None:
            index = key.arg.split(" ")

        type = node.search_one("type")
        if type.arg == 'leafref' and node.arg in index:
            #
            return

        self.emit_arg(level, 'body', "\n")
        self.emit_arg(level, 'body', "\n")
        self.emit_arg(level, 'body', "%s OBJECT-TYPE" % self.translateArg(node))

        self.emit_arg(level, 'body', "\n")
        type = node.search_one("type")
        self.emit_arg(level + 1, 'body', "SYNTAX " + self.typegen(node, type, brief = False))

        #UNITS
        unit = node.search_one('units')
        if unit is not None:
            str = "\nUNITS \"%s\"\n" % unit.arg
            self.emit_arg(level + 1, 'body', str)

        #max-access
        config = node.search_one('config')
        self.emit_arg(level, 'body', "\n")
        self.emit_arg(level + 1, 'body', "MAX-ACCESS ")

        '''
        MAX-ACCESS values: "not-accessible", "accessible-for-notify", "read-only", "read-write", "read-create"
        Only columnar object index has 'not-accessible' as its maximal level of access, other columnar has 'read-write'
        '''
        isconfig = True
        if config is not None:
            if config.arg == "true":
                isconfig = True
            else:
                isconfig = False

        access = 'read-write'
        if isconfig == False:
          access = 'read-only'
        elif key and key.arg == node.arg:
          access = 'not-accessible'

        self.emit_arg(0, 'body', access)

        #status
        status = node.search_one('status')
        statusValue = 'current'
        if status is not None:
            statusValue = status.arg
        self.emit_arg(level + 1, 'body', "\n")
        self.emit_arg(level + 1, 'body', "STATUS %s" % statusValue)

        #description
        self.emit_arg(level + 1, 'body', "\n")
        self.emit_arg(level + 1, 'body', "DESCRIPTION")
        description = node.search_one('description')
        if description is not None:
            self.emit_arg(level + 2, 'body', '\n"' + description.arg + '"\n')
        else:
            self.emit_arg(level + 2, 'body', '\n"%s"\n'%DESC)

        #reference
        reference = node.search_one('reference')
        if reference is not None:
            self.emit_arg(level + 1, 'body', "\n")
            self.emit_arg(level + 1, 'body', "REFERENCE")
            self.emit_arg(level + 1, 'body', "\n")
            self.emit_arg(level + 2, 'body', '"' + reference.arg + '"')

        defval = node.search_one(('yang-smi', 'default'))
        if defval is None:
            defval = node.search_one('default')
        if defval is not None:
            if defval.arg in SNMPv2ObjectIdentifier:
                self.queueimports['SNMPv2-SMI'].append(defval.arg)

            default = defval.arg
            if type.arg in ("string",):
                default = "\"%s\""%defval.arg
            str = "\nDEFVAL { %s }\n" % default
            self.emit_arg(level + 1, 'body', str)

        self.emit_arg(level + 1, 'body', "\n::= { %s %d }" % (self.getParentArg(parent), self.findIndexNode(parent, node)))

    def emit_identity(self, level, mod, stmt):
        self.identities.append(stmt.arg)
        base = stmt.search_one('base')
        if base is not None:
            self.baseid.append(base.arg)

    def emit_stmt(self, level, parent, stmt):
        if search_one(stmt, 'uses'):
          self.emit_grouping(level, parent, stmt)
          if hasattr(stmt, 'i_children') and len(stmt.i_children) > 0:
            for s in stmt.i_children:
              self.emit_stmt(level + 1, stmt, s)

        elif stmt.keyword == 'container':
            self.emit_container(level, parent, stmt)
            for s in stmt.substmts:
                self.emit_stmt(level + 1, stmt, s)

        elif stmt.keyword == 'augment':
             # HERE
             a = stmt.arg

             # also, since we are the root, add the module as parent
             if not (self.full_path(stmt)) in self.augmentpaths:
                 self.augmentpaths.append(self.full_path(stmt))

             # MEF
             if stmt.arg.find(':') == -1:
                 prefix = self.thismod_prefix
             else:
                 prefix = stmt.arg[1:stmt.arg.find(':')]

             node = find_target_node(self.ctx, stmt, True)
             if node is not None and prefix in self.imports:
                 self.augments.append(self.full_path(stmt) + '-->' + self.full_path(node))
             else:
                 pass

             for s in stmt.substmts:
                 self.emit_stmt(level + 1, stmt, s)

        elif stmt.keyword == 'list':
            self.emit_list(level, parent, stmt)
            for s in stmt.substmts:
                self.emit_stmt(level, stmt, s)

        elif stmt.keyword == 'choice':
            print "CHOICE:" + self.full_path(stmt)
            for child in stmt.substmts:
                self.emit_stmt(level + 1, stmt, child)

        elif stmt.keyword == 'case':
            print "case:" + self.full_path(stmt)
            for child in parent.substmts:
              self.emit_stmt(level + 1, stmt, child)

        elif stmt.keyword == 'leaf':
          self.emit_leaf(level, parent, stmt);

        elif stmt.keyword == 'identity':
            self.emit_identity(level, parent, stmt)

        if stmt.keyword == 'typedef':
            self.emit_typedef(level, parent, stmt)
        elif stmt.keyword == 'rpc':
            pass
            #print "Unknown: rpc"
            #self.emit_action(level, mod, stmt)
        elif stmt.keyword == 'notification':
            self.emit_notif(level, parent, stmt)
        elif stmt.keyword == 'feature':
            pass
            #print "Unknown: feature"
            #self.emit_feature(level, mod, stmt)
        elif stmt.keyword == 'deviation':
            pass
            #print "Unknown: deviation"
            #self.emit_feature(level, mod, stmt)

    def emit_module_header(self, level, module):

        imports = module.search('import')
        for i in imports:
            pre = i.search_one('prefix').arg
            pkg = i.arg
            self.imports[pre] = pkg

        self.emit_arg(level, 'header', "%s-MIB DEFINITIONS ::= BEGIN" % module.arg.upper())
        self.emit_arg(level + 1, 'header', "\n\nIMPORTS")

        module_identity = None

        pre = module.search_one('prefix')
        if  pre is not None:
            module_identity = pre.arg

        # Pre-check module identity and append suffix 'Module' if conflict with another yang top level
        for substmt in module.substmts:
          if (substmt.keyword == 'container') and (substmt.parent.parent is None) and (substmt.arg is not None):
            if module_identity is None:
              module_identity = substmt.arg
            elif module_identity == substmt.arg:
              module_identity ="%sModule"%module_identity
            break

        if module_identity is not None:
            self.emit_arg(level + 1, 'body', "\n")
            self.emit_arg(level + 1, 'body', "%s MODULE-IDENTITY" % module_identity)
            self.thismod_prefix = module_identity
        else:
            sys.stderr.write("Invalid YANG module file!")
            sys.exit(0)

        #Last updated
        revs = module.search("revision")
        self.emit_arg(level + 2, 'body', "\n")
        lastUpdate = "%sZ"%time.strftime("%Y%m%d%H%M")
        if len(revs) > 0:
            lastUpdate = "%s"%self.current_revision(0, revs)
        self.emit_arg(level + 2, 'body', 'LAST-UPDATED "%s"' % lastUpdate)

        #Organization
        org = module.search_one('organization')
        self.emit_arg(level + 2, 'body', "\n")
        self.emit_arg(level + 2, 'body', 'ORGANIZATION')
        if org is not None:
            self.emit_arg(level + 3, 'body', '\n"' + org.arg + '"\n')
        else: #Default
            self.emit_arg(level + 3, 'body', '\n"' + ORGANIZATION + '"\n')

        #contact info
        contact = module.search_one('contact')
        self.emit_arg(level + 2, 'body', "\n")
        self.emit_arg(level + 2, 'body', 'CONTACT-INFO')
        if contact is not None:
            self.emit_arg(level + 3, 'body', '\n"' + contact.arg + '"\n')
        else: #Default
            self.emit_arg(level + 3, 'body', '\n"' + CONTACT_INFO + '"\n')

        #Description
        self.emit_arg(level + 2, 'body', "\n")
        self.emit_arg(level + 2, 'body', 'DESCRIPTION')
        description = module.search_one('description')
        if description is not None:
            self.emit_arg(level + 3, 'body', '\n"' + description.arg + '"\n')
        else:
            self.emit_arg(level + 3, 'body', '\n"%s"\n'%DESC)

        #Revision --Revision History
        self.emit_arg(level + 2, 'body', "\n")
        self.emit_arg(level + 2, 'body', "\n")
        self.emit_arg(level + 2, 'body', '--Revision History')
        if len(revs) > 0:
            for r in revs:
                cur = [int(p) for p in r.arg.split("-")]
                rev = "%4d%02d%02d" % tuple(cur) + "" + time.strftime("%H%M") + "Z"
                self.emit_arg(level + 2, 'body', "\n")
                self.emit_arg(level + 2, 'body', 'REVISION "%s"' % rev)
                if r.search_one('description'):
                    self.emit_arg(0, 'body', "\n")
                    self.emit_arg(level + 2, 'body', "\n")
                    self.emit_arg(level + 2, 'body', 'DESCRIPTION')
                    self.emit_arg(level + 3, 'body', '\n"' + r.search_one('description').arg + '"\n')

        enterpriseOID = self.ctx.opts.yang2smi_oid
        self.emit_arg(level + 1, 'body', "\n")
        self.emit_arg(level + 1, 'body', '::= { enterprises %s }' % enterpriseOID)


    def emit_smi_footer(self, level):
        self.emit_arg(0, 'body', "\n")
        self.emit_arg(level, 'body', "END")
        self.emit_arg(0, 'body', "\n")

    def current_revision(self, level, r_stmts):
        """Pick the most recent revision date.
    
        `r_stmts` is a list of 'revision' statements.
        """
        cur = max([[int(p) for p in r.arg.split("-")] for r in r_stmts])

        return "%4d%02d%02d" % tuple(cur) + "" + time.strftime("%H%M") + "Z"

    def typebasic(self, t):
        i = 0
        if t.arg.find(":") == -1:
            while i < len(convertType):
                if (convertType[i + 2] == None and convertType[i + 3] == t.arg):
                    if convertType[i] != "":
                        if convertType[i] not in self.queueimports:
                            self.queueimports[convertType[i]] = []
                        self.queueimports[convertType[i]].append(convertType[i + 1])
                    return convertType[i + 1]
                i += 4
        else:
            [prefix, name] = t.arg.split(':', 1)
            while i < len(convertType):
                if (convertType[i + 2] == prefix or prefix == 'yang' or convertType[i + 2] == self.imports[prefix]) and (convertType[i + 3] == name):
                    if convertType[i] != "":
                        if convertType[i] not in self.queueimports:
                            self.queueimports[convertType[i]] = []
                        self.queueimports[convertType[i]].append(convertType[i + 1])
                    return convertType[i + 1]
                i += 4
        return None

    def typegen(self, node, t, brief = True):
        if t.arg == 'string' or t.arg == 'binary':
            if brief == True:
                return "OCTET STRING"
            else:
                length = t.search_one("length")
                if length is None:
                    return "OCTET STRING (SIZE(" + OctetLengthDefault + "))"
                else:
                    return "OCTET STRING (SIZE(" + length.arg + "))"
        elif t.arg == 'enumeration':
            syntax = "INTEGER"
            if brief == True:
                return syntax
            else:
                syntax = "%s {\n" % syntax
                i = 0
                while i < len(t.substmts) - 1:
                    syntax += " "*self.indent + "%s" % t.substmts[i].arg
                    value = t.substmts[i].search_one('value')
                    if value is not None:
                        syntax += "(%s)," % value.arg
                    else:
                        syntax += ","
                    i += 1
                    syntax += "\n"
                syntax += " "*self.indent + "%s" % t.substmts[i].arg
                value = t.substmts[i].search_one('value')
                if value is not None:
                    syntax += "(%s)" % value.arg
                syntax += "\n"
                syntax += "}"
            return syntax
        elif t.arg == 'leafref':
            s = t.arg + ' : '
            p = t.search_one('path')
            if p is not None:
                leafrefkey = p.arg
                leafrefkey = leafrefkey[leafrefkey.rfind("/") + 1:]

                if leafrefkey.find(':') == -1:
                    prefix = self.thismod_prefix
                else:
                    [prefix, name] = leafrefkey.split(':', 1)

                if prefix != self.thismod_prefix:
                    if self.imports[prefix] not in self.queueimports:
                        self.queueimports[self.imports[prefix]] = []
                    self.queueimports[self.imports[prefix]].append(name)
            return name

        elif t.arg == 'identityref':
            print "Unknown: identityref"
            return t.arg

        elif t.arg == 'union':
            print "Unknown: union"
            return t.arg
        else:
            typeBasic = self.typebasic(t)
            if typeBasic is not None:
                if brief == True:
                    return typeBasic

                range = t.search_one("range")
                if range is not None:
                    return typeBasic + " (" + range.arg + ")"
                else:
                    return typeBasic
            elif t.arg.find(":") == -1:
                return t.arg
            else:
                [prefix, name] = t.arg.split(':', 1)
                #If this using import module
                if prefix in self.imports:
                    #Insert into queue importing
                    if prefix != self.thismod_prefix:
                        if self.imports[prefix] not in self.queueimports:
                            self.queueimports[self.imports[prefix]] = []
                        self.queueimports[self.imports[prefix]].append(name)
                    return name
                else:
                    print "Unknown: %s" % t.arg
                    return "Unknown: %s" % t.arg

    def full_path(self, stmt):
        pathsep = "/"
        path = stmt.arg
        if (stmt.keyword != 'grouping') and (stmt.keyword != 'choice') and (stmt.keyword != 'case'):
            while stmt.parent is not None:
                stmt = stmt.parent
                if stmt.arg is not None:
                    path = stmt.arg + pathsep + path
        return path

    def findIndexNode(self, parent, node):
        i = 0
        for child in parent.substmts:
            #get child statements
            if (child.keyword == 'container') or (child.keyword == 'leaf') or (child.keyword == 'leaf-list') or (child.keyword == 'list') or (child.keyword == 'notification'):
                i += 1
                if child.arg == node.arg:
                    return i

        # Lookup child in i_children (grouping, case)
        if hasattr(parent, 'i_children') and len(parent.i_children) > 0:
          for child in parent.i_children:
              #get child statements
              if (child.keyword == 'container') or (child.keyword == 'leaf') or (child.keyword == 'leaf-list') or (child.keyword == 'list') or (child.keyword == 'notification'):
                  i += 1
                  if child.arg == node.arg:
                      return i

        return i

    def emit_child_notif(self, level, parent, node):
        if node.keyword == 'container':
            str = ''
            for params in node.substmts:
                objectName = self.emit_child_notif(level, node, params)
                if objectName != None:
                    str += objectName + ',\n'
            return str[0:len(str) - 2]

        if node.keyword == 'leaf':
            type = node.search_one('type')
            if type is None or type.arg != 'leafref':
                return node.arg
        return None

    def emit_notif(self, level, mod, node):

        self.emit_arg(level, 'body', "\n\n%s NOTIFICATION-TYPE" % (node.arg[0].lower() + node.arg[1:]))
        self.emit_arg(level + 1, 'body', "\nOBJECTS {")
        strNotif = '\n'
        for params in node.substmts:
            objectName = self.emit_child_notif(level, node, params)
            if objectName != None and objectName.strip() != '':
                strNotif += " "*self.indent + objectName + ',\n'
        self.emit_arg(level + 1, 'body', strNotif[0: len(strNotif) - 2])
        self.emit_arg(level + 1, 'body', "\n}")

        status = node.search_one('status')
        statusValue = 'current'
        if status is not None:
            statusValue = status.arg

        self.emit_arg(level + 1, 'body', "\nSTATUS %s" % statusValue)

        #description
        description = node.search_one('description')
        if description is not None:
            self.emit_arg(level + 1, 'body', "\nDESCRIPTION")
            self.emit_arg(level + 2, 'body', '\n"%s"' % description.arg)
        else:
            self.emit_arg(level + 2, 'body', '\n"%s"\n'%DESC)

        self.emit_arg(level + 1, 'body', "\n::= { %s %d }\n" % (self.thismod_prefix, self.findIndexNode(mod, node)))
    
    def getParentArg(self, node):
      return node.arg

    def translateArg(self, node):
      if SNMPv2KEYWORD_MAPPING.get(node.arg, None) is not None:
        return SNMPv2KEYWORD_MAPPING[node.arg]
      return node.arg
