###############################################################################
#
# Copyright (C) 2002-2009 by OpenClovis Inc. All  Rights Reserved.
# 
# The source code for  this program is not published  or otherwise 
# divested of  its trade secrets, irrespective  of  what  has been 
# deposited with the U.S. Copyright office.
# 
# This program is  free software; you can redistribute it and / or
# modify  it under  the  terms  of  the GNU General Public License
# version 2 as published by the Free Software Foundation.
# 
# This program is distributed in the  hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied  warranty  of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
# General Public License for more details.
# 
# You  should  have  received  a  copy of  the  GNU General Public
# License along  with  this program. If  not,  write  to  the 
# Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
#
###############################################################################
import pdb
import sys
import os
import xml.dom.minidom as minidom
import microdom
from types import *
import re, fnmatch
from mgtcommon import *
from xml.etree import ElementPath
from xml.etree import ElementTree as ET
from ncclient.operations import RPC
from ncclient.devices.default import DefaultDeviceHandler
from lxml import etree

import traceback
import logging

snippet = """
<config xmlns:xc="urn:ietf:params:xml:ns:netconf:base:1.0">
    %s
</config>
"""

class SetOperationFailed(Exception):
  def __init__(leaf, value):
    Exception.__init__("NETCONF set of [%s] to [%s] failed" % (leaf,str(value)))
    self.node = leaf
    self.value = value


class MgtObject(microdom.MicroDom):
    """
    The MgtObject class is used to store manageability objects in a hierarchical relationship.
    """

    def __init__(self, parent, attributes,children,data,keys=None):
        microdom.MicroDom.__init__(self, attributes,children,data)
        
        self.isleaf = True
        #if self.children_ != []:  You can have constant data.  So it could be I want to say if len(self.children_) > 1 here
        #    self.isleaf = False

        for child in self.children_:
            if child.__class__ == MgtObject:
                child.parent = self
                self.isleaf = False

        self.callbackmethod = None
        self.parent = parent
        self.keys = {} # Dictionary of key/value for the names of this item
        self.names = [] # Fill this field with all the keys that name this item
        self.settingTag = None
        
        # Value to identify network element (or network root object)
        self.isNetworkElement = False

        if keys:
          primaryKey = None
          for k in keys:
            chkey = self.child_.get(k,None)
            if chkey and chkey.data_:
              self.names.append(str(chkey.data_))
              self.keys[k] = str(chkey.data_)
              if not primaryKey: primaryKey = chkey.data_ 
          if primaryKey:
            self.attributes_["listkey"] = primaryKey  # set the listkey attribute which the safplus CLI uses to identify list elements. This assumes the first key is the "primary"

    def isListElem(self):
      """? Returns whether this object is an element of a list"""
      # We figure this out by seeing whether it has list keys
      return bool(self.names)

    def fullPath(self):
      myname = "/" + self.tag_
      if self.names:
        myname = myname + "/" + self.names[0] # just assume first name is the primary key
      if self.parent:
        return self.parent.fullPath() + myname
      else:
        return myname

    def diff(self,a,myName=None,hisName=None):
        if not myName: myName = "a"
        if not hisName: hisName = "b"
        intersect = False  # If intersect is true, only diff elements that are common to both trees.  If intersect is false, include elements missing from the other as part of the diff
        if len(self.child_)==0 and len(a.child_)==0:  # both are data nodes -- they have no children
          if self.children_ == a.children_:
            return None
          else:
            mec = self.copy()
            mec.attributes_["tag_"] = mec.tag_ = myName
            ac = a.copy()
            ac.attributes_["tag_"] = ac.tag_ = hisName
            return MgtObject(None,{"tag_":self.tag_, "conflict":"True"},[mec, ac],None)

        # OK these nodes are not leafs
        ac = dict(a.child_)
        diffc = {}
        for (name,child) in self.child_.items():
          tmp = ac.get(name,None)
          if tmp: # If the child has this node, then diff them
            recdiff = child.diff(tmp,myName,hisName)  
            if recdiff: diffc[name] = recdiff  # If diferences were found, remember them, otherwise drop this node
            del ac[name] # Now remove from our temporary dictionary so later we can determine if 'a' has children that I don't have
          elif not intersect:  # I'm the only one with this node, so add it if we are not doing an xdiff (intersection then diff)
            newobj = child.copy()
            newobj.attrib_["diff"] = myName  # Add an indicator describing which tree this came from
            diffc[name] = newobj

        for (name,child) in ac.items():  # Any children in a that I don't have?
            newobj = child.copy()
            newobj.attrib_["diff"] = hisName  # Add an indicator describing which tree this came from
            diffc[name] = newobj

        if diffc:          
          ret = MgtObject(None,self.attributes_,diffc.values(),self.data_)
          return ret

        return None


    def copy(self,recurse=100000):
      children = []
      for child in self.children_:
        try:
          children.append(child.copy())
        except AttributeError: # Its data
          children.append(child)
      newobj = MgtObject(self.parent,self.attributes_,children,None)
      newobj.keys.update(self.keys)
      newobj.names=list(self.names)
      return newobj

    def dump(self, recurse=True, fullPathOnce=False):
        if type(recurse) == BooleanType:  # Convert boolean to recurse depth if a number was not passed
          if recurse == True: 
            recurse = 1000  #  If the xml is deeper than 1000, its probably bad :-)
          if recurse == False:
            recurse = 0 
        if self.data_ or self.children_ and recurse > 0:
          full = 1
        else: full = 0

        # The data is also encoded as a child (so position is not lost)
        #if self.data_: datastr = str(self.data_)
        #else: datastr = ""
        if self.children_:
          chlst=[]
          if not recurse: # chstr = "<%d children>" % len(self.children_)
            unDumpedChild=0
            # pdb.set_trace()
            for c in self.children_:  # Even though we are not recursing, go though the children and dump and constant strings/numbers, the name (if it exists), and keep track of whether there is more children to dump
              try:
                tmp = c.tag_  # check to make sure it is a complex object
                if 0:  # Not needed: mgtobject sets up listkey attribute when created. c.tag_ == "name":  # WORKAROUND: if there is a child named "name" then dump it regardless of the depth parameter so netconf lists work properly
                  chlst.append(c.dump(True)) 
                else:
                  unDumpedChild+=1
              except AttributeError: # Its raw data
                chlst.append(str(c))
            if unDumpedChild: chlst.append("<more>%s</more>" % unDumpedChild)
          else:
            for c in self.children_:
              try:
                chlst.append(c.dump(recurse-1))
              except AttributeError:
                chlst.append(str(c))
          if chlst: full = 1  # If there's something in the child list, generate full xml
          chstr = "".join(chlst)
        else: chstr = ""

        attrStringLst = []
        if len(self.attributes_)>1:
          # Format attribute string, eliminating the tag_ attribute
          attrStringLst = ["%s='%s'" % (k,v) for (k,v) in filter(lambda i: i[0] != 'tag_', self.attributes_.items())]
        else:
          attrStringLst = []
        if fullPathOnce:
          attrStringLst.append("path='%s'" % self.fullPath())
        attrs = " " + " ".join(attrStringLst)

        if full:
          return "<%s%s>%s</%s>" % (self.tag_,attrs,chstr,self.tag_)
        else:
          return "<%s%s />" % (self.tag_,attrs)

    def matches(self,spec):
      if self.tag_ == spec[0]:
        if not self.names or len(spec)==1:  # not a list element or all list elements selected
          return spec[1:]
        else:
          if spec[1] in self.names or fnmatch.filter(self.names,spec[1]):  # fnmatch does advanced *? matching
            return spec[2:]
          if spec[1] == "**":
            return spec[1:] # ** matches many children so spec node is not consumed

      if fnmatch.filter(self.names,spec[0]):
        return spec[1:]
      if spec[0] == "**":  
        return spec # ** matches many children so spec node is not consumed
      return None
       

    def resolvePathToXml(self, path,depth=16):
      # pdb.set_trace()
      elementList = self.resolvePath(path)
      if not elementList:
        return "<error>Unknown path [%s]</error>" % "/".join(path)

      ret = []
      for e in elementList:
        ret.append(e.dump(depth,fullPathOnce=True))
      return "\n".join(ret)

    def resolvePath(self,path): 
      if not path:  # Nothing left so return this object
        return [self]
      else:
        mychildname = path[0]
        if mychildname == "" or mychildname == ".":
          return self.resolvePath(path[1:]) # foo//bar == foo/bar and also handle foo/, foo/./bar == foo/bar
        if mychildname == "..":
          next = self.parent if self.parent else self  # .. goes up but not beyond the top
          return next.resolvePath(path[1:])

        if self.child_.has_key(mychildname): # Try a direct lookup
          ch = self.child_[mychildname]
          if not ch.names:  # It is not a multi-named list element (if it is we need to validate that the names match)
            return ch.resolvePath(path[1:])
          
        # Lookup did not work, iterate thru all children looking for a complex match
        if 1:
          ret = []
          for c in self.children_:
            if isinstance(c,MgtObject):  # Make sure its a node, not just data
              rest = c.matches(path)
              if rest is not None:  # If [] is returned, that should eval to True.  It means that that match fully consumed the path
                ret += c.resolvePath(rest)
          return ret
          #return "<error>Unknown child [%s]</error>" % mychildname
          # pdb.set_trace()

    def convertXPath2Ele(self,xpath):
        # Add leading '/' if missing
        if xpath[0] != "/":
          xpath = "/" + xpath
    
        next = iter(ElementPath.xpath_tokenizer(xpath)).next
    
        # Tree root node
        etRoot = None

        # Map to store xpath and ElementXpath
        mapparentelem = {}
    
        #For each leaf node on xpath
        current_elemment = etRoot
    
        def key_to_xml(s):
          s = s.replace("_",":")
          #tag without array index
          s = s.split("[")[0]
          s = s.split("=")[0]
          return s
        
        """ Build element attribute """
        def _transformAttr(next, token):
            predicate = []
            attributes = {}
        
            #This is [list] attribute value of a node element
            #Ex: /a/b/c[name="foo" id="bar"] => <name>foo</name><id>bar</id>
            if token[0] == "[":
                while 1:
                    token = next()
                    if token[0] == "]":
                        break
                    if token[0] and token[0][:1] in "'\"":
                        token = "'", token[0][1:-1]
                    predicate.append(token[1])
        
                #This is array index [1] ... [2] ...(already combine with @attributes)
                if not re.match("\d+$", predicate[0]):
                    while len(predicate)>2:
                        #Ignore ',' expression token
                        if predicate[0] == ',':
                            predicate = predicate[2:]
                        key = predicate[0]
                        value = predicate[2]
                        # shift left
                        predicate = predicate[4:]
                        attributes[key_to_xml(key)] = value
        
            return attributes
    
        def _buildNode(current_xpath):
            current_elemment = None
            p_parent_xpath = None
            try:
                current_elemment = mapparentelem[current_xpath]
            except:
                try:
                    p_parent_xpath = current_xpath[:current_xpath.rfind("/")]
                    token = current_xpath[current_xpath.rfind("/") + 1:]
                    #Get current element with xpath:current_xpath if exists
                    parent_elemment = mapparentelem[p_parent_xpath]
                    current_elemment = ET.Element(key_to_xml(token))
                    mapparentelem[current_xpath] = current_elemment
                    parent_elemment.append(current_elemment)
                except Exception,e:
                    """ Backward up level of parent xpath """
                    current_elemment = _buildNode(p_parent_xpath)
    
            return current_elemment
    
        while 1:
            try:
                token = next()
                if token[0] == "/":
                    token = next()
    
                    index = xpath.rfind("/%s" %token[1])
                    if index >= 0:
                        parent_xpath = "%s" %(xpath[:index])
    
                        #Get xpath assume suffix is leaf
                        next_index = xpath.find("/", len(parent_xpath) + 1)
                        if next_index>0:
                            current_xpath = xpath[0:next_index]
                        else:
                            #Last leaf
                            current_xpath = xpath
                        if etRoot is None:
                            etRoot = ET.Element(key_to_xml(token[1]))
                            current_elemment = etRoot
                            mapparentelem[current_xpath] = current_elemment
                        else:
                            current_elemment = _buildNode(current_xpath)
                elif token[0] == "[":
                    attributes = _transformAttr(next, token)
                    for (k,v) in attributes.items():
                        child = ET.Element(k)
                        child.text = v
                        current_elemment.append(child)
    
                        # Add operation into node tag
                        current_elemment.attrib = {}
                elif token[0] == "=":
                  current_elemment.text = next()[1]
    
            except StopIteration:
                break
    
        return etRoot

    #Public methods 
    def getName(self):
        """Get name of the MGT object."""
        return self.tag_

    def getXPath(self, withkey, withroot = True, withname = True):
        """Get location path of the MGT object."""
        if self.parent == None:
            return self.tag_

        if self.parent.tag_ == '/':
            return self.tag_

        keypath = ""
        if withkey and self.keys != {}:
            if (withname == False) and len(self.keys) == 1:
                keypath = '/'
                for k,v in self.keys.iteritems():
                    keypath = keypath + v
            else:
                keypath = '[' + ",".join(map(lambda (k,v): "%s='%s'" % (k,v), self.keys.iteritems())) + ']'

        if not withroot:
            if self.parent.parent == None:
                return self.tag_
            if self.parent.parent.tag_ == '/':
                return self.tag_ + keypath

        return self.parent.getXPath(withkey, withroot, withname) + '/' + self.tag_ + keypath

    def getXmlTag(self):
        """Get location path of the MGT object."""

        # Not including network element into xml tag
        if self.isNetworkElement or self.tag_ == 'top':
            return ['','']

        openTag = "<%s>" % self.tag_
        closerTag = "</%s>" % self.tag_

        # Build full openTag/closerTag
        p = self.parent
        while p != None and p.parent != None and not p.isNetworkElement and p.tag_ != "/":
            keypath = "\n".join(map(lambda (k,v): "<%s>%s</%s>" % (k,v,k), p.keys.iteritems()))

            openTag = "<%s>%s" %(p.tag_,keypath) + openTag
            closerTag = closerTag + "</%s>" %p.tag_
            p = p.parent

        return [openTag, closerTag]

    def getKeys(self):
        xpath = self.getXPath(False, False)
        manager = self.getNetworkElement()
        keymapping = manager.keymapping
        keys = None
        try:
            keys = keymapping[xpath]
        except:
            pass

        if keys:
            for key in keys:
                self.setId(key) 

        for child in self.children_:
            if isinstance(child,MgtObject) and not child.isleaf:
                child.getKeys()

    def getParent(self):
        """Get parent object of the MGT object."""
        return self.parent

    def getChildren(self):
        """Get children of the MGT object."""
        return self.children_

    def getRoot(self):
        """Get the root "directory" of the object tree"""
        if self.parent is None:
            return self

        return self.parent.getRoot()

    def getElementList(self):
        """Get list of network elements from the object tree."""
        if self.tag_ == '/':
            return self.children_

        if self.parent is None:
            return []

        return self.parent.getElementList()

    def getNetworkElement(self):
        """Get network element that is the upper level of the current object."""
        if self.tag_ == 'top':
            return self.children_[0] # TODO: lookup network element from mgt root object

        if self.parent:
            return self.parent.getNetworkElement()

        return None

    def getByPath(self, obj_path, target_name = None):
        """Get a MGT object based on its location path."""
        targets = self.getsByPath(obj_path, target_name)
        if len(targets) == 0:
            return None

        return targets[0]

    def getsByPath(self, obj_path, target_name = None):
        """Get MGT objects based on their location path."""
        target = []

        if obj_path == '':
            if target_name is None:
                target.append(self) 
            elif self.tag_ == target_name:
                target.append(self)
        else:
            tokens = obj_path.split('/')
            search_obj = self
            if tokens[0] == '':
                if len(tokens) > 1:
                    search_obj = self.getRoot()
                tokens = tokens[1:]

            if search_obj == None:
                search_obj = self.getRoot()

            if search_obj.isleaf == True:
                targets = []
            else:
                targets = search_obj.findChildren(tokens)

            if target_name is None:
                target = targets
            else:
                target = [x for x in targets if x.tag_ == target_name]

        return target

    # Private methods

    def setId(self, keyname):
        if self.children_ == []:
            return

        keys = [child for child in self.children_ if child.isleaf and (child.tag_ == keyname)]

        for key in keys:
            self.keys[key.tag_] = key.children_[0]

    def match(self, arg):

        if len(self.keys) == 1:
            try: 
                if re.match(arg + '$', self.keys.values()[0]):
                    return True
            except:
                traceback.print_exc()
                pass

        try: 
            if re.match(arg + '$', self.tag_):
                return True
        except:
            traceback.print_exc()
            pass

        try: 
            if re.match('(.+?)\[(.+?)\]$',arg):
                tokens = re.split('[\[,\,,\]]',arg)
                if not re.match(tokens[0] + '$', self.tag_):
                    return False

                tokens = tokens[1: -1]

                ret = True
                for token in tokens:
                    par_items = re.split('=',token)
                    if len(par_items) != 2:
                        ret = False
                        break

                    key = par_items[0].strip()
                    value = par_items[1].strip().replace('\"', '').replace('\'', '')

                    childs = []
                    for child in self.children_:
                        if child.isleaf and (child.tag_ == key):
                            if (len(child.children_) == 1):
                                if re.match(value + '$',child.children_[0]):
                                    childs.append(child)
                            elif (len(child.children_) > 1):
                                if child.children_.count(value) > 0:
                                    childs.append(child)
                    if childs == []:
                        ret = False
                        break

                return ret
        except:
            traceback.print_exc()
            pass

        return False
    
    def rpcRequest(self, rpcCommandsDict, *sp):
        """send rpc request to Netconf server with it's attribute
           rpcCommandDict: key: loaded rpc command - value: rpcStatement
           *sp: list of argument includes a command and command's attributes
        """
        if (len(sp) <= 0 or len(rpcCommandsDict) <= 0):
            return ""
        cmd = sp[0]
        rpcStatement = None
        ret = ""        
        for key in rpcCommandsDict:
            #Check if the rpc command is loaded
            if(cmd == key):
                rpcStatement = rpcCommandsDict[cmd]
                break
        if not rpcStatement:
            return "<error>Rpc command [%s] is not loaded</error>" % sp[0]
        else:
            manager = self.getNetworkElement()
            str = ""
            try:
                top = rpcStatement.top
                ns = top.search_one('namespace').arg
                if not ns:
                    return "<error>Namespace is not defined</error>"
                rpc = RPC(manager.manager._session, device_handler=DefaultDeviceHandler(), async=manager.manager._async_mode, timeout=manager.manager._timeout, raise_mode=manager.manager._raise_mode)
                openTag = "<%s xmlns=\"%s\">\n" % (cmd, ns)
                closeTag = "</%s>" %(cmd)
                str += openTag
                for i in sp[1:]:
                    k, v = "", ""
                    if '=' not in i:
                        return "<error>Wrong format for input {%s}</error>" %i
                    else:
                        k,v = i.split("=") #k,v = attribute & value
                    if (k and v):
                        attr = "<%s>%s</%s>\n" % (k,v,k)
                        str += attr
                    else:
                        return "<error>Wrong format for input {%s}</error>" %i
                str += closeTag
                result = rpc._request(etree.fromstring(str))
                return "%s" % result
            except Exception, e:
                print e
                return ""
          
    def getChildName(self, tokens = None):
        ret = []
        prefix = ''
        if tokens[0] == '':
            if self.tag_ == '/'  and len(tokens) > 1:
                prefix = '/'
            tokens = tokens[1:]

        try:
            if len(tokens) > 1:
                for child in self.children_:
                    if re.match(tokens[0] + '$', child.tag_) and (len(child.keys) == 1):
                        token1 = re.sub('\.','\\\.',tokens[1])
                        token1 = re.sub('\*','.*',token1)
                        if re.match(token1 + '$', child.keys.values()[0]):
                            tokens = tokens[1:]
                            break
        except:
            traceback.print_exc()
            pass

        try:
          if not tokens:
              return [prefix +  x.tag_ for x in self.children_]
          if len(tokens) == 0:
              return [prefix + x.tag_ + "/" for x in self.children_]
          elif len(tokens) == 1:
              ret = []
              for x in self.children_:
                keypath = ''
                if len(x.keys.keys()) > 0:
                  keypath = '[' + ",".join(map(lambda (k,v): "%s=%s" % (k,v), x.keys.iteritems())) + ']'
                ret.append(x.tag_ + keypath)
  
              ret = filter(lambda x: "%s%s"%(prefix, x) if x[0] == '[' else "%s%s/"%(prefix, x), filter(lambda x: x.startswith(tokens[-1]), ret))
              if (tokens[0] == '.' or tokens[0] == '..') and self.parent != None:
                  ret.append(prefix + '../')
              return ret
          else:
              if tokens[0] == '..':
                  if self.parent != None:
                      return [prefix + tokens[0] + '/' + p for p in self.parent.getChildName(tokens[1:])]
                  else:
                      return []
              else:
  
                  if re.match('\[(.+?)\]$',tokens[0]):
                      items = re.split('[\[,\,,\]]',tokens[0])
                      items = items[1:- 1]
                  else:
                      items = [tokens[0]]
  
                  lkitems = []
                  for item in items:
                      item = re.sub('\.','\\\.',item)
                      item = re.sub('\*','.*',item)
                      lkitems.append(item)
  
                  foundchild = None
                  for child in self.children_:
                      for item in lkitems:
                          if child.match(item):
                              foundchild = child
                              break
                      if foundchild is not None:
                          break
  
                  if foundchild is not None:
                      return [prefix + tokens[0] + '/' + p for p in foundchild.getChildName(tokens[1:])]
                  else:
                      return []
        except:
          traceback.print_exc()

    def findChildren(self, tokens = None):
        childs = []
        if len(tokens) > 0:
            if re.match('\[(.+?)\]$',tokens[0]):
                items = re.split('[\[,\,,\]]',tokens[0])
                items = items[1:- 1]
            else:
                items = [tokens[0]]

                try:
                    if len(tokens) > 1:
                        for child in self.children_:
                            if re.match(tokens[0] + '$', child.tag_) and (len(child.keys) == 1):
                                token1 = re.sub('\.','\\\.',tokens[1])
                                token1 = re.sub('\*','.*',token1)
                                if re.match(token1 + '$', child.keys.values()[0]):
                                    tokens = tokens[1:]
                                    items = [tokens[0]]
                                    break
                except:
                    traceback.print_exc()
                    pass

            lkitems = []
            for item in items:
                item = re.sub('\.','\\\.',item)
                item = re.sub('\*','.*',item)
                lkitems.append(item)

        if len(tokens) == 1:
            if tokens[0] == '':
                childs.append(self)
            elif tokens[0] == '..':
                if self.parent != None:
                    childs.append(self.parent)
            else:
                for child in self.children_:
                    for item in lkitems:
                        if child.match(item):
                            childs.append(child)
                            break

        elif len(tokens) > 1:
            if tokens[0] != '':
                if tokens[0] == '..':
                    if self.parent != None:
                        childs = self.parent.findChildren(tokens[1:])
                else:
                    for child in self.children_:
                        for item in lkitems:
                            if child.match(item):
                                childs = childs + child.findChildren(tokens[1:])
                                break

        return childs

    def addChild(self, c):
        c.parent = self
        self.children_.append(c)
        try:
            self.__dict__[microdom.keyify(c.tag_)] = c
            self.child_[microdom.keyify(c.tag_)] = c
        except:
            traceback.print_exc()

    def getCallbackCmd(self, cmdd):
        try:
            if self.callbackmethod is not None:
                return getattr(self.callbackmethod, cmdd)
            else:
                if self.parent is not None:
                    return self.parent.getCallbackCmd(cmdd)
                else:
                    return None
        except AttributeError:
            traceback.print_exc()
            return None

    def getCallbackName(self):
        if self.callbackmethod is not None:
            return dir(self.callbackmethod)
        else:
            if self.parent is not None:
                return self.parent.getCallbackName()
            else:
                return []

    def getNames(self):
        # This method used to pull in base class attributes
        # at a time dir() didn't do it yet.
        return dir(self.__class__)

    def ls(self):
        for child in self.children_:
            if child.__class__ == MgtObject:
                if child.isleaf:
                    if len(child.children_) == 1:
                        print "%-20s %s" %(child.tag_ + ':', child.children_[0])
                    else:
                        print "%-20s %s" %(child.tag_ + ':', child.children_)
                else:
                    if child.keys == {}:
                        print "%s" %child.tag_
                    else:
                        print "%-20s %s" %(child.tag_ + ':', child.keys)
            else:
                print "%s" %(child)

    def process(self, text, cmd):
        tokens = text.split('/')

        search_obj = self
        if tokens[0] == '':
            if len(tokens) > 1:
                search_obj = self.getRoot()

        if search_obj == None:
            search_obj = self.getRoot()

        objs = search_obj.getChildName(tokens)

        return objs

    def loadPlugin(self, arg):
        if not os.path.isfile(arg):
            print "No such file"
            return

        basename = os.path.basename(arg)
        if basename.endswith('.py'):
            basename = basename[0:-3]

        method = __import__(basename)
        method.__file__ = arg
        self.callbackmethod = method
        pass

    def deferredSet(self,val):
       """Return the server, an XML string needed to actually set this entity, and a fn to set the value locally.  This XML string needs to be wrapped in the standard netconf <confg ...> wrapper"""
  
       def doSet(val=val):
         self.children_ = [val]
         self.data_ = val

       if self.isleaf:           
         self.children_ = [val]
         self.data_ = val
         manager = self.getNetworkElement()
         xml = "%s%s%s" % (self.getXmlTag()[0], val, self.getXmlTag()[1])
         return (manager,xml,doSet)
       else:
         dropIntoPdb()
         assert(0)

    def setValue(self, val,database):
        if self.isleaf:
            
            manager = self.getNetworkElement()

            self.children_ = [val]
            self.data_ = val

            conf_xml = snippet % "%s%s%s" %(self.getXmlTag()[0], val, self.getXmlTag()[1])
            #print conf_xml
            logging.debug("Configuration xml: %s", conf_xml)

            try:
                manager.manager.edit_config(target=database, config=conf_xml)
            except:
               dropIntoPdb()
#               traceback.print_exc()
               raise SetOperationFailed(self.fullPath(),val)           
        else:
            dropIntoPdb()
            assert(0)

    def getConfigValue(self,xpath,database):
        #xpath = "/" + self.getXPath(True,False)
        #print 'getConfigValue:xpath=%s, database=%s'%(xpath,database)
        manager = self.getNetworkElement()
        output = manager.manager.get_config(source=database, filter=("xpath", xpath))
        #print 'getConfigValue:output='+str(output)
        result = ''
        doms = minidom.parseString(str(output))
        elements = doms.getElementsByTagName("data")
        if elements != [] and elements[0].childNodes:
            for element in elements[0].childNodes:
                result += element.toxml()
        return result

    def getValue(self):
        xpath = "/" + self.getXPath(True,False)
        manager = self.getNetworkElement()
        output = manager.manager.get(filter=("xpath", xpath))

        result = ''
        doms = minidom.parseString(str(output))
        elements = doms.getElementsByTagName("data")
        if elements != [] and elements[0].childNodes:
            for element in elements[0].childNodes:
                result += element.toxml()
        return result

    def deleteValue(self, value, database):
        manager = self.getNetworkElement()

        ele = self.convertXPath2Ele(value)
        # Add extra xml attribute - this need nc:config xml template 
        ele.attrib = {'xc:operation': "delete"}

        key = "\n".join(map(lambda (k,v): "<%s>%s</%s>" % (k,v,k), self.keys.iteritems()))

        #Full xml: <openTag>value</closerTag>
        xml = "%s%s%s%s" %(self.getXmlTag()[0], key, ET.tostring(ele), self.getXmlTag()[1])
        conf_xml = snippet % xml

        #print conf_xml
        logging.debug("Deleting xml: %s", conf_xml)

        try:
            manager.manager.edit_config(target=database, config=conf_xml)
        except:
            traceback.print_exc()
            return False

        return True

    '''def createValue(self, value,database):
        manager = self.getNetworkElement()

        ele = self.convertXPath2Ele(value)
        # Add extra xml attribute
        ele.attrib = {'xc:operation': "create"}

        key = "\n".join(map(lambda (k,v): "<%s>%s</%s>" % (k,v,k), self.keys.iteritems()))

        #Full xml: <openTag>value</closerTag>
        xml = "%s%s%s%s" %(self.getXmlTag()[0], key, ET.tostring(ele), self.getXmlTag()[1])
        conf_xml = snippet % xml

        #print conf_xml
        logging.debug("Creating xml: %s", conf_xml)

        try:
            manager.manager.edit_config(target=database, config=conf_xml, default_operation="merge")
        except:
            traceback.print_exc()
            return False
        return True'''


    def valueOperation(self, value,operation,database):
        manager = self.getNetworkElement()

        ele = self.convertXPath2Ele(value)
        # Add extra xml attribute
        ele.attrib = {'xc:operation': operation}

        key = "\n".join(map(lambda (k,v): "<%s>%s</%s>" % (k,v,k), self.keys.iteritems()))

        logging.debug("key=%s",key)
        logging.debug("ele=%s",ET.tostring(ele))

        #Full xml: <openTag>value</closerTag>
        xml = "%s%s%s%s" %(self.getXmlTag()[0], key, ET.tostring(ele), self.getXmlTag()[1])
        conf_xml = snippet % xml

        #print conf_xml
        logging.debug("%sing xml: %s", operation,conf_xml)

        try:
            manager.manager.edit_config(target=database, config=conf_xml, default_operation="merge")
        except:
            traceback.print_exc()
            return False
        return True

class MgtRootObject(MgtObject):
    
    def __init__(self, attributes,children,data):
        MgtObject.__init__(self, None,attributes,children,data)

    def ls(self):
        for child in self.children_:
            md = microdom.LoadMiniDom(child.config)
            host = '%s' %md.host.data_
            status = ''
            if child.connected == False:
                status = 'DOWN!'
            print "%-20s %-30s \033[0;31m%s\033[0m" %(host + ':', child.tag_, status) 

    def fullPath(self):
      return ""

class MgtGroupObject(MgtObject):
    def __init__(self, attributes,children,data):
        MgtObject.__init__(self, attributes,children,data)

    def ls(self):
        for child in self.children_:
            print "%s:" %(child.getXPath(True))
            child.ls()
            print ""

    def addChildren(self, children):
        self.children_ = children

    def findChildren(self, tokens = None):
        ret = MgtObject.findChildren(self, tokens)
        if ret != []:
            return ret

        for child in self.children_:
            childs = child.findChildren(tokens)
            ret = ret + childs
        return ret

def getDomName(dom):
    try:
        return dom.nodeName.split(':')[1]
    except Exception, ex:
        return dom.nodeName

def loadFile(filename):
    dom = minidom.parse(filename)
    elem = LoadMiniDom(dom.childNodes[0])
    return elem

def LoadMiniDom(dom):
    if type(dom) is ListType:
        return [LoadMiniDom(d) for d in dom]

    childlist = []
    dataLst=[]

    name = getDomName(dom)

    attrlist = { "tag_": name}

    if dom.childNodes:
        for child in dom.childNodes:
            if child.nodeName == '#text' or child.nodeName == '#cdata-section':  # Raw data
                if not child.nodeValue.startswith('\n '):
                    childlist.append(str(child.nodeValue))
                    dataLst.append(str(child.nodeValue))
            else:
                childlist.append(LoadMiniDom(child))

    for (key,val) in dom._attrs.items():
        attrlist[key] = val.value

    try: 
        data = dom.data
    except AttributeError:
        data = "".join(dataLst) 

    try:
        data = int(data)
        goodAsInt = True
    except ValueError:
        try:
            data = float(data)
        except ValueError:
            data = str(data)
            if data in microdom.Truths: data = True
            if data in microdom.Falses: data = False
  
    tmp = MgtObject(None, attrlist, childlist, data)
    for c in childlist:
      try:
        c.parent = tmp
      except:
        pass
    return tmp
