/*
 * Copyright (C) 2002-2012 OpenClovis Solutions Inc.  All Rights Reserved.
 *
 * This file is available  under  a  commercial  license  from  the
 * copyright  holder or the GNU General Public License Version 2.0.
 * 
 * The source code for  this program is not published  or otherwise 
 * divested of  its trade secrets, irrespective  of  what  has been 
 * deposited with the U.S. Copyright office.
 * 
 * This program is distributed in the  hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied  warranty  of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 * 
 * For more  information, see  the file  COPYING provided with this
 * material.
 *
 */

#ifndef MGTOM_HXX_
#define MGTOM_HXX_

#include <string>
#include <map>
#include <list>
#include <clMgtApi.hxx>
#include <clHandleApi.hxx>
#include "clMgtDatabase.hxx"

namespace MgtServer
{

  /*
   * This class is used to store the data binding of a component.
   * Example:
   *      componentA -> bind -> new clMgtOM() -> insert into hash map
   *
   */
  class MgtOM
  {
  public:
    Mgt::Msg::MsgBind bindData;
    ClUint32T myRev;

    MgtOM(Mgt::Msg::MsgBind *bindDataComp)
    {
      bindData = *bindDataComp;
      myRev = 0;
    }
    /**
     * Write cache data to database
     */
    ClRcT write(std::string& xml, SAFplus::MgtDatabase *db = NULL)
    {
      ClRcT rc = CL_OK;
      std::string value;
      std::string key;
      if (db == NULL)
        return CL_FALSE;

      /* Update or Insert record */
      key.assign(toKey());
      rc = db->getRecord(key, value);
      if (rc != CL_OK)
        {
          return db->insertRecord(key, xml);
        }
      else
        {
          return db->setRecord(key, xml);
        }
    }
    /**
     * Read cache data from database
     */
    ClRcT read(std::string &value, SAFplus::MgtDatabase *db = NULL)
    {
      ClRcT rc = CL_OK;
      std::string key;
      if (db == NULL)
        return CL_FALSE;
      key.assign(toKey());
      rc = db->getRecord(key, value);
      return rc;
    }
    bool valid(void)
    {
      if (this == NULL)
        return false;
      /* TODO: check valid */
      return true;
    }

    /*
     * Create Key from object
     */
    const char* toKey()
    {
      std::stringstream ss;
      ss << bindData.module().c_str();
      ss << ":/";
      ss << bindData.route().c_str();
      return ss.str().c_str();
    }

    MgtOM()
    {
       myRev = 0;
    }
  };

  class MgtModuleOM
  {
  public:
    MgtModuleOM();
    ~MgtModuleOM();
    std::map<std::string, MgtOM *> mOMList;
    MgtOM * findOM(const std::string& route)
    {
      std::map<std::string, MgtOM*>::iterator omIndex = mOMList.find(route);
      if (omIndex != mOMList.end())
        {
          return static_cast<MgtOM *>((*omIndex).second);
        }
      return NULL;
    }

    MgtOM * findOM(const std::string& route, bool exact)
    {
      if (exact)
        {
          return findOM(route);
        }
      else
        {
          std::map<std::string, MgtOM*>::iterator omEntry = mOMList.begin();
          while (omEntry != mOMList.end())
            {
              std::string omKey = static_cast<std::string>((*omEntry).first);
              if (omKey.find(route) != std::string::npos)
                {
                  return static_cast<MgtOM *>((*omEntry).second);
                }
              omEntry++;
            }
        }
      return NULL;
    }
  };

  class MgtRootOM
  {
  private:
    MgtRootOM() {};
    static MgtRootOM* instance;

  public:
    ~MgtRootOM();
    static MgtRootOM *getInstance()
    {
      if (instance == NULL)
        {
          instance = new MgtRootOM();
        }
      return instance;
    }

    static void DestroyInstance()
    {
      delete instance;
      instance = 0;
    }

    std::map<std::string, MgtModuleOM*> mModuleOMs;
    /*
     * Retrieve all routes register on module
     */
    MgtModuleOM * findModuleOM(const std::string& module)
    {
      std::map<std::string, MgtModuleOM*>::iterator moduleIndex = mModuleOMs.find(module);
      if (moduleIndex != mModuleOMs.end())
        {
          return static_cast<MgtModuleOM *>((*moduleIndex).second);
        }
      return NULL;
    }

    /*
     * Retrieve OM entry register on module and route
     */
    MgtOM * findOM(const std::string& module, const std::string& route)
    {
      MgtModuleOM *moduleOM = findModuleOM(module);
      if (moduleOM != NULL)
        {
          return moduleOM->findOM(route);
        }
      return NULL;
    }

    std::list<MgtOM*> findListOM(SAFplus::Handle findHdl)
    {
      std::list<MgtOM*> el;
      std::map<std::string, MgtModuleOM*>::iterator moduleIndex = mModuleOMs.begin();
      while (moduleIndex != mModuleOMs.end())
        {
          MgtModuleOM *moduleOM = static_cast<MgtModuleOM *>((*moduleIndex).second);
          std::map<std::string, MgtOM*>::iterator omIndex = moduleOM->mOMList.begin();
          while (omIndex != moduleOM->mOMList.end())
            {
              MgtOM *omEntry = static_cast<MgtOM *>((*omIndex).second);
              SAFplus::Handle hdl = SAFplus::Handle::create();
              hdl.id[0] = omEntry->bindData.handle().id0();
              hdl.id[1] = omEntry->bindData.handle().id1();
              if (hdl.getNode() == findHdl.getNode())
                {
                  switch (findHdl.getPort())
                    {
                  case 0: // Case node address
                    el.push_back(omEntry);
                    break;
                  default: // Case node and port
                    if (hdl.getPort() == findHdl.getPort())
                      {
                        el.push_back(omEntry);
                      }
                    break;
                    }
                }
              omIndex++;
            }
          moduleIndex++;
        }
      return el;
    }

    SAFplus::Handle findHandle(const std::string& pathSpec)
    {
      SAFplus::Handle hdl = SAFplus::Handle::create(0, 0);

      std::map<std::string, MgtModuleOM*>::iterator moduleIndex = mModuleOMs.begin();
      while (moduleIndex != mModuleOMs.end())
        {
          MgtModuleOM *moduleOM = static_cast<MgtModuleOM *>((*moduleIndex).second);
          std::map<std::string, MgtOM*>::iterator omIndex = moduleOM->mOMList.begin();
          while (omIndex != moduleOM->mOMList.end())
            {
              MgtOM *omEntry = static_cast<MgtOM *>((*omIndex).second);
              std::string omKey =  omEntry->bindData.route();

              if (pathSpec.find(omKey) != std::string::npos)
                {
                  hdl.id[0] = omEntry->bindData.handle().id0();
                  hdl.id[1] = omEntry->bindData.handle().id1();
                  return hdl;
                }

              omIndex++;
            }
          moduleIndex++;
        }

      return hdl;
    }
  };

}
#endif /* CLMGTOM_HXX_ */
