/*
 * Copyright (C) 2002-2012 OpenClovis Solutions Inc.  All Rights Reserved.
 *
 * This file is available  under  a  commercial  license  from  the
 * copyright  holder or the GNU General Public License Version 2.0.
 *
 * The source code for  this program is not published  or otherwise
 * divested of  its trade secrets, irrespective  of  what  has been
 * deposited with the U.S. Copyright office.
 *
 * This program is distributed in the  hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * For more  information, see  the file  COPYING provided with this
 * material.
 *
 */

// Standard C/C++ includes
#include "assert.h"
#include <sstream>
#include <map>

#include <boost/filesystem.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/join.hpp>

// SAFplus includes.
#include <clLogApi.hxx>
#include <clSafplusMsgServer.hxx>
#include <clMsgPortsAndTypes.hxx>

#include "clMgtError.hxx"
#include <clCkptIpi.hxx>

#include "clMgtNetconfHandler.hxx"
#include "clMgtNetconfYuma.hxx"

using namespace std;
using namespace Mgt::Msg;
using namespace SAFplus;
using namespace SAFplusI;

namespace MgtServer
{
  SAFplus::Mutex sendReplyNetconfMutex;

  static ClCharT datapath[CL_MAX_NAME_LENGTH] = { 0 };

  MgtNetconfHandler* MgtNetconfHandler::instance = 0;

  log_debug_t getLogSeverity(void)
  {
    log_debug_t ncx_log_level = LOG_DEBUG_INFO;
    switch (SAFplus::logSeverity)
    {
      case LogSeverity::LOG_SEV_EMERGENCY:
      case LogSeverity::LOG_SEV_ALERT:
      case LogSeverity::LOG_SEV_CRITICAL:
      case LogSeverity::LOG_SEV_ERROR:
        ncx_log_level = LOG_DEBUG_ERROR;
        break;

      case LogSeverity::LOG_SEV_WARNING:
      case LogSeverity::LOG_SEV_NOTICE:
      {
        ncx_log_level = LOG_DEBUG_WARN;
        break;
      }

      case LogSeverity::LOG_SEV_INFO:
      {
        ncx_log_level = LOG_DEBUG_INFO;
        break;
      }

      case LogSeverity::LOG_SEV_DEBUG:
      {
        ncx_log_level = LOG_DEBUG_DEBUG;
        break;
      }

      default:
        ncx_log_level = LOG_DEBUG_DEBUG4;
    }
    return ncx_log_level;
  }

  /**
   * MgtNetconfHandle constructor
   */
  MgtNetconfHandler::MgtNetconfHandler()
  {
  }
  /**
   * Singleton instance of mgtNetconfHandle
   */
  MgtNetconfHandler* MgtNetconfHandler::getInstance()
  {
    if (instance == NULL)
    {
      instance = new MgtNetconfHandler();
    }
    return instance;
  }

  void MgtNetconfHandler::DestroyInstance()
  {
    delete instance;
    instance = 0;
  }

  std::string MgtNetconfHandler::dumpValStr(val_value_t *val, ncx_instfmt_t fmt)
  {
    status_t res;
    xmlChar *valStr;
    std::string out = "";
    res = val_gen_instance_id(NULL, val->virtualval ? val->virtualval : val, fmt, &valStr);
    if (res == NO_ERR && valStr)
    {
      out.assign((const char *) valStr);
      free(valStr);
    }
    return out;
  }

  /**
   *
   */
  ClRcT MgtNetconfHandler::netconfInitialize(int argc, ClCharT *argv[])
  {
    status_t rc;
    ClCharT *aspdir = NULL;
    help_mode_t showhelpmode;
    boolean showver = FALSE;
    log_debug_t ncx_log_level = getLogSeverity();

    /*
     * ncx_init
     * CL_OK == NO_ERR
     */
    rc = ncx_init(false, ncx_log_level, true, NULL, argc, argv);
    if (rc == ERR_NCX_MOD_NOT_FOUND)
    {
      throw MgtError(MgtError::YUMA_ERROR, rc, "Initialization failed");
    }
    assert(rc == NO_ERR);

    /* load in the server boot parameter definition file.  This needs to go before agt_init1 since that function uses the module.
     */
    rc = ncxmod_load_module(NCXMOD_NETCONFD, NULL, NULL, NULL);
    assert(rc == NO_ERR);

    rc = agt_init1(argc, argv, &showver, &showhelpmode);

    agt_profile_t *profile = agt_get_profile();

    /* load in the NETCONF data types and RPC methods */
    rc = ncxmod_load_module(NCXMOD_YUMA_NETCONF, NULL, NULL, NULL);
    assert(rc == NO_ERR);

    /*
     * Setting default startup transaction Id into datapath
     */
    if (NULL != (aspdir = getenv("ASP_DIR")))
    {
      snprintf(datapath, CL_MAX_NAME_LENGTH, "%s/%s", aspdir, "share/netconf");
    }
    /*
     * All yang modules are located in ...
     */
    else if (NULL != (aspdir = getenv("YUMA_HOME")))
    {
      snprintf(datapath, CL_MAX_NAME_LENGTH, "%s", aspdir);
    } else
    {
      ClCharT cwd[CL_MAX_NAME_LENGTH] = { 0 };
      if (NULL == getcwd(cwd, CL_MAX_NAME_LENGTH))
      {
        logDebug("MGT", "BIND", "Getting current working directory failed");
      }
      snprintf(datapath, CL_MAX_NAME_LENGTH, "%s/%s", cwd, "share/netconf");
      boost::filesystem::path p(datapath);
      if (!boost::filesystem::is_directory(p))
      {
        snprintf(datapath, CL_MAX_NAME_LENGTH, "%s/../%s", cwd, "share/netconf");
      }

    }

    boost::filesystem::path p(datapath);
    if (!boost::filesystem::is_directory(p))
    {
      char t[256];
      snprintf(t, 256, "Data path [%s] does not exist or is not a directory.", datapath);
      throw MgtErrorStr(MgtError::SYSTEM_ERROR, ENOENT, t, __FILE__, __LINE__);
    }

    /*
     * Configure data path to store netconf data
     */
    ncxmod_set_datapath((const xmlChar*) datapath);

#if 0
    /*
     * Load netconf module default
     */
    rc = ncxmod_load_module(NCXMOD_WITH_DEFAULTS, NULL, &profile->agt_savedevQ, NULL);
    assert(rc == NO_ERR);
#endif

    rc = agt_init2();
    assert(rc == NO_ERR);

    /*
     * Configure log level (This will overwrite the param --log-level and netconfd.conf configuration)
     */
    bool userDefinedLogLevel = false;
    for (int i = 0; i < argc; ++i)
    {
      if (strstr(argv[i], "--log-level"))
      {
        userDefinedLogLevel = true;
        break;
      }
    }
    if (!userDefinedLogLevel)
    {
      logDebug("MGT", "INI", "configure netconf log level: overwrite the --log-level and netconfd.conf configuration");
      log_set_debug_level(ncx_log_level);
    }

    return rc;
  }

  /*
   * This helper function is to generate a val_value_t to an pair <xpath=value>
   */
  status_t mgtNetconfSetHelper(const val_value_t *val, std::map<std::string, std::string> &bulkMap)
  {
    status_t res = NO_ERR;
    obj_type_t objType = val->obj->objtype;

    /*
     * Bulk setting values
     */
    switch (objType)
    {
      case NCX_BT_ANY:
      case NCX_BT_CONTAINER:
      case OBJ_TYP_CONTAINER:
      case OBJ_TYP_LIST:
      {
        /*
         * Handle child values
         */
        val_value_t *chval;
        for (chval = val_get_first_child(val); chval != NULL; chval = val_get_next_child(chval))
        {
          if (!chval->index) res = mgtNetconfSetHelper(chval, bulkMap);

          if (res != NO_ERR)
            return res;
        }
      }
      break;
      case OBJ_TYP_LEAF:
      {
        xmlChar *buffXpath;
        xmlChar *buffValue;

        res = val_gen_instance_id(NULL, val, NCX_IFMT_XPATH1, &buffXpath);
        if (res != NO_ERR)
          return res;

        buffValue = val_make_sprintf_string(val);
        if (buffValue)
        {
          std::string xpath(MgtYuma::yumaXPathParse(buffXpath));
          std::string value((const char *) buffValue);
          bulkMap.insert(std::pair<std::string, std::string>(xpath, value));
        }

        if (buffValue)
          free(buffValue);

        if (buffXpath)
          free(buffXpath);
      }
      break;

      default:
      break;
    }
    return res;
  }
  xmlNodePtr xmlGetNode(const xmlNodePtr node, const xmlChar *keyword)
  {
    xmlNodePtr cur = node;
    cur = cur->xmlChildrenNode;
    while (cur != NULL)
    {
      if (!xmlStrcmp(cur->name, keyword))
      {
        return cur;
      }
      cur = cur->next;
    }
    return NULL;
  }

  status_t mgtRpcOutput(rpc_msg_t *msg, std::string output)
  {
    status_t res = NO_ERR;
    obj_template_t *entryobj;
    bool isSessionLocalCreate = false;
    xml_node_t top;
    xml_init_node(&top);
    /* create a dummy session control block */
    ses_cb_t *scbDummy = agt_ses_get_dummy_session();
    if (scbDummy == NULL)
    {
      scbDummy = agt_ses_new_dummy_session();
      if (scbDummy == NULL)
      {
        res = ERR_INTERNAL_MEM;
      } else
      {
        isSessionLocalCreate = true;
      }
    }
    if (!scbDummy)
    {
      res = ERR_INTERNAL_MEM;
    } else
    {
      scbDummy->reader = xmlReaderForMemory((const char*) output.c_str(), output.length(), NULL, NULL, 0);
      res = agt_xml_consume_node(scbDummy, &top, NCX_LAYER_NONE, NULL);
      if (res == NO_ERR)
      {
        entryobj = obj_find_child(msg->rpc_method, obj_get_mod_name(msg->rpc_method), top.elname);
        val_value_t *entryval = val_new_value();
        res = agt_val_parse_nc(scbDummy, &msg->mhdr, entryobj, &top, NCX_DC_STATE, entryval);
        if (res == NO_ERR)
        {
          msg->rpc_data_type = RPC_DATA_STD;
          dlq_enque(entryval, &msg->rpc_dataQ);
        }
      }
      if (isSessionLocalCreate) agt_ses_free_dummy_session(scbDummy);
    }
    return res;
  }

static void print_element_names(xmlNodePtr  a_node)
    {
      xmlNodePtr cur_node = NULL;
      for (cur_node = a_node; cur_node; cur_node = cur_node->next) 
      {
        if (cur_node->type == XML_ELEMENT_NODE) 
        {
            logDebug("MGT", "YUMA","node type: Element, name: %s\n", cur_node->name);
        }
        print_element_names(cur_node->children);
    }
}
  /**
   * Edit data call for object
   */
  status_t MgtNetconfHandler::mgtNetconfRpc(Mgt::Msg::MsgRpc_MgtRpcType rpcType, ses_cb_t *scb, rpc_msg_t *msg, xml_node_t *methnode)
  {
    status_t res = NO_ERR;
    res = NO_ERR;
    logDebug("MGT", "YUMA", "[%s:%s] RPC ", (const char*) methnode->module, (const char*) methnode->qname);
    const char* result;
    xmlDocPtr doc;
    xmlNodePtr cur, node;
    xmlBufferPtr buf;
    const char* buff = reinterpret_cast<const char*>(scb->readbuff);
    const char* xmlTag = strstr(buff, "<?xml");
    if (xmlTag == NULL)
    {
      logError("MGT", "RPC", "Invalid rpc request: wrong xml document");
      return ERR_XML2_FAILED;
    }
    const char* rpcClosedTag = strstr(buff, "</nc:rpc>");
    if (rpcClosedTag == NULL)
    {
      logError("MGT", "RPC", "Invalid rpc request: missing rpc closed tag");
      return ERR_XML2_FAILED;
    }
    const char* p = rpcClosedTag;
    while (p && *p++ != '>');
    int size = p - xmlTag;
    char xml[size + 1];
    strncpy(xml, xmlTag, size);
    xml[size] = '\0';
    doc = xmlParseDoc((xmlChar*)xml);
    cur = xmlDocGetRootElement(doc);
    std::string param;
    if (cur == nullptr)
    {
      xmlFreeDoc(doc);
    } else
    {
      node = xmlGetNode(cur, methnode->elname);
      if(node->children != NULL)
      {
        param = "<input>";
        node = node->children;
        do
        {
          buf = xmlBufferCreate();
          xmlNodeDump(buf,doc,node, 0, 0 );
          node=node->next;
          param.append((const char*)xmlBufferContent(buf));
        }while(node!=NULL);
        param.append("</input>");
        xmlBufferFree(buf);
      }
      else
      {
        logDebug("MGT", "YUMA", "no input parameter");
        param="";
      }
    }
    std::string xpath = "/";
    xpath.append((const char*) methnode->elname);
    std::string xmlResult=SAFplus::mgtRpc(rpcType,xpath,param);
    if (!xmlResult.empty())
    {
         if(rpcType==Mgt::Msg::MsgRpc::CL_MGT_RPC_INVOKE)
         {
           mgtRpcOutput(msg, xmlResult);
         }
    }
    else
    {
      res = ERR_INTERNAL_VAL;
    }
    xmlFreeDoc(doc);
    return res;
  }
  status_t MgtNetconfHandler::mgtNetconfSet(op_editop_t editop, val_value_t *newval)
  {
    status_t res = NO_ERR;
    const char *moduleName = (const char *) obj_get_mod_name(newval->obj);

    switch (editop)
    {
      /*
       * Apply deleted
       */
      case OP_EDITOP_CREATE:
      case OP_EDITOP_DELETE:
      {
        ClRcT retCode = CL_OK;
        xmlChar *buffXpath;
        if (newval->virtualval)
        {
          res = val_gen_instance_id(NULL, newval->virtualval, NCX_IFMT_XPATH1, &buffXpath);
        }
        else
        {
          res = val_gen_instance_id(NULL, newval, NCX_IFMT_XPATH1, &buffXpath);
        }
        if (res != NO_ERR)
          return res;

        std::string strXpath(MgtYuma::yumaXPathParse(buffXpath));
        free(buffXpath);

        if (editop == OP_EDITOP_DELETE)
          {
            retCode = SAFplus::mgtDelete(strXpath);
          }
        else
          {
            retCode = SAFplus::mgtCreate(strXpath);
          }

        if (retCode != CL_OK && retCode != CL_ERR_DOESNT_EXIST)
          {
            logError("NETC", "CRUD", "%s xpath [%s] got error [%d]", editop == OP_EDITOP_DELETE?"Delete":"Create", strXpath.c_str(), retCode);
            return SET_ERROR(ERR_INTERNAL_VAL);
          }
        logDebug("NETC", "CRUD", "%s xpath [%s] succeeded", editop == OP_EDITOP_DELETE?"Delete":"Create", strXpath.c_str());
      }
      if (editop == OP_EDITOP_DELETE) break;

      case OP_EDITOP_REPLACE:
      case OP_EDITOP_MERGE:
      {
        ClRcT retCode = CL_OK;
        std::map < std::string, std::string > bulkMap;
        res = mgtNetconfSetHelper(newval, bulkMap);
        if (res != NO_ERR)
          return res;

        /*
         * TODO: reduce multi calling to mgtSet, instead of combine into one shot for all xpaths:values
         */
        for (std::map<std::string, std::string>::iterator it = bulkMap.begin(); it!=bulkMap.end(); ++it)
        {
          retCode = SAFplus::mgtSet((*it).first, (*it).second);
          if (retCode != CL_OK)
          {
            logError("NETC", "CRUD", "Update xpath [%s] with value [%s] got error [%d]", (*it).first.c_str(), (*it).second.c_str(), retCode);
            return SET_ERROR(ERR_INTERNAL_VAL);
          }
          logDebug("NETC", "CRUD", "Update xpath [%s] with value [%s] succeeded", (*it).first.c_str(), (*it).second.c_str());
        }
      }
      break;
      default:
        break;
      }

    return res;
  }

  string MgtNetconfHandler::getKeysObj(const obj_template_t *obj)
  {
    std::vector<std::string> keyList;
    /*
     * Sanity check key if list type only
     */
    if (obj->objtype != OBJ_TYP_LIST)
    {
      return "";
    }

    /* for a list, replace key with a leaf name */
    const obj_key_t *key;
    for (key = obj_first_ckey(obj);
         key != NULL;
         key = obj_next_ckey(key))
    {
      keyList.push_back((char *)obj_get_name(key->keyobj));
    }

    return boost::algorithm::join(keyList, ",");
  }

  /*
   * This helper function to convert XML to customize data type, check and wrap into pseudo 'top' level
   */
  string MgtNetconfHandler::preformatXML(val_value_t *dstval, const char *outVal)
  {
    string parseStr(outVal);
    if (!obj_is_top(dstval->obj) || (obj_is_top(dstval->obj) && (dstval->obj->objtype == OBJ_TYP_LIST)))
    {
        // TODO: work-around to consume top node xml
        std::stringstream ss;
        ss << "<top>" << parseStr << "</top>";
        parseStr.assign(ss.str());
    }
    return parseStr;
  }

  void MgtNetconfHandler::mgtNetconfParserXmlData(const xmlChar *moduleName, val_value_t *virval, val_value_t *dstval, xmlChar *outVal)
  {
    ses_cb_t *scb;
    xml_msg_hdr_t msg;
    status_t res;
    xml_node_t top, chnode;
    boolean done = false;
    boolean isSessionLocalCreate = false;
    obj_template_t *entryobj;
    val_value_t *entryval;
    int numEntryCounter = 0; // Number mgt objs returning from mgtGet

    res = NO_ERR;

    xml_msg_init_hdr(&msg);
    xml_init_node(&top);

    /* create a dummy session control block */
    scb = agt_ses_get_dummy_session();
    if (scb == NULL)
    {
      scb = agt_ses_new_dummy_session();
      if(scb == NULL)
      {
        SET_ERROR(ERR_INTERNAL_MEM);
        return;
      }
      isSessionLocalCreate = true;
    }
    string parseStr = preformatXML(dstval, (const char *)outVal);

    scb->reader = xmlReaderForMemory((const char*) parseStr.c_str(), parseStr.length(), NULL, NULL, 0);
    res = agt_xml_consume_node(scb, &top, NCX_LAYER_RPC, NULL);
    if (res != NO_ERR)
      {
        logDebug("MGT", "GET", "Parsing Data returned error [%d]. Module [%s] Data [%s]", res, (const char*) moduleName, parseStr.c_str());
        goto out;
      }

    if (!xml_strcmp(top.elname, virval->name))
    {
      res = agt_val_parse_nc(scb, &msg, virval->obj, &top, NCX_DC_STATE, dstval);
      if (res != NO_ERR)
        SET_ERROR(ERR_INTERNAL_MEM);
      goto out;
    }

    xml_init_node(&chnode);

    /* go through each child node until the parent end node */
    while (!done)
      {
        /* get the next node which should be a child or end node */
        res = agt_xml_consume_node(scb, &chnode, NCX_LAYER_RPC, &msg);
        if (res != NO_ERR)
          {
            goto out;
          }
        else
          {
            /* validate the child member node type */
            switch (chnode.nodetyp)
              {
            case XML_NT_START:
            case XML_NT_EMPTY:
              /* any namespace OK for now, check it in get_child_node */
              break;
            case XML_NT_STRING:
              res = ERR_NCX_WRONG_NODETYP_SIM;
              goto out;
            case XML_NT_END:
              res = xml_endnode_match(&top, &chnode);
              if (res != NO_ERR)
                {
                  goto out;
                }
              else
                {
                  /* no error exit */
                  done = TRUE;
                  continue;
                }
              break;
            default:
              res = ERR_NCX_WRONG_NODETYP;
              goto out;
              }
          }

          entryobj = obj_find_child(dstval->obj, moduleName, chnode.elname);
          if (entryobj)
          {
            /* Create a new value & map to the template */
            entryval = val_new_value();
            val_init_from_template(entryval, entryobj);
            res = agt_val_parse_nc(scb, &msg, entryobj, &chnode, NCX_DC_STATE, entryval);
            entryval->cachetime = 0;
            val_add_child(entryval, dstval);
          }
          else if (obj_is_top(virval->obj) && virval->obj->objtype == OBJ_TYP_LIST)
          {
            // Sync-up data with checkpoint for the first XML entry
            if (!numEntryCounter)
            {
              res = agt_val_parse_nc(scb, &msg, virval->obj, &chnode, NCX_DC_STATE, dstval);

              vector <val_value_t*> lst;
              val_value_t *ch = val_find_child(virval->parent, moduleName, chnode.elname);
              /*
               * Marked all values with 'proxied' virtual value as deleted
               */
              while (ch)
              {
                if (ch != virval)
                  {
                    VAL_MARK_DELETED(ch);
                    lst.push_back(ch);
                  }
                ch = val_find_next_child(virval->parent, moduleName, chnode.elname, ch);
              }
              // Actually deleted
              for (vector <val_value_t*>::iterator it = lst.begin(); it != lst.end(); ++it)
              {
                val_remove_child(*it);
                val_free_value(*it);
              }
            }
            else
            {
              entryval = val_new_value();
              val_init_from_template(entryval, virval->obj);
              res = agt_val_parse_nc(scb, &msg, entryval->obj, &chnode, NCX_DC_STATE, entryval);
              val_add_child(entryval, virval->parent);
            }
            numEntryCounter++;
          }

        /*
         * Ignore all sub-childs
         */
        agt_xml_skip_subtree(scb, &chnode);
        xml_clean_node(&chnode);
      }

    out:
    /* clean up and exit */
    if ((scb != NULL) && isSessionLocalCreate)
    {
      agt_ses_free_dummy_session(scb);
    }
    xml_clean_node(&top);
    xml_msg_clean_hdr(&msg);
  }

  /**
   * Get and save data for module
   */
  void MgtNetconfHandler::mgtNetconfGet(const xmlChar *moduleName, val_value_t *virval, val_value_t *dstval)
  {
    assert(moduleName);
    assert(dstval);

    /* Find registered MOs in the database */
    //string modName((char*) moduleName);

    std::string xpath = "{n}/";
    //xpath.append(modName).append("/").append((const char*) dstval->name);
    xpath.append((const char*) dstval->name);

    logDebug("NETC", "GET", "Trying [%s]", xpath.c_str());
    std::string xmlResult = SAFplus::mgtGet(xpath);
    if (!xmlResult.empty())
      {
        logDebug("NETC", "GET", "received\n%s\n", xmlResult.c_str());

        /* Fill object data into the tree */
        mgtNetconfParserXmlData(moduleName, virval, dstval, (xmlChar *) xmlResult.c_str());
      }
  }

  void MgtNetconfHandler::mgtNetconfNotifParserXmlData(const xmlChar* moduleName, agt_not_msg_t *notifMsg, obj_template_t *notiobj, const xmlChar *outVal)
  {
    ses_cb_t *scb;
    val_value_t *entryval;
    val_value_t *child;
    xml_msg_hdr_t msg;
    status_t res;
    xml_node_t top;
    bool isSessionLocalCreate = false;
    res = NO_ERR;

    xml_msg_init_hdr(&msg);
    xml_init_node(&top);

    /* create a dummy session control block */
    scb = agt_ses_get_dummy_session();
    if (scb == NULL)
    {
      scb = agt_ses_new_dummy_session();
      if (scb == NULL)
      {
        SET_ERROR(ERR_INTERNAL_MEM);
        return;
      }
      isSessionLocalCreate = true;
    }
    /* Create a new value & map to the template */
    entryval = val_new_value();
    if (entryval == NULL) goto out;
    scb->reader = xmlReaderForMemory((const char*) outVal, xml_strlen(outVal), NULL, NULL, 0);

    res = agt_xml_consume_node(scb, &top, NCX_LAYER_RPC, NULL);

    res = agt_val_parse_nc(scb, &msg, notiobj, &top, NCX_DC_STATE, entryval);

    if (res != NO_ERR)
    {
      logDebug("NETC", "GET", "Data returning invalid!");
      goto out;
    }

    /*
     * Go through 2nd level
     */
    for (child = val_get_first_child(entryval); child != NULL; child = val_get_next_child(child))
    {
      agt_not_add_to_payload(notifMsg, child);
    }

    out:
    /* clean up and exit */
    val_free_value(entryval);

    if (scb && isSessionLocalCreate)
    {
      agt_ses_free_dummy_session(scb);
    }
    xml_msg_clean_hdr(&msg);
  }

  /**
   * Handle bind request from Mgt Objects
   */
  void MgtNetconfHandler::mgtMsgBindHandle(SAFplus::Handle from, void *pInMsg)
  {
    ClRcT rc = CL_OK;
    Mgt::Msg::MsgBind *bindData = (Mgt::Msg::MsgBind*) pInMsg;

    logDebug("NETC", "MSG", "Receive Bind MSG");
    MgtModuleOM *moduleOM = MgtRootOM::getInstance()->findModuleOM(bindData->module());
    if (moduleOM == NULL)
      {
        moduleOM = new MgtModuleOM();

        MgtRootOM::getInstance()->mModuleOMs.insert(pair<string, MgtModuleOM*>(bindData->module(), moduleOM));

        rc = MgtYuma::yumaLoadModule((const xmlChar*) bindData->module().c_str(), false);
        if (rc != CL_OK)
          {
            logError("NETC", "BIND", "Binding error.  Module [%s] Route [%s] Error [0x%x]",
                bindData->module().c_str(), bindData->route().c_str(), rc);
            return;
          }
      }

    /*
     * Build ClMgtOM from bindData
     */
    MgtOM *compBindTemp = new MgtOM(bindData);
    MgtOM *compBind = MgtRootOM::getInstance()->findOM(bindData->module(), compBindTemp->toKey());
    if (compBind == NULL)
      {
        /*
         * compBind is new, insert it into list object
         */
        moduleOM->mOMList.insert(pair<string, MgtOM *>(compBindTemp->toKey(), compBindTemp));
        MgtYuma::yumaRegisterDataSetCallback((const xmlChar*) bindData->module().c_str(), (const xmlChar*) bindData->route().c_str());
      }
    else
      {
        /*
         * Compbind is already exist, update its handle
         */
        Mgt::Msg::Handle *hdl = compBind->bindData.mutable_handle();
        hdl->set_id0(bindData->handle().id0());
        hdl->set_id1(bindData->handle().id1());
        delete (compBindTemp);
      }

    logDebug("NETC", "BIND", "Binding [%s] success", bindData->route().c_str());

#if 0
    // This is not sync message, then not to reply to src
    SAFplus::MsgServer *mgtIocInstance = &safplusMsgServer;
    mgtIocInstance->SendMsg(from, (void *) &rc, sizeof(rc), CL_MGT_MSG_TYPE);
#endif
  }

  void MgtNetconfHandler::mgtMsgNotifyHandle(SAFplus::Handle from, void *pInMsg)
  {
    ClRcT rc = CL_OK;
    logDebug("NETC", "MSG", "Receive notification message");

    if (((Mgt::Msg::MsgMgt *)pInMsg)->data_size() == 0) // No notify data
      {
        return;
      }

    Mgt::Msg::MsgBind bindData;
    bindData.ParseFromString(((Mgt::Msg::MsgMgt *) pInMsg)->bind());

    Mgt::Msg::MsgSetGet notiData;
    notiData.ParseFromString(((Mgt::Msg::MsgMgt *) pInMsg)->data(0));

    ncx_module_t *notifMod = ncx_find_module((const xmlChar *) bindData.module().c_str(), NULL);
    if (!notifMod)
    {
      rc = CL_ERR_INVALID_PARAMETER;
    }
    else
    {
      obj_template_t *notifObj = ncx_find_object(notifMod, (const xmlChar *) bindData.route().c_str());
      if (!notifObj)
      {
        rc = CL_ERR_INVALID_PARAMETER;
      }
      else
      {
        agt_not_msg_t *notifMsg = agt_not_new_notification(notifObj);
        if (!notifMsg)
        {
          rc = CL_ERR_INVALID_PARAMETER;
        }
        else
        {
          mgtNetconfNotifParserXmlData((const xmlChar*) bindData.module().c_str(), notifMsg, notifObj, (const xmlChar*) notiData.data().c_str());
          agt_not_queue_notification(notifMsg);
        }
      }
    }

    (void)rc; /*unused*/
#if 0
    // This is not sync message, then not to reply to src
    SAFplus::MsgServer *mgtIocInstance = &safplusMsgServer;
    mgtIocInstance->SendMsg(from, (void *) &rc, sizeof(rc), CL_MGT_MSG_TYPE);
#endif
  }
}
