/*
 * Copyright (C) 2002-2012 OpenClovis Solutions Inc.  All Rights Reserved.
 *
 * This file is available  under  a  commercial  license  from  the
 * copyright  holder or the GNU General Public License Version 2.0.
 *
 * The source code for  this program is not published  or otherwise
 * divested of  its trade secrets, irrespective  of  what  has been
 * deposited with the U.S. Copyright office.
 *
 * This program is distributed in the  hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * For more  information, see  the file  COPYING provided with this
 * material.
 *
 */
#include <clLogApi.hxx>
#include <clMgtApi.hxx>

#include "assert.h"

#include <boost/algorithm/string.hpp>
#include <boost/filesystem.hpp>
#include "clMgtNetconfHandler.hxx"
#include "clMgtNetconfYuma.hxx"

namespace MgtServer
{
  std::vector<std::string> MgtYuma::nodeRegistered;
  /*
   * Filter Mgt RPC data
   */
  status_t mgtFilterOutput(ses_cb_t *scb, rpc_msg_t *msg, int32 indent)
  {
    val_value_t *valRpc = (val_value_t *) msg->rpc_user2;
    val_value_t *child;
    for (child = val_get_first_child(valRpc); child != NULL; child = val_get_next_child(child))
      {
        xml_wr_full_val(scb, &msg->mhdr, child, scb->indent);
      }
    val_free_value(valRpc);
    return NO_ERR;
  }

  status_t mgtSetRpcOutput(rpc_msg_t *msg, std::string output)
  {
    status_t res = NO_ERR;
    obj_template_t *entryobj;

    xml_node_t top;
    xml_init_node(&top);
    /* create a dummy session control block */
    ses_cb_t *scbDummy = agt_ses_new_dummy_session();
    if (!scbDummy)
      {
        res = ERR_INTERNAL_MEM;
      }
    else
      {
        scbDummy->reader = xmlReaderForMemory((const char*) output.c_str(), output.length(), NULL, NULL, 0);
        res = agt_xml_consume_node(scbDummy, &top, NCX_LAYER_NONE, NULL);
        if (res == NO_ERR)
          {
            entryobj = obj_find_child(msg->rpc_method, obj_get_mod_name(msg->rpc_method), top.elname);
            val_value_t *entryval = val_new_value();
            res = agt_val_parse_nc(scbDummy, &msg->mhdr, entryobj, &top, NCX_DC_STATE, entryval);
            if (res == NO_ERR)
              {
                msg->rpc_data_type = RPC_DATA_STD;
                msg->rpc_user2 = entryval;
                msg->rpc_datacb = (void*) mgtFilterOutput;
              }
          }
        agt_ses_free_dummy_session(scbDummy);
      }
    return res;
  }

  /********************************************************************
   * FUNCTION yumaDataSetCallback
   *
   * Edit database object callback
   * Path:
   * Add object instrumentation in COMMIT phase.
   *
   * INPUTS:
   *
   *
   * RETURNS:
   *     error status
   ********************************************************************/
  status_t MgtYuma::yumaDataSetCallback(ses_cb_t *scb, rpc_msg_t *msg, agt_cbtyp_t cbtyp, op_editop_t editop, val_value_t *newval, val_value_t *curval)
  {
    status_t res = NO_ERR;
    val_value_t *errorval = (curval) ? curval : newval;
    val_value_t *val = NULL;


    logDebug("MGT", "YUMA", "Data set callback for [%s:%s] phase", agt_cbtype_name(cbtyp), op_editop_name(editop));
    if(newval!=NULL)
    {
    logDebug("MGT", "YUMA", "Data set callback for newval [%s] phase", newval->name);
    }

    switch (cbtyp)
      {
    case AGT_CB_VALIDATE:
      /* description-stmt validation here */
      break;
    case AGT_CB_APPLY:
      switch (editop)
        {
      case OP_EDITOP_LOAD:
        val = newval;
        break;
      case OP_EDITOP_MERGE:
        if (newval)
          {
            val = newval;
          }
        else if (curval)
          {
            val = curval;
          }
        else
          {
            res = SET_ERROR(ERR_INTERNAL_VAL);
          }
        break;
      case OP_EDITOP_REPLACE:
        val = newval;
        break;
      case OP_EDITOP_CREATE:
        val = newval;
        break;
      case OP_EDITOP_DELETE:
        val = curval;
        break;
      default:
        res = SET_ERROR(ERR_INTERNAL_VAL);
        break;
        }

        if (res == NO_ERR)
        {
          res = MgtNetconfHandler::getInstance()->mgtNetconfSet(editop, val);
        }
        
        if (res == NO_ERR && editop == OP_EDITOP_DELETE && val_is_virtual(val))
        {
          val_value_t *parent = val->parent;
          /*
           * Remove out of queue list 
           */
          val_remove_child(val);

          /*
           * Promote an existing value to virtual val
           */
          val_value_t *ch = val_find_child(parent, val_get_mod_name(val), val->name);
          if (ch)
          {
            ch->getcb = (void *) yumaDataGetCallback;
          }
          /*
           * Clean up
           */
          val_free_value(val);
        }

      /* database manipulation done here */
      break;
    case AGT_CB_COMMIT:
      {
        /*
         * Promote an existing value to virtual val
         */
        if(NULL == newval)
        {
          break;
        }
        val_value_t *parent = newval->parent;
        val_value_t *ch = val_find_child(parent, val_get_mod_name(newval), newval->name);
        if (ch)
        {
          ch->getcb = NULL;
        }

      }
      break;
    case AGT_CB_ROLLBACK:
      /* undo device instrumentation here */
      if(curval)
      {
         val = curval;
         if (res == NO_ERR)
         {
            res = MgtNetconfHandler::getInstance()->mgtNetconfSet(editop, val);
         }
      }
      break;
    default:
      res = SET_ERROR(ERR_INTERNAL_VAL);
      break;
      }

    if (res != NO_ERR)
      {
        agt_record_error(scb, &msg->mhdr, NCX_LAYER_CONTENT, res, NULL, (errorval) ? NCX_NT_VAL : NCX_NT_NONE, errorval, (errorval) ? NCX_NT_VAL : NCX_NT_NONE, errorval);
      }
    return res;

  }

  /********************************************************************
   * FUNCTION yumaRpcValidateCallback
   *
   * RPC validate phase
   *
   * INPUTS:
   *    see agt/agt_rpc.h
   * RETURNS:
   *    status
   *********************************************************************/
  status_t MgtYuma::yumaRpcValidate(ses_cb_t *scb, rpc_msg_t *msg, xml_node_t *methnode)
  {
    status_t res;
    val_value_t *errorval;
    const xmlChar *errorstr;

    res = NO_ERR;
    errorval = NULL;
    errorstr = NULL;
    logDebug("MGT", "YUMA", "Rpc callback for validate phase");

    /* Find registered MOs in the database */
    res = MgtNetconfHandler::getInstance()->mgtNetconfRpc(Mgt::Msg::MsgRpc_MgtRpcType::MsgRpc_MgtRpcType_CL_MGT_RPC_VALIDATE,scb,msg,methnode);
    if (res != NO_ERR)
      {
        agt_record_error(scb, &msg->mhdr, NCX_LAYER_OPERATION, res, methnode, NCX_NT_STRING, errorstr, NCX_NT_VAL, errorval);
      }

    return res;

  }

  /********************************************************************
   * FUNCTION yumaRpcInvoke
   *
   * RPC invocation phase
   * All constraints have passed at this point.
   * Call device instrumentation code in this function.
   *
   * INPUTS:
   *     see agt/agt_rpc.h for details
   *
   * RETURNS:
   *     error status
   ********************************************************************/
  status_t MgtYuma::yumaRpcInvoke(ses_cb_t *scb, rpc_msg_t *msg, xml_node_t *methnode)
  {
    status_t res;
    val_value_t *errorval;
    const xmlChar *errorstr;
    logDebug("MGT", "YUMA", "Rpc callback for Invoke phase");
    res = NO_ERR;
    errorval = NULL;
    errorstr = NULL;
    /* Find registered MOs in the database */
    res = MgtNetconfHandler::getInstance()->mgtNetconfRpc(Mgt::Msg::MsgRpc_MgtRpcType::MsgRpc_MgtRpcType_CL_MGT_RPC_INVOKE, scb,msg,methnode);
    /* if error: set the res, errorstr, and errorval parms */
    if (res != NO_ERR)
    {
      agt_record_error(scb, &msg->mhdr, NCX_LAYER_OPERATION, res, methnode, NCX_NT_STRING, errorstr, NCX_NT_VAL, errorval);
    }
    return res;
  }

  /********************************************************************
   * FUNCTION clMgtRpcPostReply
   *
   * RPC post-reply phase
   *
   * INPUTS:
   *    see agt/agt_rpc.h
   * RETURNS:
   *    status
   *********************************************************************/
  status_t MgtYuma::yumaRpcPostReply(ses_cb_t *scb, rpc_msg_t *msg, xml_node_t *methnode)
  {
    status_t res;
    val_value_t *errorval;
    const xmlChar *errorstr;
    logDebug("MGT", "YUMA", "Rpc callback for PostReply phase");
    res = NO_ERR;
    errorval = NULL;
    errorstr = NULL;
    res = MgtNetconfHandler::getInstance()->mgtNetconfRpc(Mgt::Msg::MsgRpc_MgtRpcType::MsgRpc_MgtRpcType_CL_MGT_RPC_POSTREPLY, scb,msg,methnode);
    /* if error: set the res, errorstr, and errorval parms */
    if (res != NO_ERR)
    {
      agt_record_error(scb, &msg->mhdr, NCX_LAYER_OPERATION, res, methnode, NCX_NT_STRING, errorstr, NCX_NT_VAL, errorval);
    }
    return res;
  } /* dummy_post_reply */

  status_t MgtYuma::yumaListValidate(ses_cb_t *scb, rpc_msg_t *msg, xml_node_t *methnode)
  {
    status_t res;
    val_value_t *errorval;
    const xmlChar *errorstr;

    res = NO_ERR;
    errorval = NULL;
    errorstr = NULL;

    logDebug("MGT", "YUMA", "[%s:%s] List Validate",(const char*) methnode->module,(const char*) methnode->qname);

    /* if error: set the res, errorstr, and errorval parms */
    if (res != NO_ERR)
      {
        agt_record_error(scb, &msg->mhdr, NCX_LAYER_OPERATION, res, methnode, NCX_NT_STRING, errorstr, NCX_NT_VAL, errorval);
      }

    return res;
  }


  status_t MgtYuma::yumaListInvoke(ses_cb_t *scb, rpc_msg_t *msg, xml_node_t *methnode)
  {
    status_t res;
    val_value_t *errorval;
    const xmlChar *errorstr;

    res = NO_ERR;
    errorval = NULL;
    errorstr = NULL;
    logDebug("MGT", "YUMA", "[%s:%s] List Invoke",(const char*) methnode->module,(const char*) methnode->qname);

    /* if error: set the res, errorstr, and errorval parms */
    if (res != NO_ERR)
      {
        agt_record_error(scb, &msg->mhdr, NCX_LAYER_OPERATION, res, methnode, NCX_NT_STRING, errorstr, NCX_NT_VAL, errorval);
      }

    return res;

  }

  status_t MgtYuma::yumaListPostReply(ses_cb_t *scb, rpc_msg_t *msg, xml_node_t *methnode)
  {
    status_t res;
    val_value_t *errorval;
    const xmlChar *errorstr;

    res = NO_ERR;
    errorval = NULL;
    errorstr = NULL;

    logDebug("MGT", "YUMA", "[%s:%s] List Post Reply",(const char*) methnode->module,(const char*) methnode->qname);

    /* if error: set the res, errorstr, and errorval parms */
    if (res != NO_ERR)
      {
        agt_record_error(scb, &msg->mhdr, NCX_LAYER_OPERATION, res, methnode, NCX_NT_STRING, errorstr, NCX_NT_VAL, errorval);
      }

    return res;
  } 

  status_t MgtYuma::yumaLeafValidate(ses_cb_t *scb, rpc_msg_t *msg, xml_node_t *methnode)
  {
    status_t res;
    val_value_t *errorval;
    const xmlChar *errorstr;

    res = NO_ERR;
    errorval = NULL;
    errorstr = NULL;

    logDebug("MGT", "YUMA", "[%s:%s] Leaf Validate",(const char*) methnode->module,(const char*) methnode->qname);

    /* if error: set the res, errorstr, and errorval parms */
    if (res != NO_ERR)
      {
        agt_record_error(scb, &msg->mhdr, NCX_LAYER_OPERATION, res, methnode, NCX_NT_STRING, errorstr, NCX_NT_VAL, errorval);
      }

    return res;
  }


  status_t MgtYuma::yumaLeafInvoke(ses_cb_t *scb, rpc_msg_t *msg, xml_node_t *methnode)
  {
    status_t res;
    val_value_t *errorval;
    const xmlChar *errorstr;

    res = NO_ERR;
    errorval = NULL;
    errorstr = NULL;
    logDebug("MGT", "YUMA", "[%s:%s] Leaf Invoke",(const char*) methnode->module,(const char*) methnode->qname);

    /* if error: set the res, errorstr, and errorval parms */
    if (res != NO_ERR)
      {
        agt_record_error(scb, &msg->mhdr, NCX_LAYER_OPERATION, res, methnode, NCX_NT_STRING, errorstr, NCX_NT_VAL, errorval);
      }

    return res;

  }

  status_t MgtYuma::yumaLeafPostReply(ses_cb_t *scb, rpc_msg_t *msg, xml_node_t *methnode)
  {
    status_t res;
    val_value_t *errorval;
    const xmlChar *errorstr;

    res = NO_ERR;
    errorval = NULL;
    errorstr = NULL;

    logDebug("MGT", "YUMA", "[%s:%s] Leaf Post Reply",(const char*) methnode->module,(const char*) methnode->qname);

    /* if error: set the res, errorstr, and errorval parms */
    if (res != NO_ERR)
      {
        agt_record_error(scb, &msg->mhdr, NCX_LAYER_OPERATION, res, methnode, NCX_NT_STRING, errorstr, NCX_NT_VAL, errorval);
      }

    return res;
  } 

  void MgtYuma::register_callbacks(ncx_module_t *mod)
  {
    const xmlChar *modname = ncx_get_modname(mod);
    const xmlChar *modrev = ncx_get_modversion (mod);
    status_t res = NO_ERR;

    obj_template_t *obj;

    /* initialize any RPC methods */
    for (obj = (obj_template_t *) dlq_firstEntry(&mod->datadefQ); obj != NULL; obj = (obj_template_t *) dlq_nextEntry(obj))
      {
        if (obj_is_abstract(obj))
          {
            continue;
          }

        switch (obj->objtype)
          {
          case OBJ_TYP_CHOICE:
          case OBJ_TYP_CASE:       /* last named database object */
          case OBJ_TYP_USES:
          case OBJ_TYP_REFINE:            /* child of uses only */
          case OBJ_TYP_AUGMENT:
          case OBJ_TYP_NONE:
          case OBJ_TYP_ANYXML:
            break;

          case OBJ_TYP_LEAF:
          case OBJ_TYP_LEAF_LIST:
          case OBJ_TYP_CONTAINER:
          case OBJ_TYP_LIST:
            {
             /*
              * Using heap memory
              */
              xmlChar *buff;
              res = obj_gen_object_id(obj, &buff);
              assert(res == NO_ERR);

              logDebug("MGT", "YUMA", "Registering [%s] callbacks for xpath: [%s]", (const char *)obj_get_typestr(obj), (char*) buff);

              /* register callback functions */
              res = agt_cb_register_callback(modname, buff, modrev, MgtYuma::yumaDataSetCallback);
              if (res != NO_ERR)
              {
                free(buff);
              }
              // ERR_NCX_DUP_ENTRY
              //assert(res == NO_ERR);
#if 0
            res = agt_rpc_register_method(modname, obj_get_name(obj), AGT_RPC_PH_VALIDATE, MgtYuma::yumaLeafValidate);
            assert(res == NO_ERR);
            res = agt_rpc_register_method(modname, obj_get_name(obj), AGT_RPC_PH_INVOKE, MgtYuma::yumaLeafInvoke);
            assert(res == NO_ERR);
            res = agt_rpc_register_method(modname, obj_get_name(obj), AGT_RPC_PH_POST_REPLY, MgtYuma::yumaLeafPostReply);
            assert(res == NO_ERR);
#endif
            } break;


#if 0
            logDebug("MGT", "YUMA", "Registering container/list callbacks for xpath: [%s]", obj_get_name(obj));
            /* register callback functions */
            res = agt_rpc_register_method(modname, obj_get_name(obj), AGT_RPC_PH_VALIDATE, MgtYuma::yumaListValidate);
            assert(res == NO_ERR);
            res = agt_rpc_register_method(modname, obj_get_name(obj), AGT_RPC_PH_INVOKE, MgtYuma::yumaListInvoke);
            assert(res == NO_ERR);
            res = agt_rpc_register_method(modname, obj_get_name(obj), AGT_RPC_PH_POST_REPLY, MgtYuma::yumaListPostReply);
            assert(res == NO_ERR);
#endif
            break;

          case OBJ_TYP_RPC:
            logDebug("MGT", "YUMA", "Registering RPC callbacks for xpath: [%s]", obj_get_name(obj));
            /* register callback functions */
            res = agt_rpc_register_method(modname, obj_get_name(obj), AGT_RPC_PH_VALIDATE, MgtYuma::yumaRpcValidate);
            assert(res == NO_ERR);
            res = agt_rpc_register_method(modname, obj_get_name(obj), AGT_RPC_PH_INVOKE, MgtYuma::yumaRpcInvoke);
            assert(res == NO_ERR);
            res = agt_rpc_register_method(modname, obj_get_name(obj), AGT_RPC_PH_POST_REPLY, MgtYuma::yumaRpcPostReply);
            assert(res == NO_ERR);
            break;
          case OBJ_TYP_RPCIO:
            break;
          case OBJ_TYP_NOTIF:
            break;
          }


      }
  } /* register_rpc */


  void MgtYuma::register_rpc(ncx_module_t *mod, dlq_hdr_t *objnameQ)
  {
    const xmlChar *modname = ncx_get_modname(mod);

    status_t res = NO_ERR;

    obj_template_t *obj;

    /* initialize any RPC methods */
    for (obj = (obj_template_t *) dlq_firstEntry(&mod->datadefQ); obj != NULL; obj = (obj_template_t *) dlq_nextEntry(obj))
      {

        if (!obj_is_rpc(obj) || obj_is_abstract(obj))
          {
            continue;
          }

        /* register validate function */
        logDebug("MGT", "YUMA", "Register RPC phase validate for xpath: [%s]", obj_get_name(obj));
        res = agt_rpc_register_method(modname, obj_get_name(obj), AGT_RPC_PH_VALIDATE, MgtYuma::yumaRpcValidate);
        assert(res == NO_ERR);

        /* register invoke function */
        logDebug("MGT", "YUMA", "Register RPC phase invoke for xpath: [%s]", obj_get_name(obj));
        res = agt_rpc_register_method(modname, obj_get_name(obj), AGT_RPC_PH_INVOKE, MgtYuma::yumaRpcInvoke);
        assert(res == NO_ERR);

        /* register post-reply function */
        logDebug("MGT", "YUMA", "Register RPC phase post-reply for xpath: [%s]", obj_get_name(obj));
        res = agt_rpc_register_method(modname, obj_get_name(obj), AGT_RPC_PH_POST_REPLY, MgtYuma::yumaRpcPostReply);
        assert(res == NO_ERR);
      }
  } /* register_rpc */

  void MgtYuma::unregister_rpc(ncx_module_t *mod)
  {
    const xmlChar *modname = ncx_get_modname(mod);

    /* cleanup any RPC methods */
    obj_template_t *obj;
    for (obj = (obj_template_t *) dlq_firstEntry(&mod->datadefQ); obj != NULL; obj = (obj_template_t *) dlq_nextEntry(obj))
      {

        if (!obj_is_rpc(obj) || obj_is_abstract(obj))
          {
            continue;
          }

        logDebug("MGT", "YUMA", "Unregister RPC for xpath: [%s]", obj_get_name(obj));

        /* unregister any RPCs methods */
        agt_rpc_unregister_method(modname, obj_get_name(obj));
      }

  } /* unregister_rpc */

  status_t MgtYuma::yumaDataGetCallback(ses_cb_t *scb, getcb_mode_t cbmode, val_value_t *virval, val_value_t *dstval)
  {
    status_t rc = NO_ERR;

    xmlChar * moduleName;
    moduleName = (xmlChar *) obj_get_mod_name(dstval->obj);

    logDebug("MGT", "YUMA", "Get value of [%s] for module [%s]", dstval->name,moduleName);

    /* Always refresh value */
    //virval->cachetime = 0;

    /* Get data for the module */
    MgtNetconfHandler::getInstance()->mgtNetconfGet((const xmlChar *) moduleName, virval, dstval);
    return rc;
  }

  /*
   * Any request load module from IMM, loadSIL always false (we already have dummy SIL for them)
   */
  status_t MgtYuma::yumaLoadModule(const xmlChar* moduleName, bool loadSIL)
  {
    status_t rc = NO_ERR;
    agt_profile_t *agt_profile;
    ncx_module_t *mod;
    val_value_t *modval, *newval;
    boolean errdone, sil_loaded;
    xmlChar * moduri;
    newval = NULL;
    mod = NULL;
    errdone = FALSE;
    sil_loaded = FALSE;
    agt_profile = agt_get_profile();
    boolean added = FALSE;
    obj_template_t *obj;

    /*
     * Still not test for revision for this network.yang
     */
    agt_profile = agt_get_profile();

    /* mandatory module name */
    modval = val_new_value();

    /* check for any version of this module already loaded */
    mod = ncx_find_module(moduleName, NULL);

    if (mod == NULL)
      {
        /*
         * Exptected all deviations loaded before doing this testing
         */

        /* module not loaded already
         * load all the deviations first
         */

        if (rc == NO_ERR)
          {
            if (loadSIL)
              {
                rc = agt_load_sil_code(moduleName, NULL, TRUE);
                if (rc == ERR_NCX_SKIPPED)
                  {
                    rc = ncxmod_load_module(moduleName, NULL, &agt_profile->agt_savedevQ, &mod);
                  }
                else if (rc == NO_ERR)
                  {
                    sil_loaded = TRUE;
                  }
              }
            else
              {
                rc = ncxmod_load_module(moduleName, NULL, &agt_profile->agt_savedevQ, &mod);
              }

            /* pre-check the module; it should be found if status == NO_ERR */
            if (rc == NO_ERR)
              {
                mod = ncx_find_module(moduleName, NULL);
                if (mod == NULL)
                  {
                    rc = SET_ERROR(ERR_INTERNAL_VAL);
                  }
              }
          }
      }

    /* generate the return value */
    if (rc == NO_ERR && mod != NULL)
      {
        newval = val_make_string(val_get_nsid(modval), NCX_EL_MOD_REVISION, (mod->version) ? mod->version : EMPTY_STRING);
        if (newval == NULL)
          {
            rc = ERR_INTERNAL_MEM;
            errdone = TRUE;
          }
      }

    if (rc == NO_ERR && mod && !sil_loaded)
      {
        /* Register dummy callback if sil not found <- IMM */
        MgtYuma::register_callbacks(mod);
      }

    if (rc == NO_ERR && mod)
      {
        /* prune all the obsolete objects */
        ncx_delete_mod_obsolete_objects(mod);
        /* add the module commit tests for this module */
        rc = agt_val_add_module_commit_tests(mod);
      }

    if (rc == NO_ERR && mod)
      {
        /* add the <schema> node in netconf-state module */
        rc = agt_state_add_module_schema(mod);
      }

    if (rc == NO_ERR && mod)
      {
        /* add the <capability> node in <hello> template */
        rc = agt_cap_add_module(mod);
      }

    if (rc == NO_ERR && mod)
      {
        /* send the capability change notification */
        moduri = cap_make_moduri(mod);
        if (!moduri)
          {
            rc = ERR_INTERNAL_MEM;
          }
        else
          {
            /* agt_sys_send_sysCapabilityChange(scb, TRUE, moduri) */;
            m__free(moduri);
          }
      }

    if (rc != NO_ERR)
      {
        if (!errdone)
          {
            goto out;
          }
      }

    if (rc == NO_ERR && mod)
      {
        /*
         * Register top node here to handle data routing (SIL Phase 2 init)
         */
        for (obj = (obj_template_t *) dlq_firstEntry(&mod->datadefQ); obj != NULL; obj = (obj_template_t *) dlq_nextEntry(obj))
        {
          if (obj_has_name(obj) && obj_is_data_db(obj) && !obj_is_cli(obj) && !obj_is_abstract(obj) && obj_is_top(obj))
          {
            val_value_t *topval = agt_add_top_node_if_missing(mod, obj_get_name(obj), &added, &rc);
            if (added)
            {
              logDebug("MGT", "YUMA", "Registering 'get' callback for node: [%s:%s]", moduleName, obj_get_name(obj));
              topval->getcb = (void*) MgtYuma::yumaDataGetCallback;
              topval->cachetime = 0;

              /* Cache virtual value into cfg->roof if it is already binding */
              MgtNetconfHandler::getInstance()->mgtNetconfGet((const xmlChar *) moduleName, topval, topval);
              nodeRegistered.push_back((char *) obj_get_name(obj));
            }
          }
        }
      }
    if(agt_profile->agt_targ != NCX_AGT_TARG_RUNNING) rc = cfg_fill_candidate_from_running();
    out:
    if (newval)
      val_free_value(newval);

    if (modval)
      val_free_value(modval);

    return rc;
  }

  std::string MgtYuma::yumaXPathParse(xmlChar *valXPath)
  {
    status_t res = NO_ERR;
    std::string objXPath = "";
    bool isAttr = false;
    bool isAttrValue = false;
    const tk_token_t *tk;

    if (!valXPath)
      return objXPath;

    tk_chain_t *tkc = tk_tokenize_xpath_string(NULL, valXPath, 1, 1, &res);
    if (!tkc || res != NO_ERR)
      return objXPath;

    for (tk = (tk_token_t *) dlq_firstEntry(&tkc->tkQ); tk != NULL; tk = (tk_token_t *) dlq_nextEntry(tk))
      {
        if (tk->typ == TK_TT_LBRACK)
        {
          isAttr = true;
        }

        if (!isAttr && !isAttrValue)
          objXPath.append((char *) tk->val);
        else if (isAttrValue)
        {
          objXPath.append("/").append((char *) tk->val);
          isAttrValue = false;
        }

        if (isAttr && tk->typ == TK_TT_EQUAL)
        {
          isAttrValue = true;
        }

        if (tk->typ == TK_TT_RBRACK)
        {
          isAttr = false;
          isAttrValue = false;
        }
      }

    tk_free_chain(tkc);
    return objXPath;
  }

  status_t MgtYuma::yumaRegisterDataSetCallback(const xmlChar* moduleName, const xmlChar* route)
  {
    status_t rc = NO_ERR;

    std::string xpath = MgtYuma::yumaXPathParse((xmlChar*) route);
    if (xpath.empty())
      {
        rc = ERR_FORM_INPUT;
        return rc;
      }

    xmlChar *xmlDefPath = xml_strdup((const xmlChar *)xpath.c_str());

    /* Register callback edit data for route.  objXPath memory is given to the routine. */
    rc = agt_cb_register_callback(moduleName, xmlDefPath, NULL, yumaDataSetCallback);

    if (rc != NO_ERR)
    {
      free(xmlDefPath);
    }

    return rc;
  }

  status_t MgtYuma::yumaUnloadModule(const xmlChar* moduleName)
  {
    status_t rc = NO_ERR;
    ncx_module_t *mod;

    /* check for any version of this module already loaded */
    mod = ncx_find_module(moduleName, NULL);

    if (mod != NULL)
      {
        /*
         * Unregister callbacks and RPCs
         */
        MgtYuma::unregister_rpc(mod);
      }

    return rc;
  }

  /*
   * Function to check a load a yang module
   */
  ClRcT MgtYuma::yumaCheckAndLoadModule(std::vector<std::string> &ckptNew, const std::vector<std::string> &paths, std::map<std::string,bool> *mapLoadFile)
  {
    ClRcT rc = CL_OK;
    std::vector<std::string>::iterator iterCkpt;
    std::map<std::string, bool>::iterator itLoadFile;
    /*
     * flag is TRUE if module is not existing in tree modules.
     */
    bool lookupNeeded = false;
    for (iterCkpt = ckptNew.begin(); iterCkpt != ckptNew.end(); ++iterCkpt)
    {
      /*
       * Compare checkpoint with current to get new changed
       */
      if (std::find(MgtYuma::nodeRegistered.begin(), MgtYuma::nodeRegistered.end(), *iterCkpt) == MgtYuma::nodeRegistered.end())
      {
        obj_template_t *obj = ncx_find_any_object((const xmlChar*)(*iterCkpt).c_str());
        if (obj != NULL)
        {
          rc = MgtServer::MgtYuma::yumaLoadModule(obj_get_mod_name(obj), false);
        }
        else
        {
          // Need to get new file changed and parse first
          lookupNeeded = true;
          break;
        }
      }
    }

    /*
     * Live update YANG modules, lookup new YANG file and do parse
     */
    if (lookupNeeded && !paths.empty())
    {
      yang_pcb_t              *pcb;
      pcb = yang_new_pcb();
      assert (pcb != NULL);
      bool checkFailed;

      for (iterCkpt = ckptNew.begin(); iterCkpt != ckptNew.end(); ++iterCkpt)
      {
        checkFailed = true;
        for (std::vector<std::string>::const_iterator iterDir = paths.begin(); iterDir != paths.end(); ++iterDir)
        {
          boost::filesystem::path yangfile(*iterDir + "/" + *iterCkpt + ".yang");
          logDebug("MOD", "CHK", "Searching YANG file [%s] ...", yangfile.string().c_str());
          if (boost::filesystem::exists(yangfile))
          {
            itLoadFile = mapLoadFile->find(yangfile.string());
            if (itLoadFile == mapLoadFile->end())
            {
              logDebug("MOD", "LOAD", "Loading YANG file [%s]", yangfile.string().c_str());
              status_t ret = yang_parse_from_filespec((const xmlChar*) yangfile.string().c_str(), pcb, YANG_PT_TOP, true);
              if (!ret)
              {
                /*
                 * Register callback for module's top node
                 */
                rc = MgtServer::MgtYuma::yumaLoadModule((const xmlChar*) yangfile.stem().string().c_str(), false);
                if (rc == CL_OK)
                {
                  checkFailed = false;
                } else logNotice("MOD", "INIT", "Register callback for module [%s] error [0x%x]", yangfile.string().c_str(), rc);
              } else logNotice("MOD", "LOAD", "Parse YANG file [%s] error [0x%x]", yangfile.string().c_str(), ret);
              /*
               * Marked as failed to re-check at next polling
               */
              mapLoadFile->insert(std::pair<std::string, bool>(*iterCkpt, checkFailed));
            }
          }
        }
      }
      yang_free_pcb(pcb);
    }

    return rc;
  }
}

//  xmlNodePtr MgtYuma::yumaRpcParseAttributeNode(const xmlNodePtr node, const xmlChar *keyword)
//  {
//    xmlNodePtr cur = node;
//    cur = cur->xmlChildrenNode;
//    while (cur != NULL)
//      {
//        if (!xmlStrcmp(cur->name, keyword))
//          {
//            return cur;
//          }
//        cur = cur->next;
//      }
//
//    return NULL;
//  }
//
////  ses_cb_t *scb, rpc_msg_t *msg, xml_node_t *methnode
//  std::string MgtYuma::yumaRpcParseAttribute(ses_cb_t *scb, xml_node_t *methnode)
//  {
//    xmlChar *xml;
//    xmlDocPtr doc;
//    xmlNodePtr cur, node;
//    std::string data = "";
//
//    std::vector<std::string> strs;
//    std::string buff(reinterpret_cast<const char*>(scb->readbuff));
//    boost::split(strs, buff, boost::is_any_of("]]"));
//
//    xml = (xmlChar*) strs[0].c_str();
//    doc = xmlParseDoc(xml);
//    cur = xmlDocGetRootElement(doc);
//
//    if (cur == nullptr)
//      {
//        xmlFreeDoc(doc);
//      }
//    else
//      {
//        node = yumaRpcParseAttributeNode(cur, methnode->elname);
//        node = node->xmlChildrenNode;
//        while (node != NULL)
//          {
//            if (node->type == XML_ELEMENT_NODE)
//              {
//                char str[30];
//                sprintf(str, "%s=%s,", node->name, xmlNodeListGetString(doc, node->xmlChildrenNode, 1));
//                data.append(str);
//              }
//            node = node->next;
//          }
//      }
//
//    return data;
//  }
//}
