#include <locale>

#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/replace.hpp>
#include <boost/filesystem.hpp>

// Snmp include
#include <clMgtRoot.hxx>
#include <clMgtCommon.hxx>
#include <clMsgPortsAndTypes.hxx>
#include <clSafplusMsgServer.hxx>
#include "clMgtSnmpHandler.hxx"

static oid snmptrap_oid[] = { 1, 3, 6, 1, 6, 3, 1, 1, 4, 1, 0 };
extern bool keep_snmpAgent_running;

// namespace
using namespace SAFplus;

namespace MgtServer
{
  /* 
   * This token is to replace when convert YANG to SMIv2 that conflict definied SMIv2's keyword.
   * So, when calling mgt function, we have to revert back to original name - without 'y2mib'
   */
  const std::string y2mibToken = "y2mib";

  /********************************
   * Global variables
   ********************************/
  SAFplus::Mutex sendReplySnmpMutex;
  MgtSnmpHandler* MgtSnmpHandler::instance = 0;
  std::list<std::string *> dataRows;
  std::map<MgtSnmpScalarInfo *, std::string *> dataMap;

  SAFplus::Mutex gSnmpReqGetLock;
  SAFplus::Mutex gSnmpReqSetLock;

  /* convert string "1.2.3" to snmp oid */
  void str2oid(const string &oid_in, ClUint32T *outOid, ClUint32T *outLen)
  {
    std::vector<string> tmpList;

    char *splitStr = strdup(oid_in.c_str());
    char *saveptr = NULL;
    char *token = strtok_r(splitStr, ".", &saveptr);
    while (token != NULL)
    {
      tmpList.push_back(token);
      token = strtok_r(NULL, ".", &saveptr);
    }
    free(splitStr);

    if (outLen)
    {
      *outLen = tmpList.size();
    }

    if (outOid)
    {
      /*
       * Copy out snmp oid
       */
      for (ClUint32T i = 0; i < tmpList.size(); i++)
      {
        outOid[i] = atoi(tmpList[i].c_str());
      }
    }
  }

  /* convert snmp oid 1.2.3 to string "1.2.3" */
  void oid2str(string &oidStr, const ClUint32T *oid, const ClUint32T len)
  {
    stringstream sstream;
    ClUint32T i = 0;

    if (!oid || len < 1)
    {
      oidStr.assign("");
      return;
    }

    sstream << oid[0];
    for (i = 1; i < len; i++)
    {
      sstream << ".";
      sstream << oid[i];
    }
    oidStr.assign(sstream.str());
  }
  
  void oid2str(string &oidStr, const oid *oid, const ClUint32T len)
  {
    stringstream sstream;
    ClUint32T i = 0;

    if (!oid || len < 1)
    {
      oidStr.assign("");
      return;
    }

    sstream << oid[0];
    for (i = 1; i < len; i++)
    {
      sstream << ".";
      sstream << oid[i];
    }
    oidStr.assign(sstream.str());
  }

  /*
   * Params:
   * inOid: Array of ClUint32T
   * outOid: Array of ClUint32T
   *
   */
  void copy2oid(const ClUint32T *inOid, ClUint32T inLen, ClUint32T *outOid, ClUint32T *outLen)
  {
    if (outLen)
    {
      *outLen = inLen;
    }
    memcpy(outOid, inOid, inLen*sizeof(ClUint32T));
  }

  /*
   * Convert ClUint32T to snmp oid (unsigned long int)
   */
  void int2oid(const ClUint32T *inOid, ClUint32T inLen, oid *outOid)
  {
    assert(outOid != NULL);
    /*
     * Copy out snmp oid
     */
    for (ClUint32T i = 0; i < inLen; i++)
    {
      outOid[i] = inOid[i];
    }
  }

  void snmpDataReset(void)
  {
    logDebug("SNP", "OPE", "Reseting the static data structure.");
    SAFplus::ScopedLock<SAFplus::Mutex> lock(gSnmpReqGetLock);
    for (std::list<std::string*>::iterator it=dataRows.begin(); it!=dataRows.end(); ++it)
    {
      delete *it;
    }
    dataRows.clear();
  }

  /**
   * Process the get request. This is called in the mode end callback.
   */
  int snmpProcessRequest(netsnmp_request_info *requests)
  {
    ClRcT rc = CL_SNMP_ERR_NOERROR;
    {
      SAFplus::ScopedLock<SAFplus::Mutex> lock(gSnmpReqSetLock);
      map<MgtSnmpScalarInfo *, std::string *>::iterator it = dataMap.begin();
      while (it != dataMap.end())
      {
        MgtSnmpScalarInfo *s = static_cast<MgtSnmpScalarInfo *>((*it).first);
        string *scalarData = static_cast<string *>((*it).second);

#if 0
        //Find and send request to OM
        MgtSnmpOM *mgtOM = MgtSnmpHandler::getInstance()->findOM(s->module, s->oid, s->len);
        if (mgtOM)
        {
          //TODO: Should combine into an mgtOM for multi requests to reduce bandwidth
          rc = MgtSnmpHandler::getInstance()->mgtSnmpDataSet(mgtOM, *scalarData);
        }
#endif
        //TODO: should work?
        rc = SAFplus::mgtSet(s->xpath, *scalarData);

        delete scalarData;
        it++;
      }

      //Reset data
      dataMap.clear();

    }
    snmpDataReset();
    return rc;
  }

  void smiErrorCallback(char *path, int line, int severity, char *msg, char *tag)
  {
    logDebug("SMN", "CALLBACK", "SMI Error [%s]", msg);
  }

  /*
   * Fill xml value into scalar column
   */
  ClRcT snmpScalarSetXmlData(netsnmp_variable_list *vptr, const string &dataOMXml, const string &scalarName, ClUint32T columnType, bool isIdx = false)
  {
    ClRcT rc = CL_OK;
    int ret;
    xmlTextReaderPtr reader;
    const xmlChar *nodeName, *nodeValue;

    reader = xmlReaderForMemory(dataOMXml.c_str(), dataOMXml.length(), NULL, NULL, 0);
    if (reader != NULL)
    {
      ret = xmlTextReaderRead(reader);
      bool isValueFound = false;
      while (ret && !isValueFound)
      {
        nodeName = xmlTextReaderConstName(reader);
        if (!strcmp(scalarName.c_str(), (const ClCharT*) nodeName))
        {
          ret = xmlTextReaderRead(reader);
          if (ret)
          {
            nodeValue = xmlTextReaderConstValue(reader);
            stringstream valStr;
            valStr << nodeValue;

            isValueFound = true;
            switch (columnType)
            {
              case ASN_INTEGER:
              {
                ClInt32T scalarVal = 0;
                valStr >> scalarVal;
                if (isIdx)
                {
                  snmp_set_var_value(vptr, (u_char *) &scalarVal, sizeof(scalarVal));
                }
                else
                {
                  snmp_set_var_typed_value(vptr, columnType, (u_char *) &scalarVal, sizeof(scalarVal));
                }
                break;
              }
              case ASN_COUNTER:
              case ASN_TIMETICKS:
              case ASN_UNSIGNED: /* also ASN_GAUGE */
              {
                ClUint32T scalarVal = 0;
                valStr >> scalarVal;
                if (isIdx)
                {
                  snmp_set_var_value(vptr, (u_char *) &scalarVal, sizeof(scalarVal));
                }
                else
                {
                  snmp_set_var_typed_value(vptr, columnType, (u_char *) &scalarVal, sizeof(scalarVal));
                }
                break;
              }
              case ASN_OCTET_STR:
              {
                ClCharT scalarVal[CL_MAX_NAME_LENGTH] = { 0 };
                if (nodeValue)
                  strncat(scalarVal, (const ClCharT*) nodeValue, strlen((const ClCharT*) nodeValue));

                if (isIdx)
                {
                  snmp_set_var_value(vptr, (u_char *) &scalarVal, strlen(scalarVal));
                }
                else
                {
                  snmp_set_var_typed_value(vptr, columnType, (u_char *) &scalarVal, strlen(scalarVal));
                }
                break;
              }
              case ASN_OBJECT_ID:
              {
                ClUint32T scalarVal[CL_SNMP_MAX_OID_LEN] = { 0 };
                for (ClUint32T i = 0; i < strlen((const ClCharT*) nodeValue); i++)
                {
                  scalarVal[i] = ((const ClCharT*) nodeValue)[i];
                }
                if (isIdx)
                {
                  snmp_set_var_value(vptr, (u_char *) &scalarVal, sizeof(scalarVal));
                }
                else
                {
                  snmp_set_var_typed_value(vptr, columnType, (u_char *) &scalarVal, sizeof(scalarVal));
                }
                break;
              }
              case ASN_COUNTER64:
              {
                ClUint64T scalarVal = 0;
                valStr >> scalarVal;
                if (isIdx)
                {
                  snmp_set_var_value(vptr, (u_char *) &scalarVal, sizeof(scalarVal));
                }
                else
                {
                  snmp_set_var_typed_value(vptr, columnType, (u_char *) &scalarVal, sizeof(scalarVal));
                }
                break;
              }
              case ASN_IPADDRESS:
              {
                ClUint8T scalarVal[4] = { 0 };
                int ip[4];
                char ch;
                valStr >> ip[0] >> ch >> ip[1] >> ch >> ip[2] >> ch >> ip[3];
                for (int i = 0; i < 4; i++)
                {
                  scalarVal[i] = ip[i];
                }
                if (isIdx)
                {
                  snmp_set_var_value(vptr, (u_char *) &scalarVal, sizeof(scalarVal));
                }
                else
                {
                  snmp_set_var_typed_value(vptr, columnType, (u_char *) &scalarVal, sizeof(scalarVal));
                }
                break;
              }
              case ASN_OPAQUE:
              {
                ClUint8T scalarVal = (ClUint8T) atoi((const ClCharT*) nodeValue);
                if (isIdx)
                {
                  snmp_set_var_value(vptr, (u_char *) &scalarVal, sizeof(scalarVal));
                }
                else
                {
                  snmp_set_var_typed_value(vptr, columnType, (u_char *) &scalarVal, sizeof(scalarVal));
                }
                break;
              }
              default:
              {
                logDebug("SNM", "SUB", "Invalid data type [%d] for scalar [%s]", columnType, scalarName.c_str());
                break;
              }
            }
          }
          break;
        }
        ret = xmlTextReaderRead(reader);
      }
      xmlFreeTextReader(reader);

      if (!isValueFound)
      {
        rc = CL_SNMP_ERR_WRONGVALUE;
      }
    }
    else
    {
      rc = CL_SNMP_ERR_RESOURCEUNAVAILABLE;
    }

    return rc;
  }

  ClRcT snmpScalarGetXmlData(netsnmp_variable_list *vptr, const string &scalarName, ClUint32T columnType, string &xmlOut)
  {
    ClRcT rc = CL_OK;
    stringstream valStr;

    switch (columnType)
    {
      case ASN_INTEGER:
      case ASN_OPAQUE:
      case ASN_COUNTER:
      case ASN_TIMETICKS:
      case ASN_UNSIGNED: /* also ASN_GAUGE */
      {
        valStr << *(vptr->val.integer);
        break;
      }
      case ASN_OCTET_STR:
      {
        ClCharT scalarVal[CL_MAX_NAME_LENGTH] = { 0 };
        strncat(scalarVal, (const ClCharT*) vptr->val.string, vptr->val_len);

        valStr << scalarVal;
        break;
      }
      case ASN_OBJECT_ID:
      {
        for (ClUint32T i = 0; i < vptr->val_len; i++)
        {
          valStr.put('.');
          valStr.put(((const ClCharT*) vptr->val.objid)[i]);
        }
        break;
      }
      case ASN_COUNTER64:
      {
        valStr << *(ClUint64T*)(vptr->val.counter64);
        break;
      }
      case ASN_IPADDRESS:
      {
        valStr << ((ClUint8T*) vptr->val.integer)[0] << "." << ((ClUint8T*) vptr->val.integer)[1] << "."
                << ((ClUint8T*) vptr->val.integer)[2] << "." << ((ClUint8T*) vptr->val.integer)[3];
        break;
      }
      default:
      {
        logDebug("SNM", "SUB", "Invalid data type [%d] for scalar [%s]", columnType, scalarName.c_str());
        break;
      }
    }
    xmlOut.append("<").append(scalarName).append(">");
    xmlOut.append(valStr.str());
    xmlOut.append("</").append(scalarName).append(">");
    return rc;
  }

  /*
   * Fill xml value into scalar columns
   */
  ClRcT snmpNotificationGetValue(MgtServer::MgtSnmpNotificationInfo *snmpNotifInfo, netsnmp_variable_list *notification_vars,
                                 const ClUint8T *dataObjectInstance)
  {
    ClRcT rc = CL_OK;
    int ret;
    xmlTextReaderPtr reader;
    const xmlChar *nodeName, *nodeValue;

    reader = xmlReaderForMemory((const char*) dataObjectInstance, strlen((const char*) dataObjectInstance), NULL, NULL, 0);
    if (reader != NULL)
    {
      ret = xmlTextReaderRead(reader);
      while (ret)
      {
        nodeName = xmlTextReaderConstName(reader);
        map<string, MgtSnmpScalarInfo*>::iterator scalarColumn = snmpNotifInfo->mColumns.find(string((const ClCharT*) nodeName));
        if (scalarColumn != snmpNotifInfo->mColumns.end())
        {
          MgtSnmpScalarInfo *s = static_cast<MgtSnmpScalarInfo *>((*scalarColumn).second);
          oid scalarOid[CL_SNMP_MAX_OID_LEN] = {0};
          int2oid(s->oid, s->len, scalarOid);

          ret = xmlTextReaderRead(reader);
          if (ret)
          {
            nodeValue = xmlTextReaderConstValue(reader);
            stringstream valStr;
            valStr << nodeValue;

            switch (s->type)
            {
              case ASN_INTEGER:
              {
                ClInt32T scalarVal = 0;
                valStr >> scalarVal;
                snmp_varlist_add_variable(&notification_vars, scalarOid, s->len, s->type, (u_char *) &scalarVal, sizeof(scalarVal));
                break;
              }
              case ASN_COUNTER:
              case ASN_TIMETICKS:
              case ASN_UNSIGNED: /* also ASN_GAUGE */
              {
                ClUint32T scalarVal = 0;
                valStr >> scalarVal;
                snmp_varlist_add_variable(&notification_vars, scalarOid, s->len, s->type, (u_char *) &scalarVal, sizeof(scalarVal));
                break;
              }
              case ASN_OCTET_STR:
              {
                ClCharT scalarVal[CL_MAX_NAME_LENGTH] = { 0 };
                if (nodeValue)
                  strncat(scalarVal, (const ClCharT*) nodeValue, strlen((const ClCharT*) nodeValue));

                logDebug("NOT", "FIL", "Scalar value ASN_OCTET_STR [%s]", scalarVal);
                snmp_varlist_add_variable(&notification_vars, scalarOid, s->len, s->type, (u_char *) &scalarVal, strlen(scalarVal));
                break;
              }
              case ASN_OBJECT_ID:
              {
                ClUint32T scalarVal[CL_SNMP_MAX_OID_LEN] = { 0 };
                for (ClUint32T i = 0; i < strlen((const ClCharT*) nodeValue); i++)
                {
                  scalarVal[i] = ((const ClCharT*) nodeValue)[i];
                }
                snmp_varlist_add_variable(&notification_vars, scalarOid, s->len, s->type, (u_char *) &scalarVal, sizeof(scalarVal));
                break;
              }
              case ASN_COUNTER64:
              {
                ClUint64T scalarVal = 0;
                valStr >> scalarVal;
                snmp_varlist_add_variable(&notification_vars, scalarOid, s->len, s->type, (u_char *) &scalarVal, sizeof(scalarVal));
                break;
              }
              case ASN_IPADDRESS:
              {
                ClUint8T scalarVal[4] = { 0 };
                int ip[4];
                char ch;
                valStr >> ip[0] >> ch >> ip[1] >> ch >> ip[2] >> ch >> ip[3];
                for (int i = 0; i < 4; i++)
                {
                  scalarVal[i] = ip[i];
                }
                snmp_varlist_add_variable(&notification_vars, scalarOid, s->len, s->type, (u_char *) &scalarVal, sizeof(scalarVal));
                break;
              }
              case ASN_OPAQUE:
              {
                ClUint8T scalarVal = (ClUint8T) atoi((const ClCharT*) nodeValue);
                snmp_varlist_add_variable(&notification_vars, scalarOid, s->len, s->type, (u_char *) &scalarVal, sizeof(scalarVal));
                break;
              }
              default:
              {
                logDebug("SNM", "SUB", "Invalid data type [%d] for scalar [%s]", s->type, s->name.c_str());
                break;
              }
            }
          }
          break;
        }
        ret = xmlTextReaderRead(reader);
      }
      xmlFreeTextReader(reader);
    }
    return rc;
  }

  // Sending trap
  ClRcT snmpSendTrap(MgtServer::MgtSnmpNotificationInfo *snmpNotifInfo, const ClUint8T * dataObjectInstance)
  {
    netsnmp_variable_list *var_list = NULL;
    /*
     * Set the snmpTrapOid.0 value
     */
    oid notifOid[CL_SNMP_MAX_OID_LEN] = {0};
    int2oid(snmpNotifInfo->oid, snmpNotifInfo->len, notifOid);

    snmp_varlist_add_variable(&var_list, snmptrap_oid, OID_LENGTH(snmptrap_oid), ASN_OBJECT_ID, (u_char *) notifOid,
            snmpNotifInfo->len * sizeof(oid));

    snmpNotificationGetValue(snmpNotifInfo, var_list, dataObjectInstance);

    /*
     * Send the trap to the list of configured destinations and clean up
     */
    send_v2trap(var_list);
    snmp_free_varbind(var_list);
    return SNMP_ERR_NOERROR;
  }

  /**
   * The mode end function which is called after every mode is finished.
   * This is used currently to do bulk get. When the vbcount of the requests
   * is equal to the varcount.
   */
  int snmpModeEndCallback(netsnmp_mib_handler *handler, netsnmp_handler_registration *reginfo, netsnmp_agent_request_info *reqinfo,
                          netsnmp_request_info *requests)
  {
    netsnmp_request_info * request = NULL;
    static ClUint16T noOfOps = 0;

    logDebug("SNM", "OPE", "Inside MODE-END callback [%s] ...mode [%d]", __FUNCTION__, reqinfo->mode);

    switch (reqinfo->mode)
    {
      case MODE_GETBULK:
        logDebug("SNM", "OPE", "Inside mode BULK-GET");
      break;
      case MODE_GETNEXT:
      case MODE_GET:
      {
        for (request = requests; request != NULL; request = request->next)
        {
          noOfOps++;
        }
        logDebug("SNM", "OPE", "Inside mode [%s] vbC[%d] noOfOps [%d]", reqinfo->mode == MODE_GET ? "GET" : "GET-NEXT",
                reqinfo->asp->vbcount, noOfOps);

        if (reqinfo->asp->vbcount == noOfOps)
        {
          logDebug("SNP", "OPE", "Sending the get request now as the vbCount and varCount are equal.");
          noOfOps = 0;
          snmpDataReset();
        }
      }
      break;
      case MODE_SET_RESERVE1:
        logDebug("SNM", "OPE", "Inside mode RESERVE1");
      break;
      case MODE_SET_RESERVE2:
        logDebug("SNM", "OPE", "Inside mode RESERVE2");
      break;
      case MODE_SET_ACTION:
        logDebug("SNM", "OPE", "Inside mode ACTION");
      break;
      case MODE_SET_COMMIT:
        for (request = requests; request != NULL; request = request->next)
        {
          noOfOps++;
        }
        logDebug("SNM", "OPE", "Inside mode COMMIT vbC[%d] noOfOps [%d]", reqinfo->asp->vbcount, noOfOps);
        if (reqinfo->asp->vbcount == noOfOps)
        {
          logDebug("SNP", "OPE", "Sending the set requests now as the vbCount and varCount are equal.");
          noOfOps = 0;
          snmpProcessRequest(requests);
        }
      break;
      case MODE_SET_UNDO:
        logDebug("SNM", "OPE", "Inside mode UNDO");
      break;
      case MODE_SET_FREE:
        logDebug("SNM", "OPE", "Inside mode FREE");
      break;
      default:
        logDebug("SNM", "OPE", "Inside default case");
      break;
    }

    return SNMP_ERR_NOERROR;
  }

  int snmpGenericTableHandler(netsnmp_mib_handler *handler, netsnmp_handler_registration *reginfo, netsnmp_agent_request_info *reqinfo,
                              netsnmp_request_info *requests)
  {
    int rc = SNMP_ERR_NOERROR;
    netsnmp_request_info *request = NULL;
    MgtSnmpTableInfo *t = (MgtSnmpTableInfo*) reginfo->my_reg_void;

    logDebug("SNM", "SUB", "Received request for table [%s] mode [%d]", t->name.c_str(), reqinfo->mode);

    if (!requests || !reqinfo)
    {
      rc = SNMP_ERR_NOERROR;
      goto failure;
    }

    for (request = requests; request; request = request->next)
    {
      netsnmp_variable_list *reqVar = request->requestvb;
      netsnmp_table_request_info *table_info = NULL;
      netsnmp_variable_list *index = NULL;

      if (request->processed)
      {
        logDebug("SNM", "SUB", "The job is already processed, skipping this job.");
        continue;
      }

      void *data = netsnmp_extract_iterator_context(request);
      if (reqinfo->mode == MODE_GET && !data)
      {
        netsnmp_set_request_error(reqinfo, request, SNMP_NOSUCHINSTANCE);
        logDebug("SNM", "SUB", "No iterator context for the GET job, skipping it.");
        continue;
      }

      table_info = netsnmp_extract_table_info(request);
      if (!table_info)
      {
        logDebug("SNM", "SUB", "Not able to find the table_info for this job, skipping it.");
        continue;
      }

      // Lookup scalar column
      MgtSnmpScalarInfo *s = t->columns[table_info->colnum - 1];

      if (!(table_info->indexes))
      {
        rc = SNMP_ERR_GENERR;
        goto failure;
      }
      index = table_info->indexes;

      switch (reqinfo->mode)
      {
        case MODE_GET:
        {
          logDebug("SNM", "SUB", "Received GET request for column [%s] module [%s] xpath [%s]", s->name.c_str(), s->module.c_str(), s->xpath.c_str());

          string *dataContext((string *)data);
          rc = snmpScalarSetXmlData(request->requestvb, *dataContext, s->name, s->type);
          if (rc != CL_OK)
          {
            netsnmp_request_set_error(request, CL_SNMP_ERR_NOSUCHNAME);
          }
          break;
        }
        case MODE_SET_RESERVE1:
        {
          ClUint32T columnType = 0;
          int ret = 0;

          logDebug("SNM", "SUB", "Received RESERVE1 request for column [%s]", s->name.c_str());

          columnType = s->type;

          switch (columnType)
          {
            /* TODO: verification for columns for which enums, ranges are defined */
            default:
            {
              break;
            }
          }

          ret = netsnmp_check_vb_type(reqVar, columnType);
          if (ret != SNMP_ERR_NOERROR)
          {
            netsnmp_request_set_error(request, ret);
            rc = ret;
            goto failure;
          }
          break;
        }
        case MODE_SET_FREE:
        {
          logDebug("SNM", "SUB", "Received SET FREE request for column [%s]", s->name.c_str());
          break;
        }
        case MODE_SET_RESERVE2:
        {
          logDebug("SNM", "SUB", "Received SET RESERVE2 request for column [%s]", s->name.c_str());
          break;
        }
        case MODE_SET_ACTION:
        {
          logDebug("SNM", "SUB", "Received SET ACTION request for column [%s]", s->name.c_str());
          break;
        }
        case MODE_SET_COMMIT:
        {
          string xmlRequest;

          // Follow XML: <interfaces><adminStatus>true</adminStatus><ename>eth0</ename></interface>
          // Start tag
          xmlRequest.append("<").append(t->name).append(">");
          logDebug("SNM", "SUB", "Received COMMIT request for column [%s]", s->name.c_str());

          for (ClUint32T i = 0; i < t->indexes.size(); i++)
          {
            map<string, MgtSnmpScalarInfo*>::iterator indexColumn = t->mColumns.find(t->indexes[i]);
            if (indexColumn != t->mColumns.end())
            {
              MgtSnmpScalarInfo *idxScalar = static_cast<MgtSnmpScalarInfo *>((*indexColumn).second);
              // Get and filling index
              snmpScalarGetXmlData(index, idxScalar->name, idxScalar->type, xmlRequest);
            }
            index = index->next_variable;
          }

          // Filling xml for data setting
          snmpScalarGetXmlData(request->requestvb, s->name, s->type, xmlRequest);

          // End tag
          xmlRequest.append("</").append(t->name).append(">");

          // Push back to queue settings data
          {
            SAFplus::ScopedLock<SAFplus::Mutex> lock(gSnmpReqSetLock);
            dataMap.insert(pair<MgtSnmpScalarInfo *, string *>(s, new string(xmlRequest)));
          }

          logDebug("SNM", "MOD", "Scalar value setting [%s]", xmlRequest.c_str());
          break;
        }
        case MODE_SET_UNDO:
        {
          logDebug("SNM", "SUB", "Received UNDO request for column [%s]", s->name.c_str());
          break;
        }
        default:
        {
          logError("SNM", "SUB", "Unknown mode [%d] in generic table handler", reqinfo->mode);
          rc = SNMP_ERR_GENERR;
          goto failure;
        }
      }
    }

    return SNMP_ERR_NOERROR;

    failure:
    return rc;
  }

  void snmpFreeLoopTableContext(void *loopctx, netsnmp_iterator_info *iinfo)
  {
    logDebug("SNM","TBL", "Free loop table context");
    /*
     * iinfo->free_data_context gets called more than one time for the same data context.
     * To prevent this, having check before calling free
     */
    delete (std::list<std::string>*)loopctx;
  }

  void snmpXmlParserIdx(const string &dataOMXml, const char *xpathExpr, std::list<std::string> *rows)
  {
    xmlTextReaderPtr reader;
    xmlDocPtr doc;
    xmlXPathContextPtr xpathCtx;
    xmlXPathObjectPtr xpathObj;
    int ret;

    reader = xmlReaderForMemory(dataOMXml.c_str(), dataOMXml.length(), NULL, NULL, 0);
    if (reader != NULL)
    {
      /*
       * Parse and traverse the tree, collecting the nodes in the process
       */
      ret = xmlTextReaderRead(reader);
      if (ret == 0)
      {
        // TODO: Data empty!
        return;
      }

      doc = xmlTextReaderCurrentDoc(reader);

      if (doc == NULL)
      {
        // TODO: throw exception!
        return;
      }

      /* Create xpath evaluation context */
      xpathCtx = xmlXPathNewContext(doc);
      if (!xpathCtx)
      {
        // TODO: throw exception!
        xmlFreeDoc(doc);
        return;
      }

      /* Evaluate xpath expression */
      xpathObj = xmlXPathEvalExpression(BAD_CAST xpathExpr, xpathCtx);
      if(xpathObj == NULL)
      {
        // TODO: throw exception!
        xmlXPathFreeObject(xpathObj);
        xmlFreeDoc(doc);
        return;
      }

      xmlNodePtr cur;
      if (xpathObj->nodesetval->nodeNr > 0)
      {
        xmlBufferPtr buff = xmlBufferCreate();
        for(int i = 0; i < xpathObj->nodesetval->nodeNr; ++i)
        {
          cur = xpathObj->nodesetval->nodeTab[i];
          xmlNodeDump(buff, doc, cur, 0 ,0);
          rows->push_back((const char*)buff->content);
          xmlBufferEmpty(buff);
        }
        xmlBufferFree(buff);
      }

      /* Cleanup */
      /*
       * Free up the reader
       */
      xmlFreeTextReader(reader);
      xmlXPathFreeObject(xpathObj);
      xmlXPathFreeContext(xpathCtx);
      xmlFreeDoc(doc);
    }
  }

  netsnmp_variable_list *snmpGenericGetNextDataPoint(void **my_loop_context, void **my_data_context, netsnmp_variable_list *put_index_data,
                                                     netsnmp_iterator_info *mydata)
  {
    MgtSnmpTableInfo *t = (MgtSnmpTableInfo *) mydata->myvoid;
    std::list<std::string> *rows = (std::list<std::string> *)*my_loop_context;

    if (rows->empty()) 
      return NULL;

    while (!rows->empty())
    {
      string rowData = rows->back();

      //Remove processed row
      rows->pop_back();

      if (!rowData.empty())
      {
        netsnmp_variable_list *vptr;
        vptr = put_index_data;

        for (ClUint32T i = 0; i < t->indexes.size(); i++)
        {
          map<string, MgtSnmpScalarInfo*>::iterator indexColumn = t->mColumns.find(t->indexes[i]);
          if (indexColumn != t->mColumns.end())
          {
            MgtSnmpScalarInfo *s = static_cast<MgtSnmpScalarInfo *>((*indexColumn).second);
            snmpScalarSetXmlData(vptr, rowData, s->name, s->type, true);
          }
          vptr = vptr->next_variable;
        }
        string *data = new string(rowData);

        // Put and free later
        {
          SAFplus::ScopedLock<SAFplus::Mutex> lock(gSnmpReqGetLock);
          dataRows.push_back(data);
        }

        *my_data_context = data;
        return put_index_data;
      }
    }

    put_index_data = NULL;
    return NULL;
  }

  netsnmp_variable_list *snmpGenericGetFirstDataPoint(void **my_loop_context, void **my_data_context, netsnmp_variable_list *put_index_data,
                                                      netsnmp_iterator_info *mydata)
  {
    MgtSnmpTableInfo *t = (MgtSnmpTableInfo *) mydata->myvoid;
    std::list<std::string> *rows = new std::list<std::string>();

    logDebug("SNM", "TBL", "Trying table xpath [%s]", t->xpath.c_str());

    std::string xpath = "{n}";
    xpath.append(t->xpath);
    std::string xmlResult = SAFplus::mgtGet(xpath);

    if (!xmlResult.empty())
    {
      // TODO: work-around to consume top node xml
      std::stringstream ss;
      ss << "<top>" << xmlResult << "</top>";
      xmlResult.assign(ss.str());

      // Add prefix '/' (slash) to evaluation xpath
      string evalStr("/"); 
      evalStr.append(t->xpath.substr(t->xpath.find_last_of('/')));
      logDebug("SNM", "IDX", "Get data rows value for [%s]", evalStr.c_str());
      snmpXmlParserIdx(xmlResult, evalStr.c_str(), rows);
    }

    *my_loop_context = rows;

    return snmpGenericGetNextDataPoint(my_loop_context, my_data_context, put_index_data, mydata);
  }

  ClRcT snmpRegisterTable(SmiNode *smiNode, const string &moduleName)
  {
    ClRcT rc = CL_OK;
    SmiNode *childNode;
    int numColumn = 0;
    string sep("/");

    MgtMibModuleOM *moduleOM = MgtSnmpHandler::getInstance()->findModuleOM(moduleName);
    if (moduleOM == NULL)
    {
      logError("SNM", "LOAD", "Module [%s] not found!", moduleName.c_str());
      return CL_ERR_NOT_EXIST;
    }

    SmiNode *nodeRow = smiGetFirstChildNode(smiNode);

    logDebug("SNM", "TBL", "SNMP register table row [%s]", nodeRow->name);

    string xpathStr;
    MgtSnmpHandler::getParentXpathNode(xpathStr, nodeRow, moduleName, sep);
    xpathStr= xpathStr.substr(xpathStr.find('/', 1)); //TODO: trip module name

    MgtSnmpTableInfo *regSnmpTableInfo = new MgtSnmpTableInfo(moduleName, nodeRow->name, xpathStr);
    copy2oid(smiNode->oid, smiNode->oidlen, regSnmpTableInfo->oid, &regSnmpTableInfo->len);

    if (nodeRow && nodeRow->nodekind == SMI_NODEKIND_ROW)
    {
      for (childNode = smiGetFirstChildNode(nodeRow); childNode; childNode = smiGetNextChildNode(childNode))
      {
        if (childNode->nodekind == SMI_NODEKIND_COLUMN)
        {
          ++numColumn;

          string childXpathStr;
          string childName(childNode->name);

          // Change childName back to YANG if SMIv2 keyword conflict
          boost::replace_all(childName, y2mibToken, "");

          childXpathStr.append(xpathStr).append(sep).append(childName);

          MgtSnmpScalarInfo *snmpScalarInfo = new MgtSnmpScalarInfo(moduleName, childName, childXpathStr);
          copy2oid(childNode->oid, childNode->oidlen, snmpScalarInfo->oid, &snmpScalarInfo->len);

          SmiType *smiType = smiGetNodeType(childNode);

          if (smiType)
          {
            snmpScalarInfo->type = (ClUint8T) MgtSnmpHandler::getBaseTypeString(smiType->basetype);
          }
          else
          {
            snmpScalarInfo->type = ASN_NULL;
          }

          regSnmpTableInfo->columns.push_back(snmpScalarInfo);
          regSnmpTableInfo->mColumns.insert(pair<string, MgtSnmpScalarInfo*>(childName, snmpScalarInfo));
        }
      }
    }

    netsnmp_mib_handler *cb_handler = NULL;
    netsnmp_mib_handler *cb = NULL;
    netsnmp_handler_registration *reg = NULL;
    netsnmp_iterator_info *iinfo = NULL;
    netsnmp_table_registration_info *table_info = NULL;

    table_info = SNMP_MALLOC_TYPEDEF(netsnmp_table_registration_info);
    iinfo = SNMP_MALLOC_TYPEDEF(netsnmp_iterator_info);
    iinfo->flags |= NETSNMP_ITERATOR_FLAG_SORTED;

    if (!table_info || !iinfo)
    {
      SNMP_FREE(table_info);
      SNMP_FREE(iinfo);
      logError("SNM", "SUB", "Failed to create table info");
      return rc;
    }

    oid tableOid[CL_SNMP_MAX_OID_LEN] = {0};
    int2oid(regSnmpTableInfo->oid, regSnmpTableInfo->len, tableOid);

    reg = netsnmp_create_handler_registration(smiNode->name, snmpGenericTableHandler, tableOid, regSnmpTableInfo->len, HANDLER_CAN_RWRITE);
    if (!reg)
    {
      logError("SNM", "SUB", "Failed to create registration handler");
      SNMP_FREE(table_info);
      SNMP_FREE(iinfo);
      return rc;
    }

    SmiElement *iterElem;
    unsigned int numIdx = 1;
    for (iterElem = smiGetFirstElement(nodeRow); iterElem; iterElem = smiGetNextElement(iterElem))
    {
      SmiNode *smiIndexNode = smiGetElementNode(iterElem);
      SmiType *smiType = smiGetNodeType(smiIndexNode);
      if (smiType)
      {
        netsnmp_table_helper_add_index(table_info, MgtSnmpHandler::getBaseTypeString(smiType->basetype));
      }
      regSnmpTableInfo->indexes.push_back(smiIndexNode->name);
      logDebug("SNM", "REG", "Adding index [%s] for table [%s]", smiIndexNode->name, smiNode->name);
      numIdx++;
    }

    table_info->min_column = numIdx;
    table_info->max_column = numColumn;

    iinfo->get_first_data_point = snmpGenericGetFirstDataPoint;
    iinfo->get_next_data_point = snmpGenericGetNextDataPoint;
    iinfo->make_data_context = NULL;
    iinfo->free_data_context = NULL;
    iinfo->free_loop_context_at_end = snmpFreeLoopTableContext;

    iinfo->table_reginfo = table_info;
    iinfo->myvoid = regSnmpTableInfo;

    rc = netsnmp_register_table_iterator(reg, iinfo);

    cb = netsnmp_create_handler("mode_end_callback_handler", snmpModeEndCallback);
    cb->myvoid = regSnmpTableInfo;

    cb_handler = netsnmp_get_mode_end_call_handler(netsnmp_mode_end_call_add_mode_callback(NULL, NETSNMP_MODE_END_ALL_MODES, cb));
    reg->my_reg_void = regSnmpTableInfo;
    netsnmp_inject_handler(reg, cb_handler);

    if (rc != MIB_REGISTERED_OK)
    {
      logError("SNM", "REG", "Failed to register MIB table handler [%s]", smiNode->name);
    }
    else
    {
      logDebug("SNM", "REG", "Register MIB table handler [%s] successful!", smiNode->name);
    }

    moduleOM->mSnmpInfos.insert(pair<string, MgtSnmpBaseInfo*>(smiNode->name, regSnmpTableInfo));

    return rc;
  }

  /*
   * We are never called for a GETNEXT if it's registered as a
   * "instance", as it's "magically" handled for us.
   * a instance handler also only hands us one request at a time, so
   * we don't need to loop over a list of requests; we'll only get one.
   *
   */
  int snmpGenericScalarHandler(netsnmp_mib_handler *handler, netsnmp_handler_registration *reginfo, netsnmp_agent_request_info *reqinfo,
                               netsnmp_request_info *requests)
  {
    int rc = SNMP_ERR_NOERROR;
    MgtSnmpScalarInfo *s = (MgtSnmpScalarInfo*) reginfo->my_reg_void;

    logDebug("SNM", "OPE", "Generic scalar handler");

    if (!requests || !reqinfo)
    {
      return CL_SNMP_RC(CL_ERR_NULL_POINTER);
    }
    logDebug("SNM", "OPE", "Received request for scalar [%s:%s]", s->xpath.c_str(), s->name.c_str());

#if 0
    /* Find registered MOs in the database */
    MgtSnmpOM *omEntry = MgtSnmpHandler::getInstance()->findOM(s->module, s->oid, s->len);

    if (omEntry == NULL)
    {
      logError("MED", "OPE", "Instance not found!");
      return SNMP_ERR_NOERROR;
    }

    
    /* Get data of the MO */
    string dataContext;
    MgtSnmpHandler::getInstance()->mgtSnmpDataGet(omEntry, dataContext);
#endif
    std::string xpath = "{n}";
    xpath.append(s->xpath);
    std::string dataContext = SAFplus::mgtGet(xpath);

    switch (reqinfo->mode)
    {
      case MODE_GET:
      {
        logDebug("SNM", "OPE", "Received GET request");
        snmpScalarSetXmlData(requests->requestvb, dataContext, s->name, s->type);
        break;
      }
      case MODE_SET_RESERVE1:
      {
        int ret = 0;
        ClUint8T scalarType = s->type;

        logDebug("SNM", "SUB", "Received RESERVE1 request");

        ret = netsnmp_check_vb_type(requests->requestvb, scalarType);
        if (ret != SNMP_ERR_NOERROR)
        {
          netsnmp_request_set_error(requests, ret);
          rc = ret;
          goto failure;
        }
        break;
      }
      case MODE_SET_FREE:
      {
        logDebug("SNM", "SUB", "Received SET FREE request");
        break;
      }
      case MODE_SET_RESERVE2:
      {
        logDebug("SNM", "SUB", "Received RESERVE2 request");
        break;
      }
      case MODE_SET_ACTION:
      {
        break;
      }
      case MODE_SET_COMMIT:
      {
        logDebug("SNM", "SUB", "Received COMMIT request");
        break;
      }
      case MODE_SET_UNDO:
      {
        logDebug("SNM", "SUB", "Received UNDO request");
        break;
      }
      default:
      {
        logError("SNM", "SUB", "Unknown mode [%d] in generic scalar handler", reqinfo->mode);
        rc = SNMP_ERR_GENERR;
        goto failure;
      }
    }

    return SNMP_ERR_NOERROR;
    failure:
      return rc;
  }

  ClRcT snmpRegisterScalar(SmiNode *smiNode, const string &moduleName)
  {
    ClRcT rc = CL_OK;
    string sep("/");

    MgtMibModuleOM *moduleOM = MgtSnmpHandler::getInstance()->findModuleOM(moduleName);
    if (moduleOM == NULL)
    {
      logError("SNM", "LOAD", "Module [%s] not found!", moduleName.c_str());
      return CL_ERR_NOT_EXIST;
    }

    string xpathStr;
    string nodeName(smiNode->name);
    boost::replace_all(nodeName, y2mibToken, "");

    MgtSnmpHandler::getParentXpathNode(xpathStr, smiNode, moduleName, sep);
    xpathStr= xpathStr.substr(xpathStr.find('/', 1)); //TODO: trip module name

    netsnmp_mib_handler *cb_handler = NULL;
    netsnmp_handler_registration *reg = NULL;

    MgtSnmpScalarInfo *regSnmpScalarInfo = new MgtSnmpScalarInfo(moduleName, nodeName, xpathStr);
    copy2oid(smiNode->oid, smiNode->oidlen, regSnmpScalarInfo->oid, &regSnmpScalarInfo->len);

    SmiType *smiType = smiGetNodeType(smiNode);
    regSnmpScalarInfo->type = (ClUint8T) MgtSnmpHandler::getBaseTypeString(smiType->basetype);

    oid scalarOid[CL_SNMP_MAX_OID_LEN] = {0};
    int2oid(regSnmpScalarInfo->oid, regSnmpScalarInfo->len, scalarOid);
    reg = netsnmp_create_handler_registration(nodeName.c_str(), snmpGenericScalarHandler, scalarOid, regSnmpScalarInfo->len,
            HANDLER_CAN_RWRITE);
    rc = netsnmp_register_scalar(reg);

    cb_handler = netsnmp_get_mode_end_call_handler(netsnmp_mode_end_call_add_mode_callback(NULL,
    NETSNMP_MODE_END_ALL_MODES, netsnmp_create_handler("my_end_callback_handler", snmpModeEndCallback)));
    reg->my_reg_void = regSnmpScalarInfo;
    netsnmp_inject_handler(reg, cb_handler);

    if (rc != MIB_REGISTERED_OK)
    {
      logError("SNM", "REG", "Failed to register MIB scalar handler [%s] rc=[%#x]", nodeName.c_str(), rc);
    }
    else
    {
      logDebug("SNM", "REG", "Register MIB scalar handler [%s] successful!", nodeName.c_str());
    }

    moduleOM->mSnmpInfos.insert(pair<string, MgtSnmpBaseInfo*>(nodeName, regSnmpScalarInfo));

    return rc;
  }

  /*
   * Declare notification
   */
  ClRcT snmpRegisterNotification(SmiNode *smiNode, const string &moduleName)
  {
    ClRcT rc = CL_OK;
    SmiElement *iterElem;
    string sep("/");

    MgtMibModuleOM *moduleOM = MgtSnmpHandler::getInstance()->findModuleOM(moduleName);
    if (moduleOM == NULL)
    {
      logError("SNM", "LOAD", "Module [%s] not found!", moduleName.c_str());
      return CL_ERR_NOT_EXIST;
    }

    string oidStr;
    string xpathStr;

    oid2str(oidStr, smiNode->oid, smiNode->oidlen);
    MgtSnmpHandler::getParentXpathNode(xpathStr, smiNode, moduleName, sep);
    xpathStr= xpathStr.substr(xpathStr.find('/', 1)); //TODO: trip module name

    logDebug("SNM", "REG", "SNMP register notification [%s] oid [%s]", smiNode->name, oidStr.c_str());

    MgtServer::MgtSnmpNotificationInfo *regSnmpNotifInfo = new MgtServer::MgtSnmpNotificationInfo(moduleName, smiNode->name, xpathStr);
    copy2oid(smiNode->oid, smiNode->oidlen, regSnmpNotifInfo->oid, &regSnmpNotifInfo->len);

    for (iterElem = smiGetFirstElement(smiNode); iterElem; iterElem = smiGetNextElement(iterElem))
    {

      SmiNode *childNode = smiGetElementNode(iterElem);
      string childXpathStr;
      childXpathStr.append(xpathStr).append(sep).append(childNode->name);

      MgtSnmpScalarInfo *snmpScalarInfo = new MgtSnmpScalarInfo(moduleName, childNode->name, childXpathStr);
      copy2oid(childNode->oid, childNode->oidlen, snmpScalarInfo->oid, &snmpScalarInfo->len);

      SmiType *smiType = smiGetNodeType(childNode);

      if (smiType)
      {
        snmpScalarInfo->type = (ClUint8T) MgtSnmpHandler::getBaseTypeString(smiType->basetype);
      }
      else
      {
        snmpScalarInfo->type = ASN_NULL;
      }

      regSnmpNotifInfo->columns.push_back(snmpScalarInfo);
      regSnmpNotifInfo->mColumns.insert(pair<string, MgtSnmpScalarInfo*>(childNode->name, snmpScalarInfo));
    }

    moduleOM->mNotifOMList.insert(pair<string, MgtServer::MgtSnmpNotificationInfo*>(oidStr, regSnmpNotifInfo));

    return rc;
  }

  /**
   * MgtNetconfHandle constructor
   */
  MgtSnmpHandler::MgtSnmpHandler()
  {
    /* Init libxml */
    xmlInitParser();
  }

  /**
   * MgtNetconfHandle destructor
   */
  MgtSnmpHandler::~MgtSnmpHandler()
  {
    // Cleanup smi
    smiExit();

    /* Shutdown libxml */
    xmlCleanupParser();

    // Clean map modules manage
    destroyMap(mModuleOMs);
  }

  /**
   * Singleton instance of mgtNetconfHandle
   */
  MgtSnmpHandler* MgtSnmpHandler::getInstance()
  {
    if (instance == NULL)
    {
      instance = new MgtSnmpHandler();
    }
    return instance;
  }

  void MgtSnmpHandler::DestroyInstance()
  {
    delete instance;
    instance = 0;
  }

  ClRcT MgtSnmpHandler::snmpInitialize(int argc, ClCharT *argv[])
  {
    int logSev = (SAFplus::logSeverity - 1); // Convert from SAFplus's logSeverity to SMI's errorSeverity

    smiInit("snmpsmi");
    smiSetErrorHandler(smiErrorCallback);
    smiSetErrorLevel(logSev);

    /* MIB paths separate to lookup */
    char *mibsEnvPath = NULL;
    string curDir(boost::filesystem::current_path().string());
    string mibdirs(curDir);

    // ENV MIBDIRS if exporting
    if ((mibsEnvPath = SAFplus::clParseEnvStr("MIBDIRS")) != NULL)
    {
      mibdirs.append(":").append(mibsEnvPath);
    }

    if (1)
    {
      mibdirs.append(":").append(curDir).append("/share/snmp/mibs");
      mibdirs.append(":").append(curDir).append("/../share/snmp/mibs");
    }

    logDebug("SNM", "SRV","MIB search path: %s", mibdirs.c_str());
    smiSetPath(mibdirs.c_str());

    keep_snmpAgent_running = true;
    logDebug("SNM", "INIT", "SNMP master agentx initialize done!");
    return CL_OK;
  }

  ClRcT MgtSnmpHandler::CheckAndLoadMibModule(std::vector<std::string> &ckptNew, const std::vector<std::string> &paths, std::map<std::string,bool> *mapLoadFailed)
  {
    ClRcT rc = CL_OK;
    std::vector<std::string>::iterator iterCkpt;
    bool checkFailed;

    for (iterCkpt = ckptNew.begin(); iterCkpt!= ckptNew.end(); ++iterCkpt)
    {
      checkFailed = true;
      /*
       * Lookup throughout MIB searh paths location
       */
      for (std::vector<std::string>::const_iterator iterDir = paths.begin(); iterDir != paths.end(); ++iterDir)
      {
        boost::filesystem::path mibfile(*iterDir + "/" + toUpper(*iterCkpt) + "-MIB.txt");
        logDebug("MIB", "CHK", "Searching MIB file [%s] ...", mibfile.string().c_str());

        SmiModule *mibM = smiGetModule(mibfile.string().c_str());
        /* Sanity checking MIB file */
        if (mibM)
        {
            logDebug("MIB", "LOAD", "Loading MIB file [%s] [%s]", mibfile.stem().string().c_str(), mibfile.c_str());

            /* Load MIB module and register with snmp deamon */
            MgtMibModuleOM *moduleOM = MgtSnmpHandler::getInstance()->findModuleOM(mibfile.stem().string());
            if (moduleOM == NULL) 
            {
              moduleOM = new MgtMibModuleOM();
              MgtSnmpHandler::getInstance()->mModuleOMs.insert(pair<string, MgtMibModuleOM*>(mibfile.stem().string(), moduleOM));
              MgtSnmpHandler::getInstance()->LoadMIBModule(mibfile.stem().string());
            }
            checkFailed = false;
          }
          else
            logNotice("MIB", "LOAD", "Loading MIB file [%s] failed", mibfile.string().c_str());
      }
      /*
       * Marked as failed to re-check at next polling
       */
      if (checkFailed) mapLoadFailed->insert(pair<string, bool>(*iterCkpt,checkFailed));
    }
    return rc;
  }

  /**
   * Bind request handle
   */
  void MgtSnmpHandler::mgtMsgBindHandle(SAFplus::Handle from, void *pInMsg)
  {
    ClRcT rc = CL_OK;

    Mgt::Msg::MsgBind *bindData = (Mgt::Msg::MsgBind *) pInMsg;

    logDebug("SNM", "BIND", "MIB module [%s] route [%s] handle [%#" PRIx64 ":%#" PRIx64 "]", bindData->module().c_str(),
            bindData->route().c_str(), bindData->handle().id0(), bindData->handle().id1());

    /* Load MIB module and register with snmp deamon */
    MgtMibModuleOM *moduleOM = MgtSnmpHandler::getInstance()->findModuleOM(bindData->module());
    if (moduleOM == NULL)
    {
      moduleOM = new MgtMibModuleOM();
      MgtSnmpHandler::getInstance()->mModuleOMs.insert(pair<string, MgtMibModuleOM*>(bindData->module(), moduleOM));
      MgtSnmpHandler::getInstance()->LoadMIBModule(bindData->module().c_str());
    }

    MgtSnmpOM *compBindTemp = new MgtSnmpOM(bindData);

    /* Convert to real snmp oid */
    str2oid(bindData->route(), compBindTemp->oid, &compBindTemp->len);

    MgtSnmpOM *compBind = MgtSnmpHandler::getInstance()->findOM(bindData->module(), compBindTemp->oid, compBindTemp->len);

    /* Store data binding for lookup */
    if (compBind == NULL)
    {
      moduleOM->mOMList.insert(pair<string, MgtSnmpOM *>(bindData->route(), compBindTemp));
    }
    else /* Update route table */
    {
      Mgt::Msg::Handle *hdl = compBind->bindData.mutable_handle();
      hdl->set_id0(bindData->handle().id0());
      hdl->set_id1(bindData->handle().id1());
      delete (compBindTemp);
    }

    SAFplus::SafplusMsgServer *mgtIocInstance = &safplusMsgServer;
    mgtIocInstance->SendMsg(from, (void *) &rc, sizeof(rc), SAFplusI::CL_MGT_MSG_TYPE);
  }
  /**
   * Notification handle
   */
  void MgtSnmpHandler::mgtMsgNotifHandle(SAFplus::Handle from, void *pInMsg)
  {
    Mgt::Msg::MsgMgt *notify = (Mgt::Msg::MsgMgt *) pInMsg;
    logDebug("SNM", "MSG", "Receive notification message");
    if (notify->data_size() == 0)
    {
      logError("SNM", "NOTI", "No data in notification");
      return;
    }
    Mgt::Msg::MsgBind bindData;
    bindData.ParseFromString(notify->bind());

    Mgt::Msg::MsgSetGet notiData;
    notiData.ParseFromString(notify->data(0));

    MgtSnmpNotificationInfo *snmpNotifInfo = MgtSnmpHandler::getInstance()->findNotifOM(bindData.module(), bindData.route());
    if (snmpNotifInfo != NULL)
    {
      snmpSendTrap(snmpNotifInfo, (const ClUint8T *) notiData.data().c_str());
    }
    else
    {
      logNotice("SNM", "NOTI", "Notification [%s] not found in [%s]", bindData.route().c_str(), bindData.module().c_str());
    }
  }

  string MgtSnmpHandler::toUpper(const string &str)
  {
    string result;
    unsigned int i;
    std::locale loc;

    for (i = 0; i < str.length(); i++)
    {
      result += std::toupper(str.at(i), loc);
    }
    return result;
  }

  string MgtSnmpHandler::toCamelcase(const string &str)
  {
    string result;
    unsigned int i = 0;
    bool isUpcase = false;
    std::locale loc;

    while (i < str.length())
    {
      if (str[i] == '-')
      {
        isUpcase = true;
      }
      else
      {
        if (isUpcase)
        {
          result += std::toupper(str.at(i), loc);
        }
        else
        {
          result += std::tolower(str.at(i), loc);
        }
        isUpcase = false;
      }
      i++;
    }
    return result;
  }

  ClCharT MgtSnmpHandler::getBaseTypeString(SmiBasetype basetype)
  {
    switch (basetype)
    {
      case SMI_BASETYPE_UNKNOWN:
        return ASN_NULL;
      case SMI_BASETYPE_POINTER:
        return ASN_NULL;
      case SMI_BASETYPE_INTEGER32:
      case SMI_BASETYPE_ENUM:
        return ASN_INTEGER;
      case SMI_BASETYPE_OCTETSTRING:
      case SMI_BASETYPE_BITS:
        return ASN_OCTET_STR;
      case SMI_BASETYPE_OBJECTIDENTIFIER:
        return ASN_OBJECT_ID;
      case SMI_BASETYPE_UNSIGNED32:
        return ASN_UNSIGNED;
      case SMI_BASETYPE_INTEGER64:
        return ASN_INTEGER;
      case SMI_BASETYPE_UNSIGNED64:
        return ASN_UNSIGNED64;
      default:

      break;
    }

    return ASN_NULL;
  }

  void MgtSnmpHandler::getParentXpathNode(string &name, SmiNode *node, const string &moduleName, string &sep)
  {
    SmiNode *parent = smiGetParentNode(node);
    if (parent != NULL && parent->nodekind == SMI_NODEKIND_TABLE)
    {
      parent = smiGetParentNode(parent);
    }

    if (parent != NULL && strcmp(moduleName.c_str(), node->name) && strcmp("enterprises", node->name))
    {
      getParentXpathNode(name, parent, moduleName, sep);
      name.append(sep).append(node->name);
    }
  }

  void MgtSnmpHandler::registerSMIModule(SmiModule *smiModule, const string &moduleName)
  {
    SmiNode *smiNode;
    SmiNodekind nodeKind = SMI_NODEKIND_SCALAR | SMI_NODEKIND_TABLE | SMI_NODEKIND_NOTIFICATION;

    for (smiNode = smiGetFirstNode(smiModule, nodeKind); smiNode; smiNode = smiGetNextNode(smiNode, nodeKind))
    {
      if (smiNode->nodekind == SMI_NODEKIND_TABLE)
      {
        snmpRegisterTable(smiNode, moduleName);
      }
      else if (smiNode->nodekind == SMI_NODEKIND_NOTIFICATION)
      {
        snmpRegisterNotification(smiNode, moduleName);
      }
      else
      {
        snmpRegisterScalar(smiNode, moduleName);
      }
    }
  }

  void MgtSnmpHandler::LoadMIBModule(const string &name)
  {
    SmiModule *smiModule = smiGetModule(name.c_str());
    if (smiModule)
    {
      MgtSnmpHandler::registerSMIModule(smiModule, name);
    }
  }

  MgtMibModuleOM* MgtSnmpHandler::findModuleOM(const string &module)
  {
    map<string, MgtMibModuleOM*>::iterator moduleIndex = mModuleOMs.find(module);
    if (moduleIndex != mModuleOMs.end())
    {
      return static_cast<MgtMibModuleOM *>((*moduleIndex).second);
    }
    return NULL;
  }

  MgtSnmpOM* MgtSnmpHandler::findOM(const string &module, const ClUint32T *oid, ClUint32T len)
  {
    MgtMibModuleOM *moduleOM = findModuleOM(module);
    if (moduleOM != NULL)
    {
      map<string, MgtSnmpOM*>::iterator omEntry = moduleOM->mOMList.begin();
      while (omEntry != moduleOM->mOMList.end())
      {
        MgtSnmpOM *snmpOMEntry = static_cast<MgtSnmpOM *>((*omEntry).second);
        if (snmpOMEntry && snmpOMEntry->len <= len && !memcmp(snmpOMEntry->oid, oid, snmpOMEntry->len * sizeof(ClUint32T)))
          return snmpOMEntry;

        omEntry++;
      }
    }
    return NULL;
  }

  MgtSnmpNotificationInfo *MgtSnmpHandler::findNotifOM(const std::string &module, const std::string &route)
  {
    MgtMibModuleOM *moduleOM = findModuleOM(module);
    if (moduleOM != NULL)
    {
      map<string, MgtSnmpNotificationInfo*>::iterator notiEntry = moduleOM->mNotifOMList.find(route);
      if (notiEntry != moduleOM->mNotifOMList.end())
      {
        return static_cast<MgtSnmpNotificationInfo *>((*notiEntry).second);
      }
    }
    return NULL;
  }

  /*
   * Find all matches oid with these objects binding have the same route
   */
  void MgtSnmpHandler::findListOM(const string &module, const ClUint32T *oid, const ClUint32T len, std::list<MgtSnmpOM*> &elems)
  {
    MgtMibModuleOM *moduleOM = findModuleOM(module);
    if (moduleOM != NULL)
    {
      map<string, MgtSnmpOM*>::iterator omEntry = moduleOM->mOMList.begin();
      while (omEntry != moduleOM->mOMList.end())
      {
        MgtSnmpOM *snmpOMEntry = static_cast<MgtSnmpOM *>((*omEntry).second);
        if (snmpOMEntry->len <= len && !memcmp(snmpOMEntry->oid, oid, snmpOMEntry->len * sizeof(ClUint32T)))
        {
          elems.push_back(snmpOMEntry);
        }
        omEntry++;
      }
    }
    else
    {
      logDebug("SNM", "LKO", "Not found module %s", module.c_str());
    }
  }

  void MgtSnmpHandler::findListOM(SAFplus::Handle findHdl, std::list<MgtSnmpOM*> &elems)
  {
    map<string, MgtMibModuleOM*>::iterator moduleIndex = mModuleOMs.begin();
    while (moduleIndex != mModuleOMs.end())
    {
      MgtMibModuleOM *moduleOM = static_cast<MgtMibModuleOM *>((*moduleIndex).second);
      map<string, MgtSnmpOM*>::iterator omIndex = moduleOM->mOMList.begin();
      while (omIndex != moduleOM->mOMList.end())
      {
        MgtSnmpOM *omEntry = static_cast<MgtSnmpOM *>((*omIndex).second);
        SAFplus::Handle hdl;
        hdl.id[0] = omEntry->bindData.handle().id0();
        hdl.id[1] = omEntry->bindData.handle().id1();
        if (hdl.getNode() == findHdl.getNode())
        {
          switch (findHdl.getPort())
          {
            case 0: // Case node address
              elems.push_back(omEntry);
            break;
            default: // Case node and port
              if (hdl.getPort() == findHdl.getPort())
              {
                elems.push_back(omEntry);
              }
            break;
          }
        }

        omIndex++;
      }
      moduleIndex++;
    }
  }

#if 0
  void MgtSnmpHandler::mgtSnmpDataGet(MgtSnmpOM *omEntry, string &output)
  {
    Mgt::Msg::MsgMgt mgtMsgReq;
    Mgt::Msg::MsgSetGet setData;
    string oidStr;
    oid2str(oidStr, omEntry->oid, omEntry->len);

    findCacheData(oidStr, output);
    if (!output.empty())
    {
      return;
    }

    string strBind, request, strSet;
    stringstream ss;
    ss << omEntry->rev;
    string strRev = ss.str();

    SAFplus::Handle hdl;
    hdl.id[0] = omEntry->bindData.handle().id0();
    hdl.id[1] = omEntry->bindData.handle().id1();

    omEntry->bindData.SerializeToString(&strBind);

    mgtMsgReq.set_type(Mgt::Msg::MsgMgt::CL_MGT_MSG_GET);
    mgtMsgReq.set_bind(strBind);
    //Send mine data revision
    mgtMsgReq.add_data(strRev);
    mgtMsgReq.SerializeToString(&request);

    SAFplus::SafplusMsgServer* mgtIocInstance = &safplusMsgServer;
    try
    {
      SAFplus::ScopedLock<SAFplus::Mutex> lock(sendReplySnmpMutex);
      MsgReply *msgReply = mgtIocInstance->sendReply(hdl, (void *) request.c_str(), request.size(), SAFplusI::CL_MGT_MSG_TYPE);
      Mgt::Msg::MsgGeneral rxMsg;
      rxMsg.ParseFromArray(msgReply->buffer, strlen(msgReply->buffer));
      if (rxMsg.data_size() == 1) // Received revision only
      {
        omEntry->read(output);
      }
      else if (rxMsg.data_size() == 2) //Received revision and data
      {
        strRev = rxMsg.data(0);
        // Clear string stream
        ss.str("");
        ss << strRev;
        // Update new revision
        ss >> omEntry->rev;
        output.assign(rxMsg.data(1));
        omEntry->write(output);
      }
      else
      {
        logDebug("MGT", "GET", "Invalid message with data size %d", rxMsg.data_size());
      }
    }
    catch (SAFplus::Error &ex)
    {
      logDebug("MGT", "GET", "Try to used old data but failed");
    }

    // TODO: protect share resource data
    updateCacheData(oidStr, output);
  }


  ClRcT MgtSnmpHandler::mgtSnmpDataSet(const MgtSnmpOM *omEntry, string &data)
  {
    ClRcT ret = CL_OK;
    string bindStr, requestStr, setStr;

    SAFplus::Handle hdl;
    hdl.id[0] = omEntry->bindData.handle().id0();
    hdl.id[1] = omEntry->bindData.handle().id1();

    Mgt::Msg::MsgMgt msgRequest;
    Mgt::Msg::MsgSetGet msgSet;

    logDebug("SNM", "MOD", "Data setting instance object: [%s]", data.c_str());

    // Packing the request
    omEntry->bindData.SerializeToString(&bindStr);
    msgSet.set_data(data);
    msgSet.SerializeToString(&setStr);
    msgRequest.set_type(Mgt::Msg::MsgMgt::CL_MGT_MSG_SET);
    msgRequest.set_bind(bindStr);
    msgRequest.add_data(setStr);
    msgRequest.SerializeToString(&requestStr);

    SAFplus::SafplusMsgServer* mgtIocInstance = &safplusMsgServer;
    try
    {
      SAFplus::ScopedLock<SAFplus::Mutex> lock(sendReplySnmpMutex);
      SAFplus::MsgReply *msgReply = mgtIocInstance->sendReply(hdl, (void *) requestStr.c_str(), requestStr.size());
      return *(ClRcT *) msgReply->buffer;
    }
    catch (SAFplus::Error &ex)
    {
      ret = ex.clError;
      logError("SNM", "MOD", "Send SET message failed rc =[0x%x]", ret);
      return ret;
    }
  }
#endif
}
